/*! \file ddag_impl.hpp
 *  \brief Definitions of the object DDAG_impl used by the user. Three template specializations are available.
 *  DDAG_impl is the structure of the tree, not data. data are stored in DProperty objects.
 */
#ifndef DDAG_IMPL_H
#define DDAG_IMPL_H

#include "../../util/mpi_.hpp"
#include "../../iterators/DDAG/iterator.hpp"
#include <math.h>
#include<vector>

namespace skelgis{
  
  //-------------------------------------------------------------------------------
  //!  base template class of Distributed DAG
  /*!
    This template class defines the base distributed dag
    This basic template does not have an implementation, only specializations have
    direction of the DAG is not usefull the input file of the user precise the direction
    no parameter on the physical border because roots and leafs nodes are always managed in a different way than standard elements (different neighborhood etc.)
    \tparam node is a boolean parameter to indicate if computations on nodes are needed for this DDAG_impl
    \tparam edge is a boolean parameter to indicate if computations on edges are needed for this DDAG_impl
  */
  //-------------------------------------------------------------------------------
  template<bool node, bool edge> struct DDAG_impl;
  //-------------------------------------------------------------------------------

  //-------------------------------------------------------------------------------
  //!  First specialization of the DDAG_impl class
  /*!
    This is the more complex case where both nodes and edges have to be computed
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG_impl<true,true>
  //-------------------------------------------------------------------------------
  {
  protected:

    /*!< to keep the glocal size of the DAG */
    unsigned int dim_edge_g;
    unsigned int dim_node_g;

    /*!< to manage physical border
      physical border in DDAG_impl for scientific simulations are leafs and roots of the global DDAG_impl */
    unsigned int dim_roots; //! number of roots elements in the local DDAG_impl (global notion of roots)
    unsigned int dim_leafs; //! number of leafs elements int the local DDAG_impl (global notion of leafs)
    /*!< As distribution is not perfect, especially for huge DDAG_impl, it is important to make an overlap of communications by local computations
      For this reason pure local elements (except roots and leafs) are distinguished from elements that need communications */
    unsigned int dim_loc; //! number of internal nodes without any communication
    unsigned int dim_comm; //! number of internal nodes with communications
    
    //DATA STRUCTURE
    //nodes
    unsigned int dim_node; //! number of nodes
    /*!< cumulative degrees for inside neighbors nodes and edges
      foreach node you can get the number of inside nodes and inside edges (same) neighbors */
    unsigned int *degcum_in;
    /*!< cumulative degrees for outside neighbors nodes and edges
      foreach node you can get the number of outside nodes and inside edges (same) neighbors */
    unsigned int *degcum_out;
    /*!< associated table to degcum_in
      This table is the list of input neighbors nodes indexes for each node of the local DDAG_impl */
    unsigned int *neigh_in;
    /*!< associated table to degcum_out
      This table is the list of output neighbors nodes indexes for each node of the local DDAG_impl */
    unsigned int *neigh_out;
    /*!< associated table to degcum_in
      This table is the list of input neighbors edges indexes for each node of the local DDAG_impl */
    unsigned int *edges_out;
    /*!< associated table to degcum_out
      This table is the list of output neighbors edges indexes for each node of the local DDAG_impl */
    unsigned int *edges_in;
    
    //DATA STRUCTURE
    //edges
    unsigned int dim_edge; //! number of edges
    unsigned int *src_nodes; //! source node of each edge
    unsigned int *dst_nodes; //! destination node of each edge

    //DATA STRUCTURE
    //indexation
    unsigned int dim_ind_ed,dim_ind_no; //! number of indexes for nodes and edges (with additionnal communication indexes)
    /*!< Associations between new local nodes indexes values and nodes indexes of the global DAG
      Order of the new nodes indexes are very important for a transparent iteration on local elements
      All local nodes tables use those new indexes (neighborhood etc)
      indexes are stored this way :
      ******************************************************************************************************************************************
      ||  pure local elements  |  elements with communications  |  leafs  |  roots  |  comunicated elements from other processors IN then OUT ||
      ******************************************************************************************************************************************
     */
    unsigned int *indexes_nodes;
    /*!< Associations between new local edges indexes values and edges indexes of the global DAG 
      Edges are indexed following the nodes indexes
      output edges of Node 0 are indexed from 0 to degcum_out[0]
      output edges of Node k are indexed from degcum_out[k-1] to degcum_out[k]
    */
    unsigned int *indexes_edges;

    //DATA STRUCTURE
    //parallelization
    /*!< Number of elements to receive : outputs and inputs elements */
    unsigned int dim_tor_out,dim_tor_in;
    /*!< Number of elements to send : outputs and inputs elements */
    unsigned int dim_tos_out,dim_tos_in;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of output and input elements to send */
    unsigned int *degcum_tos;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of elements to send INSIDE another processor (this processor is up the other processor)
     for now difference between input and output is made for edges only. It is needed for DPMap_edges on pointers types */
    unsigned int * degcum_tos_in;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of elements to send OUTSIDE another processor (this processor is down the other processor)
     for now difference between input and output is made for edges only. It is needed for DPMap_edges on pointers types */
    unsigned int * degcum_tos_out;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of output and input elements to receive */
    unsigned int *degcum_tor;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of input elements to receive 
     for now difference between input and output is made for edges only. It is needed for DPMap_edges on pointers types*/
    unsigned int *degcum_tor_in;
    /*!< For each processor of the current MPI execution, it represents the cumulative degrees of the number of output elements to receive 
     for now difference between input and output is made for edges only. It is needed for DPMap_edges on pointers types*/
    unsigned int *degcum_tor_out;
    /*!< associated table to degcum_tos for edges
      This table is the list of edges local indexes to send to each other processor 
      */
    //unsigned int *edges_tos;
    /*!< associated table to degcum_tos_in for edges
      This table is the list of edges to send that goes in another processor, local indexes to send to each other processor */
    unsigned int *edges_tos_in;
    /*!< associated table to degcum_tos_out for edges
      This table is the list of edges to send that goes out of another processor, local indexes to send to each other processor */
    unsigned int *edges_tos_out;
    /*!< associated table to degcum_tos for nodes
      This table is the list of nodes local indexes to send to each other processor */
    unsigned int *nodes_tos;
    /*!< associated table to degcum_tor for edges
      This table is the list of edges local indexes to receive to each other processor  */
    //unsigned int *edges_tor;
    /*!< associated table to degcum_tor_in for edges
      This table is the list of edges local indexes to receive in from each other processor  */
    unsigned int *edges_tor_in;
    /*!< associated table to degcum_tor_out for edges
      This table is the list of edges local indexes to receive out from each other processor  */
    unsigned int *edges_tor_out;
    /*!< associated table to degcum_tos for nodes
      This table is the list of nodes local indexes to receive to each other processor  */
    unsigned int *nodes_tor;

    //-------------------------------------------------------------------------------
    //! find position tool
    /*!
      Find the position of element in table
      \param table is the pointer table in which the element has to be searched
      \param size is the size of the table
      \param element is the element value to search
      \param position is a reference to the position found
      \return true if the position has been found, false otherwise
    */
    //-------------------------------------------------------------------------------
    bool find_position(unsigned int * table,unsigned int size,unsigned int element,unsigned int &position)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<size;i++)
	{
	  if(table[i]==element)
	    {
	      position=i;
	      return true;
	    }
	}
      return false;
    }
    //-------------------------------------------------------------------------------
    //! find position tool
    /*!
      Find the position of element in table but do not get the position value
      \param table is the pointer table in which the element has to be searched
      \param size is the size of the table
      \param element is the element value to search
      \return true if the position has been found, false otherwise
    */
    //-------------------------------------------------------------------------------
    bool find_position(unsigned int * table,unsigned int size,unsigned int element)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<size;i++)
	{
	  if(table[i]==element)
	      return true;
	}
      return false;
    }
    //-------------------------------------------------------------------------------
    //! find from tool
    /*!
      Search an element in a table with two positions between which to search
      \param table is the pointer table in which the element has to be searched
      \param position1 is the index value in the table where to start the search
      \param position2 is the index value in the table where to end the search (included)
      \param element is the element value to search in table
      \return true if the position has been found, false otherwise
    */
    //-------------------------------------------------------------------------------
    bool find_from(unsigned int * table,unsigned int position1,unsigned int position2,unsigned int element)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=position1;i<position2+1;i++)
	{
	  if(table[i]==element)
	      return true;
	}
      return false;
    }
    //-------------------------------------------------------------------------------
    //! get maximum value tool
    /*!
      search the minimum value between three values
      \param a is the first value
      \param b is the second value
      \param c is the third value
      \return the minimum value between a, b and c
    */
    //-------------------------------------------------------------------------------
    unsigned int max(unsigned a,unsigned b,unsigned c)
    //-------------------------------------------------------------------------------
    {
      a=(a>b) ? a : b;
      return (a>c) ? a : c;
    }
    //-------------------------------------------------------------------------------
    //! recursively increment the number of time a subtree has been visited
    /*!
      Is used to count the number of total edges under a subtree
      \param index is the index of the root element of the subtree
      \param visited is the table to increment for the subtree
      \param degcum_block is the cumulative degrees for links between blocks of sister edges (equivalent of degcum_out for blocks)
      \param linked_blocks is the associated table to degcum_block, it is the list of neighbor blocks
    */
    //-------------------------------------------------------------------------------
    void is_visited(unsigned int index, unsigned int * visited, unsigned int * degcum_block, unsigned int *linked_blocks)
    //-------------------------------------------------------------------------------
    {
      unsigned int nb = degcum_block[index+1]-degcum_block[index];
      for(unsigned int i=0;i<nb;i++)
	  is_visited(linked_blocks[degcum_block[index]+i],visited,degcum_block,linked_blocks);
      visited[index]+=1;
    }
    //-------------------------------------------------------------------------------
    //! recursively count the number of edges in a subtree
    /*!
      Is used to modify is_visited error
      \param index is the index of the root element of the subtree
      \param degcum_block is the cumulative degrees for links between blocks of sister edges (equivalent of degcum_out for blocks)
      \param linked_blocks is the associated table to degcum_block, it is the list of neighbor blocks
      \param edge_block is the cumulative degrees for the number of edges in each block of sister edges
      \return The number of edges in the subtree
    */
    //-------------------------------------------------------------------------------
    unsigned int all_edges_under(unsigned int index, unsigned int * degcum_block, unsigned int *linked_blocks, unsigned int * weight_block)
    //-------------------------------------------------------------------------------
    {
      unsigned int nb = degcum_block[index+1]-degcum_block[index];
      unsigned int res = weight_block[index];
      for(unsigned int i=0;i<nb;i++)
	res+=all_edges_under(linked_blocks[degcum_block[index]+i],degcum_block,linked_blocks,weight_block);

      return res;
    }
    //-------------------------------------------------------------------------------
    //! Take all the subtree for the same processor
    /*!
      Used in the distribution of the DDAG_impl, if the number of edges in the subtree < ideal take it entirely
      \param index is the index of the root element of the subtree
      \param edge_block is the cumulative degrees for the number of edges in each block of sister edges
      \param degcum_block is the cumulative degrees for links between blocks of sister edges (equivalent of degcum_out for blocks)
      \param linked_blocks is the associated table to degcum_block, it is the list of neighbor blocks
      \param taken is the table to note to which processor is assigned each block
      \param count is the reference to the count of edges for the current processor
      \param cur_proc is the current processor index
    */
    //-------------------------------------------------------------------------------
    void takeall(unsigned int index,unsigned int *edge_block,unsigned int *weight_block,unsigned int *degcum_block,unsigned int *linked_blocks,unsigned int *taken,unsigned int &count,unsigned int &count_edge, unsigned int cur_proc)
      //-------------------------------------------------------------------------------
    {
      //std::cout<<"index="<<index<<std::endl;
      if(taken[index]==0)
	{
	  count += weight_block[index];
	  count_edge += edge_block[index+1]-edge_block[index];
	  taken[index] = cur_proc+1;
	}
      for(unsigned int i=0;i<degcum_block[index+1]-degcum_block[index];i++)
	takeall(linked_blocks[degcum_block[index]+i],edge_block,weight_block,degcum_block,linked_blocks,taken,count,count_edge,cur_proc);
    }
    //-------------------------------------------------------------------------------
    //! to verify that no isolated edges are stuck
    /*!
      When the ideal number of edges is reached, verify that no isolated edges are stuck in the subtree taken by the processor
      \param index is the index of the root element of the subtree
      \param edge_b_loc 
      \param degcum_block is the cumulative degrees for links between blocks of sister edges (equivalent of degcum_out for blocks)
      \param linked_blocks is the associated table to degcum_block, it is the list of neighbor blocks
      \param taken is the table to note to which processor is assigned each block
      \param count is the reference to the count of edges for the current processor
    */
    //-------------------------------------------------------------------------------
    void verif(unsigned int *index,unsigned int *edge_b_loc,unsigned int *weight_b_loc, unsigned int size, unsigned int * degcum_block, unsigned int *linked_blocks, 
	       unsigned int *taken, unsigned int &count, unsigned int &count_edge)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<size;i++)
	{
	  if(taken[index[i]]==0 && degcum_block[index[i]+1]-degcum_block[index[i]]==1 && taken[linked_blocks[degcum_block[index[i]]]]>0)
	    {
	      count+=weight_b_loc[i];
	      count_edge+=edge_b_loc[i];
	      taken[index[i]] = taken[linked_blocks[degcum_block[index[i]]]];
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! recursive function for distribution when a root has been chosen or when a leaf has been reached (so an additionnal root is chosen)
    /*!
      \param mean_edges is the mean of the number of edges in blocks, it is used for an error on the ideal number of edges
      \param edges_r
      \param index is the index where to start the recursive distribution
      \param edge_b_loc
      \param size
      \param ideal is the ideal number of edges per processor
      \param taken is the table to note which block is assigned to which processor
      \param edges_subtree
      \param count is the number of edges taken by the current processor
      \pram cur_proc is the current processor
      \return false if the ideal number of edges is reached
    */
    //-------------------------------------------------------------------------------
    bool dist(unsigned int mean_edges, unsigned int & weights_r,unsigned int *index,unsigned int *edge_b_loc,unsigned int *weight_b_loc, unsigned int size, unsigned int ideal, 
		      unsigned int *edge_block,unsigned int *weight_block, unsigned int * degcum_block, unsigned int *linked_blocks, 
		      unsigned int *taken, unsigned int * edges_subtree,unsigned int &count,unsigned int &count_edge,int &cur_proc)
    //-------------------------------------------------------------------------------
    {
      /*
	1) try to take output blocks in edge_b_loc
	2) first search for subtrees (with roots in edge_b_loc) that can be get entirely without reaching ideal
	3) then enter the recrusive call of dist(...) itself for biger subtree
	This way isolated branches are avoided
       */

      if(size>0)
	{
	  //first take the blocks that we can entirely take (<ideal)
	  unsigned int bigger=0;
	  for(unsigned int i=0;i<size;i++)
	    {
	      if(taken[index[i]]==0)
		{
		  if(edges_subtree[index[i]]>edges_subtree[index[bigger]])
		    bigger = i;

		  if(count<ideal)
		    {
		      //if we can directly add the block and its subtree
		      //if(count+edges_subtree[index[i]]<=ideal+mean_edges+floor(ideal*0.1))
		      if(count+edges_subtree[index[i]]<=ideal/*+mean_edges*/)
			takeall(index[i],edge_block,weight_block,degcum_block,linked_blocks,taken,count,count_edge,cur_proc);
		    }
		  else if(count>=ideal)
		    {
		      verif(index,edge_b_loc,weight_b_loc,size,degcum_block,linked_blocks,taken,count,count_edge);
		      //free(index);
		      //free(edge_b_loc);
		      return false;
		    }
		}
	    }
	  //bigger case
	  if(taken[index[bigger]]==0 && count<ideal)
	    {
	      //if(count<ideal && count+edge_b_loc[bigger]<=ideal+mean_edges+floor(ideal*0.1))
	      if(count<ideal && count+weight_b_loc[bigger]<=ideal/*+mean_edges*/)
		{
		  count+=weight_b_loc[bigger];
		  count_edge+=edge_b_loc[bigger];
		  taken[index[bigger]] = cur_proc+1;
		  //recursive call
		  if(count<ideal)
		    {
		      unsigned int new_size=degcum_block[index[bigger]+1]-degcum_block[index[bigger]];
		      unsigned int *edge_block_loc = (unsigned int *)calloc(new_size,sizeof(unsigned int));
		      unsigned int *weight_block_loc = (unsigned int *)calloc(new_size,sizeof(unsigned int));
		      unsigned int *index_loc = (unsigned int *)calloc(new_size,sizeof(unsigned int));
		      if(new_size>0)
			{
			  for(unsigned int j=0;j<degcum_block[index[bigger]+1]-degcum_block[index[bigger]];j++)
			    {
			      index_loc[j] = linked_blocks[degcum_block[index[bigger]]+j];
			      edge_block_loc[j] = edge_block[index_loc[j]+1]-edge_block[index_loc[j]];
			      weight_block_loc[j] = weight_block[index_loc[j]];
			    }
			}

		      //appel récursif
		      bool ret = dist(mean_edges,weights_r,index_loc,edge_block_loc,weight_block_loc,new_size,ideal,
				      edge_block,weight_block,degcum_block,linked_blocks,taken,edges_subtree,count,count_edge,cur_proc);
		      verif(index,edge_b_loc,weight_b_loc,size,degcum_block,linked_blocks,taken,count,count_edge);
		      
		      free(index_loc);
		      free(edge_block_loc);
		      free(weight_block_loc);
		      if(ret==false)
			return ret;
		    }
		  else if(count>=ideal)
		    {
		      verif(index,edge_b_loc,weight_b_loc,size,degcum_block,linked_blocks,taken,count,count_edge);
		      //free(index);
		      //free(edge_b_loc);
		      return false;
		    }
		}
	      else if(count>=ideal)
		{
		  verif(index,edge_b_loc,weight_b_loc,size,degcum_block,linked_blocks,taken,count,count_edge);
		  //free(index);
		  //free(edge_b_loc);
		  return false;
		}
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Read the input DAG file .dot
    /*!
      \param file is the path to the file to read
      \param dim_node_g is the number of nodes in the global DDAG_impl
      \param dim_edge_g is the number of edges in the global DDAG_impl
      \param src_nodes_g is the list of source nodes for each edge in the global DDAG_impl
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
    */
    //-------------------------------------------------------------------------------
    void read(const char * file,unsigned int *&src_nodes_g,unsigned int *&dst_nodes_g,unsigned int *&weight_g, unsigned int &weights)
    //-------------------------------------------------------------------------------
    {
      std::ifstream dotfile;
      dotfile.open(file);
      
      std::string line;

      while(!dotfile.eof()) 
      {
	src_nodes_g=(unsigned int *)realloc(src_nodes_g,(dim_edge_g+1)*sizeof(unsigned int));
	src_nodes_g[dim_edge_g]=0;
	dst_nodes_g=(unsigned int *)realloc(dst_nodes_g,(dim_edge_g+1)*sizeof(unsigned int));
	dst_nodes_g[dim_edge_g]=0;
	weight_g=(unsigned int *)realloc(weight_g,(dim_edge_g+1)*sizeof(unsigned int));
	weight_g[dim_edge_g]=0;
	std::getline(dotfile,line);
	sscanf(line.data(),"%u -> %u : %u ;", &src_nodes_g[dim_edge_g], &dst_nodes_g[dim_edge_g],&weight_g[dim_edge_g]);

	weights+=weight_g[dim_edge_g];

	dim_node_g=max(dim_node_g,src_nodes_g[dim_edge_g],dst_nodes_g[dim_edge_g]);
	dim_edge_g++;
      }

      dotfile.close();
    }
    //-------------------------------------------------------------------------------
    //! Find roots and leafs of the global DDAG_impl
    /*!
      \param dim_node_g is the number of nodes in the global DDAG_impl
      \param dim_edge_g is the number of edges in the global DDAG_impl
      \param src_nodes_g is the list of source nodes for each edge in the global DDAG_impl
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
      \param dim_roots_g is the reference to the global number of roots elements in the global DDAG_impl
      \param dim_leafs_g is the reference to the global number of leafs elements in the global DDAG_impl
    */
    //-------------------------------------------------------------------------------
    void findRootsLeafs(unsigned int * src_nodes_g, unsigned int * dst_nodes_g, 
			unsigned int &dim_roots_g, unsigned int &dim_leafs_g, unsigned int *&roots_g, unsigned int *&leafs_g)
    //-------------------------------------------------------------------------------
    {
      for(unsigned int j=0;j<dim_node_g;j++)
	{
	  int found_src=0;
	  int found_dst=0;
	  for(unsigned int i=0;i<dim_edge_g;i++)
	    {
	      if(src_nodes_g[i]==j)
		found_src++;
	      if(dst_nodes_g[i]==j)
		found_dst++;
	    }
	  if(found_src>0 && found_dst==0)
	    {
	      dim_roots_g++;
	      roots_g = (unsigned int *)realloc(roots_g,(dim_roots_g)*sizeof(unsigned int));
	      roots_g[dim_roots_g-1]=j;
	    }
	  if(found_src==0 && found_dst>0)
	    {
	      dim_leafs_g++;
	      leafs_g = (unsigned int *)realloc(leafs_g,(dim_leafs_g)*sizeof(unsigned int));
	      leafs_g[dim_leafs_g-1]=j;
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //! Construction of the global DAG data structure
    /*!
      \param dim_edge_g is the number of edges in the global DDAG_impl
      \param dim_node_g is the number of nodes in the global DDAG_impl
      \param src_nodes_g is the list of source nodes for each edge in the global DDAG_impl
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
      \param degcum_in_g equivalent of degcum_in for the global DAG_impl
      \param degcum_out_g equivalent of degcum_out for the global DAG_impl
      \param neigh_out_g equivalent of the neigh_out for the global DAG_impl
      \param neigh_in_g equivalent of the neigh_in for the global DAG_impl
      \param edges_g equivalent of edges_out and edges_in for the global DAG_impl (no distinction needed)
      \param edge_block cumulative degrees of the number of edges per block
    */
    //-------------------------------------------------------------------------------
    void globalDS(unsigned int *src_nodes_g,unsigned int *dst_nodes_g,
		  unsigned int *&degcum_in_g,unsigned int *&degcum_out_g,unsigned int *&neigh_out_g,unsigned int *&neigh_in_g,unsigned int *&edges_g, unsigned int *&edge_block,
		  unsigned int &mean_edges, unsigned int &max_edges, unsigned int &b)
    //-------------------------------------------------------------------------------
    {
      //--------------------------------------------------------------------------------
      //construction of degcum_out_g and degcum_in_g :
      //represent the cumulative number of inputs and outputs nodes and edges of a node
      //construction of edge_block :
      //represent the number of edges in each group of sister edges
      //--------------------------------------------------------------------------------
      unsigned int *deg_in=(unsigned int *)calloc(dim_node_g,sizeof(unsigned int));
      unsigned int *deg_out=(unsigned int *)calloc(dim_node_g,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge_g;i++)
	{
	  deg_out[src_nodes_g[i]]++;
	  deg_in[dst_nodes_g[i]]++;
	}

      degcum_in_g=(unsigned int *)calloc(dim_node_g+1,sizeof(unsigned int));
      degcum_out_g=(unsigned int *)calloc(dim_node_g+1,sizeof(unsigned int));
      edge_block=(unsigned int *)realloc(edge_block,1*sizeof(unsigned int));
      edge_block[0] = degcum_out_g[0];
      for (int i=1;i<dim_node_g+1;i++)
	{
	  degcum_out_g[i]=degcum_out_g[i-1]+deg_out[i-1];
	  degcum_in_g[i]=degcum_in_g[i-1]+deg_in[i-1];
	  if(i<dim_node_g && degcum_out_g[i]>degcum_out_g[i-1])
	    {
	      b++;
	      edge_block=(unsigned int *)realloc(edge_block,(b+1)*sizeof(unsigned int));
	      edge_block[b] = degcum_out_g[i];
	      /*if(edge_block[b]-edge_block[b-1]>max_edges)
		max_edges = edge_block[b]-edge_block[b-1];

		mean_edges += edge_block[b]-edge_block[b-1];*/
	    }
	  
	}
      free(deg_in),free(deg_out);
      //mean_edges = static_cast<unsigned int>(floor(mean_edges/b));
      //--------------------------------------------------------------------------------
      //construction of neigh_out_g, neigh_in_g and edges
      //represent the associated neighbour nodes and edges of a node with use of degcum tables
      //--------------------------------------------------------------------------------
      unsigned int *t_in=(unsigned int *)calloc(dim_node_g,sizeof(unsigned int));
      unsigned int *t_out=(unsigned int *)calloc(dim_node_g,sizeof(unsigned int));
      unsigned int *t=(unsigned int *)calloc(dim_node_g,sizeof(unsigned int));

      neigh_in_g=(unsigned int *)calloc(dim_edge_g,sizeof(unsigned int));
      neigh_out_g=(unsigned int *)calloc(dim_edge_g,sizeof(unsigned int));
      edges_g=(unsigned int *)calloc(dim_edge_g,sizeof(unsigned int));

      for (int i=0;i<dim_edge_g;i++) 
	{
	  neigh_out_g[ degcum_out_g[src_nodes_g[i]] + t_out[ src_nodes_g[i] ]++ ]=dst_nodes_g[i];
	  neigh_in_g[ degcum_in_g[dst_nodes_g[i]] + t_in[ dst_nodes_g[i] ]++ ]=src_nodes_g[i];
	  edges_g[ degcum_out_g[src_nodes_g[i]] + t[ src_nodes_g[i] ]++ ]=i;
	}
      free(t_out),free(t_in),free(t);
    }
    //-------------------------------------------------------------------------------
    //! Prepare data for distribution
    /*!
      \param b is the number of blocks in the global DAG
      \param edges_g equivalent of edges_out and edges_in for the global DAG (no distinction)
      \param edge_block cumulative degrees of the number of edges per block
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
      \param degcum_out_g equivalent of degcum_out for the global DAG_impl
      \param degcum_block is the cumulative degrees for links between blocks of sister edges (equivalent of degcum_out for blocks)
      \param linked_blocks is the associated table to degcum_block, it is the list of neighbor blocks
      \param edges_subtree
    */
    //-------------------------------------------------------------------------------
    void dataDistr(unsigned int b,unsigned int *edges_g,unsigned int *edge_block,unsigned int *weight_block,unsigned int *dst_nodes_g,unsigned int *degcum_out_g,
		   unsigned int *&degcum_block,unsigned int *&linked_blocks,unsigned int *&edges_subtree)
    //-------------------------------------------------------------------------------
    {
      //--------------------------------------------------------------------------------
      //construction of degcum_block and linked_block
      //represent the equivalent of degcum_out and neigh_out of nodes, but for blocks of sister edges
      //--------------------------------------------------------------------------------
      for(unsigned int i=0;i<b;i++)
	{
	  //initialization of degcum_block i+1 with the previous value
	  degcum_block[i+1] = degcum_block[i];

	  //with edge_block and edges we know which edges are in each block
	  unsigned int end = edge_block[i+1] - edge_block[i];
	  unsigned int * cur_edges = (unsigned int *)calloc(end,sizeof(unsigned int));
	  for(unsigned int j=0;j<end;j++)
	    cur_edges[j] = edges_g[edge_block[i]+j];

	  for(unsigned int j=0;j<end;j++)
	    {
	      //with dst_nodes we know the destination node the edges is connected to
	      unsigned int node_dst = dst_nodes_g[cur_edges[j]];
	      //with degcum_out and edges we know the out edges of the destination node
	      unsigned int en = degcum_out_g[node_dst+1] - degcum_out_g[node_dst];
	      unsigned int *out_ed = (unsigned int *)calloc(en,sizeof(unsigned int));
	      for(unsigned int k=0;k<en;k++)
		out_ed[k] = edges_g[degcum_out_g[node_dst]+k];

	      //now we identify which edges block has these output edges
	      if(en>0)
		{
		  for(unsigned int l=0;l<b;l++)
		    {
		      if(en >0 && edge_block[l+1]==out_ed[0])
			{
			  degcum_block[i+1]++;
			  linked_blocks = (unsigned int *)realloc(linked_blocks,degcum_block[i+1]*sizeof(unsigned int));
			  linked_blocks[degcum_block[i+1]-1]=l+1;
			}
		    }
		}
	      free(out_ed);
	    }
	  free(cur_edges);
	}
      //--------------------------------------------------------------------------------
      //construction of the last table edges_subtree
      //represent the total number of edges under a block if this block is the root of the subtree (we count its own edges too)
      //--------------------------------------------------------------------------------
      for(unsigned int i=0;i<b;i++)
	{
	  //naive calculation of edges_subtree
	  edges_subtree[i] = all_edges_under(i,degcum_block,linked_blocks,weight_block);
	  //error correction
	  unsigned int * visited = (unsigned int *)calloc(b,sizeof(unsigned int));
	  is_visited(i,visited,degcum_block,linked_blocks);
	  for(unsigned int j=0;j<b;j++)
	    {
	      if(visited[j]>1)
		edges_subtree[i] -= (visited[j]-1)*(weight_block[j]);
	    }
	  free(visited);
	}
    }
    //-------------------------------------------------------------------------------
    //! Fill src_nodes and dst_nodes and construct temporary pure local indexes
    /*!
      \param b is the number of blocks in the global DAG
      \param taken is the table to note which block is assigned to which processor
      \param edge_block is the cumulative degrees for the number of edges in each block of sister edges
      \param src_nodes_g is the list of source nodes for each edge in the global DDAG_impl
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
      \param edges_g equivalent of edges_out and edges_in for the global DAG (no distinction)
      \param dim_roots_g is the number of roots in the global DAG_impl
      \param dim_leafs_g is the number of leafs in the global DAG_impl
      \param roots_g is the list of roots in the global DAG_impl
      \param leafs_g is the list of leafs in the global DAG_impl
    */
    //-------------------------------------------------------------------------------
    void localSrcDstNodes(unsigned int b,unsigned int *taken,unsigned int *edge_block,unsigned int *src_nodes_g,unsigned int *dst_nodes_g,unsigned int *edges_g,
			  unsigned int dim_roots_g,unsigned int dim_leafs_g,unsigned int *roots_g,unsigned int *leafs_g)
    //-------------------------------------------------------------------------------
    {
      /*
	src_nodes and dst_nodes are taken from globals src_nodes_g and dst_nodes_g and the table taken
	indexes of edges are also taken from this
	BUT indexation are modified three time during the construction of the data structure
	1) at the beginning (in this method) to take into acount local elements that are global roots and leafs 
	(these elements do not need communications and in addition to this they have to be managed differently than other elements)
	2) later communications are calculated and indexation of new elements received from other processors are calculated
	3) finally in the same code different indexation are made between local elements that do not need any communication to be computed and elements that need communications
	Indexation for nodes is finally made this way :
	******************************************************************************************************************************************
	||  pure local elements  |  elements with communications  |  leafs  |  roots  |  comunicated elements from other processors IN then OUT ||
	******************************************************************************************************************************************

	Indexation for edges depends on indexation of nodes and are modify too
	output edges of Node 0 are indexed from 0 to degcum_out[0]
	output edges of Node k are indexed from degcum_out[k-1] to degcum_out[k]
       */

      //get src_nodes, dst_nodes and basic indexes_edges from src_nodes_g dst_nodes_g and taken
      unsigned int ind=0;
      for(unsigned int i=0;i<b;i++)
	{
	  if(taken[i]==Mpi_::mpi_rank+1)
	    {
	      for(unsigned int j=0;j<edge_block[i+1]-edge_block[i];j++)
		{
		  src_nodes[ind+j] = src_nodes_g[edges_g[edge_block[i]+j]];
		  dst_nodes[ind+j] = dst_nodes_g[edges_g[edge_block[i]+j]];
		  indexes_edges[ind+j] = edges_g[edge_block[i]+j];
		}
	      ind += edge_block[i+1]-edge_block[i];
	    }
	}
      
      //sort the node indexes (src and dst) without double values
      //it represents the new basic indexes
      unsigned int * temp = (unsigned int*)calloc(2*dim_edge,sizeof(unsigned int));
      indexes_nodes = (unsigned int*)calloc(1,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	temp[i] = src_nodes[i];
      for(unsigned int i=dim_edge;i<2*dim_edge;i++)
	temp[i] = dst_nodes[i-dim_edge];
      //sort temp
      bool mod=true;
      while(mod)
	{
	  mod=false;
	  for(unsigned int i=1;i<2*dim_edge;i++)
	    {
	      if(temp[i]<temp[i-1])
		{
		  unsigned int t = temp[i];
		  temp[i] = temp[i-1];
		  temp[i-1] = t;
		  mod=true;
		}
	    }
	}
      //fill sort
      unsigned int * s_indexes_nodes = (unsigned int*)calloc(1,sizeof(unsigned int));
      s_indexes_nodes[0] = temp[0];
      dim_node=1;
      for(unsigned int i=1;i<2*dim_edge;i++)
	{
	  //erase double values
	  if(temp[i]!=temp[i-1])
	    {
	      dim_node++;
	      indexes_nodes = (unsigned int*)realloc(indexes_nodes,dim_node*sizeof(unsigned int));
	      s_indexes_nodes = (unsigned int*)realloc(s_indexes_nodes,dim_node*sizeof(unsigned int));
	      s_indexes_nodes[dim_node-1]=temp[i];
	      indexes_nodes[dim_node-1]=temp[i];
	    }
	}
      free(temp);
      
      //change those new indexes, taking care of roots and leafs elements
      unsigned int rev = dim_node-1;
      unsigned int in=0;
      for(unsigned int i=0;i<dim_node;i++)
	{
	  bool l=false;
	  for(unsigned int j=0;j<dim_leafs_g;j++)
	    {
	      if(leafs_g[j]==s_indexes_nodes[i])
		{
		  dim_leafs++;
		  indexes_nodes[rev]=s_indexes_nodes[i];
		  rev--;
		  l=true;
		}
	    }
	  bool r = false;
	  for(unsigned int j=0;j<dim_roots_g;j++)
	    {
	      if(roots_g[j]==s_indexes_nodes[i])
		{
		  dim_roots++;
		  indexes_nodes[rev]=s_indexes_nodes[i];
		  rev--;
		  r=true;
		}
	    }
	  if(!r && !l)
	    {
	      indexes_nodes[in] = s_indexes_nodes[i];
	      in++;
	    }
	}
      free(s_indexes_nodes);

      //modification of src_nodes and dst_nodes with local indexes
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  bool f_src=false;
	  bool f_dst=false;
	  for(unsigned int j=0;j<dim_node;j++)
	    {
	      if(!f_src && src_nodes[i]==indexes_nodes[j])
		{
		  src_nodes[i] = j;
		  f_src=true;
		}
	      if(!f_dst && dst_nodes[i]==indexes_nodes[j])
		{
		  dst_nodes[i] = j;
		  f_dst=true;
		}
	    }
	}
      //re-order src_nodes and dst_nodes to get growing src_nodes
      //usefull to get the local data structure with good output edges values
      unsigned int * old_src_nodes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      unsigned int * old_dst_nodes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  old_src_nodes[i]=src_nodes[i];
	  old_dst_nodes[i]=dst_nodes[i];
	}
      bool modif=true;
      while(modif)
	{
	  modif=false;
	  for(unsigned int i=1;i<dim_edge;i++)
	    {
	      if(src_nodes[i]<src_nodes[i-1])
		{
		  unsigned int s = src_nodes[i];
		  unsigned int d = dst_nodes[i];
		  src_nodes[i] = src_nodes[i-1];
		  src_nodes[i-1] = s;
		  dst_nodes[i] = dst_nodes[i-1];
		  dst_nodes[i-1] = d;
		  modif=true;
		}
	    }
	}
      //change indexes_edges with this new order of src_nodes and dst_nodes
      unsigned int * new_indexes_edges = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  for(unsigned int j=0;j<dim_edge;j++)
	    {
	      if(src_nodes[i]==old_src_nodes[j] && dst_nodes[i]==old_dst_nodes[j])
		{
		  new_indexes_edges[i] = indexes_edges[j];
		}
	    }
	}

      free(old_src_nodes);free(old_dst_nodes);
      free(indexes_edges);
      indexes_edges=new_indexes_edges;
    }
    //-------------------------------------------------------------------------------
    //! Fill the local data structure DDAG_impl
    /*!
      \param b is the number of blocks in the global DAG_impl
      \param dim_edge_g number of edges in the global DAG_impl
      \param dst_nodes_g is the list of destination nodes for each edge in the global DDAG_impl
      \param src_nodes_g is the list of source nodes for each edge in the global DDAG_impl
      \param edge_block is the cumulative degrees for the number of edges in each block of sister edges
      \param taken is the table to note which block is assigned to which processor
      \param degcum_out_t is a temporary equivalent of degcum_out to keep the pure local version
      \param degcum_in_t is a temporary equivalent of degcum_in to keep the pure local version
      \param edges_t_out is a temporary equivalent of edges_out to keep the pure local version
      \param edges_t_in is a temporary equivalent of edges_in to keep the pure local version
      \param neigh_out_t is a temporary equivalent of neigh_out to keep the pure local version
      \param neigh_in_t is a temporary equivalent of neigh_in to keep the pure local version
      \param roots_g is the list of roots in the global DAG_impl
      \param leafs_g is the list of leafs in the global DAG_impl
      \param dim_roots_g is the number of roots in the global DAG_impl
      \param dim_leafs_g is the number of leafs in the global DAG_impl
    */
    //-------------------------------------------------------------------------------
    void fillLocalDS(unsigned int b, unsigned int *dst_nodes_g,unsigned int *src_nodes_g,unsigned int *edge_block,unsigned int *taken,
		     unsigned int *degcum_out_t,unsigned int *degcum_in_t,unsigned int *edges_t_out,unsigned int *edges_t_in,unsigned int *neigh_out_t,unsigned int *neigh_in_t,
		     unsigned int *roots_g,unsigned int *leafs_g,unsigned int dim_roots_g,unsigned int dim_leafs_g)
    //-------------------------------------------------------------------------------
    {
      /*
	for each node of the local data structure an analysis of communications needed is made

	1) search in all the global edges that are not taken by the current processor if the node is found. In this case it represents an edge and a node to receive from another processor.
             a) fill degcum_tor edges_tor and nodes_tor with in and out communications to receive
	     b) remember those communications in tor_out and tor_in and stor_out stor_in to modify local data structures
	     c) remember the processors connected to complete the point 2)
	2) search in the local edges if the node is found. In this case it represents an edge and a node to send to another processor.
	     a) fill degcum_tos edges_tos and nodes_tos helping with 1-c
	     b) distinguish communication nodes from pure local nodes
        3) get the good order of edges for input edges
	4) modify local data structure degcum_out,degcum_in,neigh_out,neigh_in,edges_out and edges_in with communications
	5) change indexation taking into acount communication and pure local nodes of 2-b

	After that the new indexation can be applied to all local data structure
       */

      //indexes_nodes and indexes_edges has been filled for local elements
      //from dim_edge and dim_node to get new indexes for communicated elements
      dim_ind_ed=dim_edge;
      dim_ind_no=dim_node;
      //to note the nodes with comunication in order to change indexation after
      unsigned int * comm_nodes = (unsigned int *)calloc(dim_comm,sizeof(unsigned int));
      unsigned int * loc_nodes = (unsigned int *)calloc(dim_loc,sizeof(unsigned int));
      
      //foreach node of the local structure
      for(unsigned int i=0;i<dim_node;i++)
	{
	  //tables to receive out and in (edges)
	  unsigned int stor_out=0;
	  unsigned int * tor_out = (unsigned int *)calloc(stor_out,sizeof(unsigned int));
	  unsigned int stor_in=0;
	  unsigned int * tor_in = (unsigned int *)calloc(stor_in,sizeof(unsigned int));
	  //tables to send out and in (edges)
	  unsigned int stos_out=0;
	  unsigned int stos_in=0;
	  unsigned int nb_proc=0;
	  //unsigned int nb_proc_out=0;
	  unsigned int * procs = (unsigned int *)calloc(nb_proc,sizeof(unsigned int));
	  //unsigned int * procs_out = (unsigned int *)calloc(nb_proc_out,sizeof(unsigned int));
	  //look at all edges in the global structure
	  for(unsigned int j=0;j<dim_edge_g;j++)
	    {
	      unsigned int block = 0;
	      //if the node is found in an edges
	      if(indexes_nodes[i]==dst_nodes_g[j] || indexes_nodes[i]==src_nodes_g[j])
		{
		  //which block is associated to this edge
		  for(unsigned int k=1;k<b+1;k++)
		    {
		      if(j>=edge_block[k-1] && j<edge_block[k])
			{
			  block=k-1;
			  k=b+1;
			}
		    }
		  //if the edge is not one of the current node then communications modifications are needed
		  //to receive informations
		  if(taken[block]-1!=Mpi_::mpi_rank)
		    {
		      //if the proc has not been noticed before in procs, add it
		      unsigned int t;
		      if(!find_position(procs,nb_proc,taken[block]-1,t))
			{
			  nb_proc++;
			  procs = (unsigned int *)realloc(procs,nb_proc*sizeof(unsigned int));
			  procs[nb_proc-1]=taken[block]-1;
			}

		      //receive an element that goes out of this node
		      if(indexes_nodes[i]==src_nodes_g[j])
			{
			  //to receive
			  stor_out++;
			  tor_out = (unsigned int *)realloc(tor_out,stor_out*sizeof(unsigned int));
			  tor_out[stor_out-1] = j;
			  
			  //construction of degcum_tor,degcum_tos (simple modification of values)
			  for(unsigned int k=taken[block];k<Mpi_::mpi_nb+1;k++)
			    {
			      degcum_tor[k]++;
			      degcum_tor_out[k]++;
			    }
			  //construction of edges_tor,edges_tos,nodes_tor,nodes_tos (reallocation of new tables)
			  //unsigned int *edges_tor_new = (unsigned int *)calloc(degcum_tor[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *edges_tor_out_new = (unsigned int *)calloc(degcum_tor_out[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *nodes_tor_new = (unsigned int *)calloc(degcum_tor[Mpi_::mpi_nb],sizeof(unsigned int));
			  //values
			  //for processors before this one simply make a copy of values
			  for(unsigned int k=0;k<taken[block]-1;k++)
			    {
			      for(unsigned int l=0;l<degcum_tor[k+1]-degcum_tor[k];l++)
				nodes_tor_new[degcum_tor[k]+l] = nodes_tor[degcum_tor[k]+l];
			      for(unsigned int l=0;l<degcum_tor_out[k+1]-degcum_tor_out[k];l++)
				edges_tor_out_new[degcum_tor_out[k]+l] = edges_tor_out[degcum_tor_out[k]+l];
			    }
			  //for this processor add the new value
			  for(unsigned int l=0;l<degcum_tor[taken[block]]-degcum_tor[taken[block]-1]-1;l++)
			    nodes_tor_new[degcum_tor[taken[block]-1]+l] = nodes_tor[degcum_tor[taken[block]-1]+l];
			  for(unsigned int l=0;l<degcum_tor_out[taken[block]]-degcum_tor_out[taken[block]-1]-1;l++)
			      edges_tor_out_new[degcum_tor_out[taken[block]-1]+l] = edges_tor_out[degcum_tor_out[taken[block]-1]+l];
			  //add the new value
			  unsigned int last_tor = degcum_tor[taken[block]-1]+degcum_tor[taken[block]]-degcum_tor[taken[block]-1]-1;
			  nodes_tor_new[last_tor] = dst_nodes_g[j];
			  last_tor = degcum_tor_out[taken[block]-1]+degcum_tor_out[taken[block]]-degcum_tor_out[taken[block]-1]-1;
			  edges_tor_out_new[last_tor] = j;
			  //for the next processors copy values at a position +1
			  for(unsigned int k=taken[block];k<Mpi_::mpi_nb;k++)
			    {
			      for(unsigned int l=0;l<degcum_tor[k+1]-degcum_tor[k];l++)
				{
				  //edges_tor_new[degcum_tor[k]+l] = edges_tor[degcum_tor[k]+l-1];
				  nodes_tor_new[degcum_tor[k]+l] = nodes_tor[degcum_tor[k]+l-1];
				}
			      for(unsigned int l=0;l<degcum_tor_out[k+1]-degcum_tor_out[k];l++)
				edges_tor_out_new[degcum_tor_out[k]+l] = edges_tor_out[degcum_tor_out[k]+l-1];
			    }

			  /*free(edges_tor);*/free(nodes_tor);
			  /*edges_tor = edges_tor_new;*/ nodes_tor = nodes_tor_new;
			  free(edges_tor_out);
			  edges_tor_out = edges_tor_out_new;
			}
		      //receive an element that goes in of this node
		      if(indexes_nodes[i]==dst_nodes_g[j])
			{
			  //to receive
			  stor_in++;
			  tor_in = (unsigned int *)realloc(tor_in,stor_in*sizeof(unsigned int));
			  tor_in[stor_in-1] = j;

			  //construction of degcum_tor,degcum_tos (simple modification of values)
			  for(unsigned int k=taken[block];k<Mpi_::mpi_nb+1;k++)
			    {
			      degcum_tor[k]++;
			      degcum_tor_in[k]++;
			    }
			  //construction of edges_tor,edges_tos,nodes_tor,nodes_tos (reallocation of new tables)
			  //unsigned int *edges_tor_new = (unsigned int *)calloc(degcum_tor[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *edges_tor_in_new = (unsigned int *)calloc(degcum_tor_in[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *nodes_tor_new = (unsigned int *)calloc(degcum_tor[Mpi_::mpi_nb],sizeof(unsigned int));
			  //values
			  //for processors before this one simply make a copy of values
			  for(unsigned int k=0;k<taken[block]-1;k++)
			    {
			      for(unsigned int l=0;l<degcum_tor[k+1]-degcum_tor[k];l++)
				{
				  //edges_tor_new[degcum_tor[k]+l] = edges_tor[degcum_tor[k]+l];
				  nodes_tor_new[degcum_tor[k]+l] = nodes_tor[degcum_tor[k]+l];
				}
			      for(unsigned int l=0;l<degcum_tor_in[k+1]-degcum_tor_in[k];l++)
				edges_tor_in_new[degcum_tor_in[k]+l] = edges_tor_in[degcum_tor_in[k]+l];
			    }
			  //for this processor add the new value
			  for(unsigned int l=0;l<degcum_tor[taken[block]]-degcum_tor[taken[block]-1]-1;l++)
			    {
			      //edges_tor_new[degcum_tor[taken[block]-1]+l] = edges_tor[degcum_tor[taken[block]-1]+l];
			      nodes_tor_new[degcum_tor[taken[block]-1]+l] = nodes_tor[degcum_tor[taken[block]-1]+l];
			    }
			  for(unsigned int l=0;l<degcum_tor_in[taken[block]]-degcum_tor_in[taken[block]-1]-1;l++)
			    edges_tor_in_new[degcum_tor_in[taken[block]-1]+l] = edges_tor_in[degcum_tor_in[taken[block]-1]+l];
			  //add the new value
			  unsigned int last_tor = degcum_tor[taken[block]-1]+degcum_tor[taken[block]]-degcum_tor[taken[block]-1]-1;
			  nodes_tor_new[last_tor] = src_nodes_g[j];
			  last_tor = degcum_tor_in[taken[block]-1]+degcum_tor_in[taken[block]]-degcum_tor_in[taken[block]-1]-1;
			  edges_tor_in_new[last_tor] = j;
			  //for the next processors copy values at a position +1
			  for(unsigned int k=taken[block];k<Mpi_::mpi_nb;k++)
			    {
			      for(unsigned int l=0;l<degcum_tor[k+1]-degcum_tor[k];l++)
				{
				  //edges_tor_new[degcum_tor[k]+l] = edges_tor[degcum_tor[k]+l-1];
				  nodes_tor_new[degcum_tor[k]+l] = nodes_tor[degcum_tor[k]+l-1];
				}
			      for(unsigned int l=0;l<degcum_tor_in[k+1]-degcum_tor_in[k];l++)
				edges_tor_in_new[degcum_tor_in[k]+l] = edges_tor_in[degcum_tor_in[k]+l-1];
			    }

			  /*free(edges_tor);*/free(nodes_tor);
			  /*edges_tor = edges_tor_new;*/ nodes_tor = nodes_tor_new;
			  free(edges_tor_in);
			  edges_tor_in = edges_tor_in_new;
			}
		    }
		}
	    }
	  //sending informations
	  if(nb_proc>0)
	    {
	      bool ro = find_position(roots_g,dim_roots_g,indexes_nodes[i]);
	      bool le = find_position(leafs_g,dim_leafs_g,indexes_nodes[i]);
	      if(!ro && !le)
		{
		  //communication node
		  dim_comm++;
		  comm_nodes = (unsigned int *)realloc(comm_nodes,dim_comm*sizeof(unsigned int));
		  //current local indexation noted
		  comm_nodes[dim_comm-1] = i;
		}
	      
	      //sort procs
	      bool mod=true;
	      while(mod)
		{
		  mod=false;
		  for(unsigned int j=1;j<nb_proc;j++)
		    {
		      if(procs[j]<procs[j-1])
			{
			  unsigned int t = procs[j];
			  procs[j] = procs[j-1];
			  procs[j-1] = t;
			  mod=true;
			}
		    }
		}
	      for(unsigned int j=0;j<dim_edge;j++)
		{
		  //send an element that is out of another processor
		  if(i==dst_nodes[j] || i==src_nodes[j])
		    {
		      //Needed for communications in advanced DPMap with pointers
		      if(i==dst_nodes[j])
			stos_in+=nb_proc;
		      else if(i==src_nodes[j])
			stos_out+=nb_proc;

		      //construction of degcum_tor,degcum_tos (simple modification of values)
		      for(unsigned int p=0;p<nb_proc;p++)
			{
			  for(unsigned int k=procs[p]+1;k<Mpi_::mpi_nb+1;k++)
			    {
			      degcum_tos[k]++;
			      if(i==dst_nodes[j])
				degcum_tos_in[k]++;
			      else if(i==src_nodes[j])
				degcum_tos_out[k]++;
			    }
		    
			  //construction of edges_tor,edges_tos,nodes_tor,nodes_tos (reallocation of new tables)
			  //unsigned int *edges_tos_new = (unsigned int *)calloc(degcum_tos[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *edges_tos_in_new; unsigned int *edges_tos_out_new;
			  if(i==src_nodes[j])
			    edges_tos_out_new = (unsigned int *)calloc(degcum_tos_out[Mpi_::mpi_nb],sizeof(unsigned int));
			  else if(i==dst_nodes[j])
			    edges_tos_in_new = (unsigned int *)calloc(degcum_tos_in[Mpi_::mpi_nb],sizeof(unsigned int));
			  unsigned int *nodes_tos_new = (unsigned int *)calloc(degcum_tos[Mpi_::mpi_nb],sizeof(unsigned int));
			  //values
			  //for processors before this one simply make a copy of values
			  for(unsigned int k=0;k<procs[p];k++)
			    {
			      for(unsigned int l=0;l<degcum_tos[k+1]-degcum_tos[k];l++)
				nodes_tos_new[degcum_tos[k]+l] = nodes_tos[degcum_tos[k]+l];
			      if(i==src_nodes[j])
				{
				  for(unsigned int l=0;l<degcum_tos_out[k+1]-degcum_tos_out[k];l++)
				    edges_tos_out_new[degcum_tos_out[k]+l] = edges_tos_out[degcum_tos_out[k]+l];
				}
			      else if(i==dst_nodes[j])
				{
				  for(unsigned int l=0;l<degcum_tos_in[k+1]-degcum_tos_in[k];l++)
				    edges_tos_in_new[degcum_tos_in[k]+l] = edges_tos_in[degcum_tos_in[k]+l];
				}
			    }
			  //for this processor add the precedent values and the new value
			  for(unsigned int l=0;l<degcum_tos[procs[p]+1]-degcum_tos[procs[p]]-1;l++)
			    nodes_tos_new[degcum_tos[procs[p]]+l] = nodes_tos[degcum_tos[procs[p]]+l];
			  if(i==src_nodes[j])
			    {
			      for(unsigned int l=0;l<degcum_tos_out[procs[p]+1]-degcum_tos_out[procs[p]]-1;l++)
				edges_tos_out_new[degcum_tos_out[procs[p]]+l] = edges_tos_out[degcum_tos_out[procs[p]]+l];
			    }
			  else if(i==dst_nodes[j])
			    {
			      for(unsigned int l=0;l<degcum_tos_in[procs[p]+1]-degcum_tos_in[procs[p]]-1;l++)
				edges_tos_in_new[degcum_tos_in[procs[p]]+l] = edges_tos_in[degcum_tos_in[procs[p]]+l];
			    }
			  //add the new value
			  unsigned int last_tos = degcum_tos[procs[p]+1]-1;
			  if(i==dst_nodes[j])
			    nodes_tos_new[last_tos] = src_nodes[j];
			  else if(i==src_nodes[j])
			    nodes_tos_new[last_tos] = dst_nodes[j];
			  //edges_tos_new[last_tos] = j;
			  if(i==src_nodes[j])
			    {
			      last_tos = degcum_tos_out[procs[p]+1]-1;
			      edges_tos_out_new[last_tos] = j;
			    }
			  else if(i==dst_nodes[j])
			    {
			      last_tos = degcum_tos_in[procs[p]+1]-1;
			      edges_tos_in_new[last_tos] = j;
			    }

			  //for the next processors copy values at a position +1
			  for(unsigned int k=procs[p]+1;k<Mpi_::mpi_nb;k++)
			    {
			      for(unsigned int l=0;l<degcum_tos[k+1]-degcum_tos[k];l++)
				{
				  //edges_tos_new[degcum_tos[k]+l] = edges_tos[degcum_tos[k]+l-1];
				  nodes_tos_new[degcum_tos[k]+l] = nodes_tos[degcum_tos[k]+l-1];
				}
			      if(i==src_nodes[j])
				{
				  for(unsigned int l=0;l<degcum_tos_out[k+1]-degcum_tos_out[k];l++)
				    edges_tos_out_new[degcum_tos_out[k]+l] = edges_tos_out[degcum_tos_out[k]+l-1];
				}
			      else if(i==dst_nodes[j])
				{
				  for(unsigned int l=0;l<degcum_tos_in[k+1]-degcum_tos_in[k];l++)
				    edges_tos_in_new[degcum_tos_in[k]+l] = edges_tos_in[degcum_tos_in[k]+l-1];
				}
			    }
			  
			  /*free(edges_tos);*/free(nodes_tos);
			  /*edges_tos = edges_tos_new;*/ nodes_tos = nodes_tos_new;
			  if(i==src_nodes[j])
			    {
			      free(edges_tos_out);
			      edges_tos_out = edges_tos_out_new;
			    }
			  else if(i==dst_nodes[j])
			    {
			      free(edges_tos_in);
			      edges_tos_in = edges_tos_in_new;
			    }
			}
		    }
		}
	    }
	  else
	    {
	      bool ro = find_position(roots_g,dim_roots_g,indexes_nodes[i]);
	      bool le = find_position(leafs_g,dim_leafs_g,indexes_nodes[i]);
	      if(!ro && !le)
		{
		  dim_loc++;
		  loc_nodes = (unsigned int *)realloc(loc_nodes,dim_loc*sizeof(unsigned int));
		  //current local indexation noted
		  loc_nodes[dim_loc-1] = i;
		}
	    }
	  free(procs);

	  //edges_t_in
	  //edges_t_out is from 0 to dim_edge but with new indexation edges_t_in can't be so simple and has to be changed
	  //for each node find in all local edges the node as destination
	  unsigned int pl=0;
	  for(unsigned int j=0;j<dim_edge;j++)
	    {
	      if(i==dst_nodes[j])
		{
		  edges_t_in[degcum_in_t[i]+pl]=j;
		  pl++;
		}
	    }
	  	  
	  //for this node modify degcum_out,degcum_in,neigh_out,neigh_in,edges_out and edges_in
	  for(unsigned int j=i+1;j<dim_node+1;j++)
	    {
	      degcum_out[j]+=stor_out;
	      degcum_in[j]+=stor_in;
	    }

	  dim_tor_out+=stor_out;
	  dim_tor_in+=stor_in;
	  dim_tos_out+=stos_out;
	  dim_tos_in+=stos_in;
	  
	  //for the current node, modify neigh_out/neigh_in/edges_out/edges_in and indexes_nodes/indexes_edges
	  neigh_out = (unsigned int *)realloc(neigh_out,degcum_out[i+1]*sizeof(unsigned int));
	  neigh_in = (unsigned int *)realloc(neigh_in,degcum_in[i+1]*sizeof(unsigned int));
	  edges_out = (unsigned int *)realloc(edges_out,degcum_out[i+1]*sizeof(unsigned int));
	  edges_in = (unsigned int *)realloc(edges_in,degcum_in[i+1]*sizeof(unsigned int));

	  //these varaibles are used to not hav multiple times the same value in indexes tables and to count the number of elements to add in indexes
	  unsigned int ind_ed=0;
	  unsigned int ind_no=0;
	  //add new elements first for the current node
	  //in elements to receive
	  for(unsigned int j=0;j<stor_in;j++)
	    {
	      //change the size of indexes tables
	      indexes_nodes = (unsigned int *)realloc(indexes_nodes,(dim_ind_no+ind_no+1)*sizeof(unsigned int));
	      indexes_edges = (unsigned int *)realloc(indexes_edges,(dim_ind_ed+ind_ed+1)*sizeof(unsigned int));

	      //for edges
	      //if the index is not already stored add it to indexes when modifying neigh and edges
	      unsigned int position;
	      bool found = find_position(indexes_edges,dim_ind_ed+ind_ed,tor_in[j],position);
	      if(!found)
		{
		  //edges_in[degcum_in[i]+j]=dim_ind_ed+ind_ed;
		  edges_in[degcum_in[i]+j]=dim_ind_ed+ind_ed;
		  indexes_edges[dim_ind_ed+ind_ed]=tor_in[j];
		  ind_ed++;
		}
	      //else use the existing position value
	      else
		edges_in[degcum_in[i]+j]=position;
	      //same for nodes
	      found = find_position(indexes_nodes,dim_ind_no+ind_no,src_nodes_g[tor_in[j]],position);
	      if(!found)
		{
		  neigh_in[degcum_in[i]+j]=dim_ind_no+ind_no;
		  indexes_nodes[dim_ind_no+ind_no]=src_nodes_g[tor_in[j]];
		  ind_no++;
		}
	      else
		neigh_in[degcum_in[i]+j]=position;
	    }
	  //out elements to receive
	  dim_ind_no+=ind_no;
	  dim_ind_ed+=ind_ed;
	  ind_no=0;
	  ind_ed=0;
	  for(unsigned int j=0;j<stor_out;j++)
	    {
	      //change the size of indexes tables
	      indexes_nodes = (unsigned int *)realloc(indexes_nodes,(dim_ind_no+ind_no+1)*sizeof(unsigned int));
	      indexes_edges = (unsigned int *)realloc(indexes_edges,(dim_ind_ed+ind_ed+1)*sizeof(unsigned int));

	      //for edges
	      //if the index is not already stored add it to indexes when modifying neigh and edges
	      unsigned int position;
	      bool found = find_position(indexes_edges,dim_ind_ed+ind_ed,tor_out[j],position);
	      if(!found)
		{
		  edges_out[degcum_out[i]+j]=dim_ind_ed+ind_ed;
		  indexes_edges[dim_ind_ed+ind_ed]=tor_out[j];
		  ind_ed++;
		}
	      //else use existing position
	      else
		edges_out[degcum_out[i]+j]=position;
	      //for nodes
	      found = find_position(indexes_nodes,dim_ind_no+ind_no,dst_nodes_g[tor_out[j]],position);
	      if(!found)
		{
		  neigh_out[degcum_out[i]+j]=dim_ind_no+ind_no;
		  indexes_nodes[dim_ind_no+ind_no]=dst_nodes_g[tor_out[j]];
		  ind_no++;
		}
	      else
		neigh_out[degcum_out[i]+j]=position;
	    }
	  dim_ind_no+=ind_no;
	  dim_ind_ed+=ind_ed;
	  //then add local elements for the current node in neigh and edges
	  for(unsigned int j=0;j<degcum_out_t[i+1]-degcum_out_t[i];j++)
	    {
	      edges_out[degcum_out[i]+stor_out+j]=edges_t_out[degcum_out_t[i]+j];
	      neigh_out[degcum_out[i]+stor_out+j]=neigh_out_t[degcum_out_t[i]+j];
	    }
	  for(unsigned int j=0;j<degcum_in_t[i+1]-degcum_in_t[i];j++)
	    {
	      edges_in[degcum_in[i]+stor_in+j]=edges_t_in[degcum_in_t[i]+j];
	      neigh_in[degcum_in[i]+stor_in+j]=neigh_in_t[degcum_in_t[i]+j];
	    }

	  free(tor_out);free(tor_in);
	}

      //change indexation of nodes with communication and local nodes
      //this is very usefull for a communication/computation time overlap
      unsigned int * new_indexes_nodes = (unsigned int *)calloc(dim_ind_no,sizeof(unsigned int));
      unsigned int * old_new = (unsigned int *)calloc(dim_ind_no,sizeof(unsigned int));
      unsigned int * new_old = (unsigned int *)calloc(dim_ind_no,sizeof(unsigned int));
      //foreach old indexation value
      unsigned int count = dim_ind_no-1;
      for(unsigned int i=0;i<dim_ind_no-dim_node;i++)
	{
	  new_indexes_nodes[count]=indexes_nodes[count];
	  old_new[count]=count;
	  new_old[count]=count;
	  count--;
	}
      for(unsigned int i=0;i<dim_leafs;i++)
	{
	  new_indexes_nodes[count]=indexes_nodes[count];
	  old_new[count]=count;
	  new_old[count]=count;
	  count--;
	}
      for(unsigned int i=0;i<dim_roots;i++)
	{
	  new_indexes_nodes[count]=indexes_nodes[count];
	  old_new[count]=count;
	  new_old[count]=count;
	  count--;
	}
      unsigned int ind=dim_comm-1;
      for(unsigned int i=0;i<dim_comm;i++)
	{
	  new_indexes_nodes[count]=indexes_nodes[comm_nodes[ind]];
	  old_new[comm_nodes[ind]]=count;
	  new_old[count]=comm_nodes[ind];
	  ind--;
	  count--;
	}
      ind = dim_loc-1;
      for(unsigned int i=0;i<dim_loc;i++)
	{
	  new_indexes_nodes[count]=indexes_nodes[loc_nodes[ind]];
	  old_new[loc_nodes[ind]]=count;
	  new_old[count]=loc_nodes[ind];
	  ind--;
	  count--;
	}
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" rcv = "<<dim_ind_no-dim_node<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" dim_leafs = "<<dim_leafs<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" dim_roots = "<<dim_roots<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" dim_comm = "<<dim_comm<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" dim_loc = "<<dim_loc<<std::endl;*/

      //delete old indexation
      free(indexes_nodes);
      indexes_nodes = new_indexes_nodes;

      //change values in data structures using old_new (only nodes concerned to to send values)
      for (int i=0;i<dim_edge;i++)
	src_nodes[i]=old_new[src_nodes[i]];
      for (int i=0;i<dim_edge;i++)
	dst_nodes[i]=old_new[dst_nodes[i]];
      for (int i=0;i<dim_tos_out+dim_tos_in;i++)
	nodes_tos[i]=old_new[nodes_tos[i]];
      //for degcum_out/degcum_in/neigh_out/neigh_in/edges_out/edges_in a little more dificult as it is cumulative
      unsigned int *new_degcum_out = (unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      unsigned int *new_degcum_in = (unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      unsigned int *new_neigh_out = (unsigned int *)calloc(dim_edge+dim_tor_out,sizeof(unsigned int));
      unsigned int *new_neigh_in = (unsigned int *)calloc(dim_edge+dim_tor_in,sizeof(unsigned int));
      unsigned int *new_edges_out = (unsigned int *)calloc(dim_edge+dim_tor_out,sizeof(unsigned int));
      unsigned int *new_edges_in = (unsigned int *)calloc(dim_edge+dim_tor_in,sizeof(unsigned int));
      for(unsigned int i=1;i<dim_node+1;i++)
	{
	  new_degcum_out[i]=new_degcum_out[i-1]+(degcum_out[new_old[i-1]+1]-degcum_out[new_old[i-1]]);
	  new_degcum_in[i]=new_degcum_in[i-1]+(degcum_in[new_old[i-1]+1]-degcum_in[new_old[i-1]]);

	  unsigned int new_st = new_degcum_out[i-1];
	  unsigned int st = degcum_out[new_old[i-1]];
	  for(unsigned int j=0;j<new_degcum_out[i]-new_degcum_out[i-1];j++)
	    {
	      new_neigh_out[new_st+j]=old_new[neigh_out[st+j]];
	      new_edges_out[new_st+j]=edges_out[st+j];
	    }
	  new_st = new_degcum_in[i-1];
	  st = degcum_in[new_old[i-1]];
	  for(unsigned int j=0;j<new_degcum_in[i]-new_degcum_in[i-1];j++)
	    {
	      new_neigh_in[new_st+j]=old_new[neigh_in[st+j]];
	      new_edges_in[new_st+j]=edges_in[st+j];
	    }
	}

      //put local indexes in nodes_tor and edges_tor (global values)
      for(unsigned int i=0;i<dim_tor_out;i++)
	{
	  for(unsigned int j=0;j<dim_ind_no;j++)
	    {
	      if(indexes_nodes[j]==nodes_tor[i])
		{
		  nodes_tor[i] = j;
		  break;
		}
	    }
	  for(unsigned int j=0;j<dim_ind_ed;j++)
	    {
	      if(indexes_edges[j]==edges_tor_out[i])
		{
		  edges_tor_out[i] = j;
		  break;
		}
	    }
	}
      for(unsigned int i=0;i<dim_tor_in;i++)
	{
	  for(unsigned int j=0;j<dim_ind_no;j++)
	    {
	      if(indexes_nodes[j]==nodes_tor[dim_tor_out+i])
		{
		  nodes_tor[dim_tor_out+i] = j;
		  break;
		}
	    }
	  for(unsigned int j=0;j<dim_ind_ed;j++)
	    {
	      if(indexes_edges[j]==edges_tor_in[i])
		{
		  edges_tor_in[i] = j;
		  break;
		}
	    }
	}
      
      

      free(degcum_out);free(degcum_in);free(neigh_out);free(neigh_in);free(edges_out);free(edges_in);
      degcum_out=new_degcum_out; degcum_in = new_degcum_in;
      neigh_out = new_neigh_out; neigh_in=new_neigh_in;
      edges_out=new_edges_out; edges_in=new_edges_in;

      free(old_new);free(new_old);
      free(loc_nodes);free(comm_nodes);
    }
    //-------------------------------------------------------------------------------
    //! Distribution
    /*!
      Read the input file, construct the global data structure, make the distribution and fill the local data structure
      \param file is the path to the .dot input file
    */
    //-------------------------------------------------------------------------------
    void distribute(const char * file)
    //-------------------------------------------------------------------------------
    {
      unsigned int weights;
      unsigned int *src_nodes_g, *dst_nodes_g, *weight_g;
      unsigned int *degcum_out_g, *degcum_in_g;
      unsigned int *neigh_out_g, *neigh_in_g;
      unsigned int * edges_g;
      unsigned int b=0;//number of blocks
      unsigned int * edge_block,*weight_block, *degcum_block, *linked_blocks, *edges_subtree;
      
      //--------------------------------------------------------------------------------
      //construction of src_nodes, dst_nodes, dim_node, dim_edge from dot file
      //for each line of type "unsigned int -> unsigned int ;" it is an edge of the DAG.
      //incrementation of dim_edge and of dim_node
      //--------------------------------------------------------------------------------
      dim_edge_g=0;
      dim_node_g=0;
      src_nodes_g=(unsigned int *)calloc(1,sizeof(unsigned int));
      dst_nodes_g=(unsigned int *)calloc(1,sizeof(unsigned int));
      weight_g=(unsigned int *)calloc(1,sizeof(unsigned int));
      weights=0;
      read(file,src_nodes_g,dst_nodes_g,weight_g,weights);
      dim_node_g++;
      //--------------------------------------------------------------------------------
      //Identify roots and leafs of the DAG
      //--------------------------------------------------------------------------------
      unsigned int * roots_g = (unsigned int *)calloc(0,sizeof(unsigned int));
      unsigned int * leafs_g = (unsigned int *)calloc(0,sizeof(unsigned int));
      unsigned int dim_roots_g=0;
      unsigned int dim_leafs_g=0;
      findRootsLeafs(src_nodes_g, dst_nodes_g, dim_roots_g, dim_leafs_g, roots_g, leafs_g);
      //--------------------------------------------------------------------------------
      //Global data structure creation + edge_block
      //--------------------------------------------------------------------------------
      unsigned int mean_edges = 0;
      unsigned int max_edges = 0;
      edge_block = (unsigned int *)calloc(0,sizeof(unsigned int));
      globalDS(src_nodes_g,dst_nodes_g,degcum_in_g,degcum_out_g,neigh_out_g,neigh_in_g,edges_g,edge_block,mean_edges,max_edges,b);
      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - last element edge_block b="<<b<<std::endl;
      unsigned int bb=0;
      weight_block=(unsigned int *)calloc(b,sizeof(unsigned int));
      for (int i=1;i<dim_node_g+1;i++)
	{
	  //new block
	  if(i<dim_node_g && degcum_out_g[i]>degcum_out_g[i-1])
	    {
	      for(unsigned int j=0;j<degcum_out_g[i]-degcum_out_g[i-1];j++)
		weight_block[bb] += weight_g[edges_g[degcum_out_g[i-1]+j]];
	      
	      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - weight_block at "<<bb<<" = "<<weight_block[bb]<<std::endl;
	      if(weight_block[bb]>max_edges)
		max_edges = weight_block[bb];
	      mean_edges += weight_block[bb];
	      bb++;
	    }
	  
	}
      mean_edges = static_cast<unsigned int>(floor(mean_edges/b));
      //std::cout<<"Mean_edges="<<mean_edges<<std::endl;
      //--------------------------------------------------------------------------------
      //Prepare data for distribution : degcum_block and linked_blocks and edges_subtree
      //--------------------------------------------------------------------------------
      degcum_block = (unsigned int *)calloc(b+1,sizeof(unsigned int));
      linked_blocks = (unsigned int *)calloc(1,sizeof(unsigned int));
      edges_subtree = (unsigned int *)calloc(b,sizeof(unsigned int));
      dataDistr(b,edges_g,edge_block,weight_block,dst_nodes_g,degcum_out_g,degcum_block,linked_blocks,edges_subtree);
      //--------------------------------------------------------------------------------
      //DISTRIBUTION
      //--------------------------------------------------------------------------------
      /*
	The algorithm works this way :
	ideal is the ideal number of edges needed per each processor for a perfect distribution
	edges of the DAG can be ponderate by a weight, the ideal number take into acount this weight
	Edges are not managed independently. All sister edges are together in a block, and each is gonna be distribute amoung processors (table taken)
	All edges of a block are taken by a same processor
	1) foreach block (this is a sort of queue of blocks to deal with) - each time we reach this point the end of a subtree has been reached or we make the distribution for another processor
	2) try to get the new block
	3) if the new block is the first one, check if a brother block is not isolated or take it
	4) construct the list of outpt blocks and enter the recursive distribution call dist(...)
	5) When you get out of this recursive call it is because the ideal number has been reached for the processor or because the end of the subtree has been reached for this block
       */
      //table to know if a block is taken by a processor
      unsigned int *taken = (unsigned int *)calloc(b,sizeof(unsigned int));
      //to count the current number of edges taken
      unsigned int count=0;
      unsigned int count_edge=0;
      //the ideal number of edges to take
      unsigned int ideal = static_cast<unsigned int>(floor(weights/Mpi_::mpi_nb)) +1;
      //std::cout<<"ideal = "<<ideal<<"\n";
      //std::cout<<"weights = "<<weights<<"\n";
      //total number of edges that are not taken
      unsigned int weights_r = weights;
      //current processor
      int cur_proc=0;

      for(unsigned int i=0;i<b;i++)
	{
	  if(count<ideal)
	    {
	      if(count==0 && cur_proc==Mpi_::mpi_nb-1)//take all the rest
		{
		  for(unsigned int k=0;k<b;k++)
		    {
		      if(taken[k]==0)
			{
			  taken[k]=cur_proc+1;
			  count+=weight_block[k];
			  count_edge+=edge_block[k+1]-edge_block[k];
			}
		    }
		  //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - Number of edges for proc "<<cur_proc<<" is "<<count<<"\n";
		  //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - count_edge last for proc "<<cur_proc<<"  = "<<count_edge<<std::endl;
		  if(cur_proc==Mpi_::mpi_rank)
		    dim_edge=count_edge;
		  weights_r=0;
		  count=0;
		  count_edge=0;
		  cur_proc++;
		}
	      if(taken[i]==0)
		{
		  //new ideal value si calculated at each knew processor to manage
		  if(count==0)
		    ideal = static_cast<unsigned int>(floor(weights_r/(Mpi_::mpi_nb-cur_proc)));

		  if(count+weight_block[i]<=ideal)
		    {
		      count+=weight_block[i];
		      count_edge+=edge_block[i+1]-edge_block[i];
		      taken[i] = cur_proc+1;

		      if(count-weight_block[i]==0)
			{
			  //particular case only if count==0 before this node
			  if(count<ideal && count+weight_block[i+1]<=ideal/*+mean_edges*/)
			    {
			      //we verify if the next block is not getting out of the current one
			      bool found=false;
			      for(unsigned int j=0;j<degcum_block[i+1]-degcum_block[i];j++)
				{
				  if(linked_blocks[degcum_block[i]+j]==1)
				    {found=true;break;}
				}
			      //if the next block is not already taken and that it does not have any output block, we take it to not isolate it
			      if(taken[i+1]==0 && found==false && degcum_block[i+2]==degcum_block[i+1])
				{
				  count+=weight_block[i+1];
				  count_edge+=edge_block[i+2]-edge_block[i+1];
				  taken[i+1]=cur_proc+1;
				}
			      //BEWARE IT WAS linked_blocks[i+2]
			      //if the next block is not already taken and that it does have a single output block
			      else if(taken[i+1]==0 && found==false && degcum_block[i+2]==degcum_block[i+1]+1 && edges_subtree[linked_blocks[degcum_block[i+2]]]==weight_block[i+1])
				{
				  count+=weight_block[i+1];
				  count_edge+=edge_block[i+2]-edge_block[i+1];
				  taken[i+1]=cur_proc+1;
				}
			    }
			}
		      if(count<ideal)
			{
			  unsigned int size = degcum_block[i+1]-degcum_block[i];
			  unsigned int *edge_b_loc = (unsigned int *)calloc(size,sizeof(unsigned int));
			  unsigned int *weight_b_loc = (unsigned int *)calloc(size,sizeof(unsigned int));
			  unsigned int *index = (unsigned int *)calloc(size,sizeof(unsigned int));

			  for(unsigned int j=0;j<size;j++)
			    {
			      index[j] = linked_blocks[degcum_block[i]+j];
			      edge_b_loc[j] = edge_block[index[j]+1]-edge_block[index[j]];
			      weight_b_loc[j] = weight_block[index[j]];
			    }
			  //recursive call
			  dist(mean_edges, weights_r,index,edge_b_loc,weight_b_loc,size,ideal,edge_block,weight_block,degcum_block,linked_blocks,taken,edges_subtree,count,count_edge,cur_proc);
			  verif(index,edge_b_loc,weight_b_loc,size,degcum_block,linked_blocks,taken,count,count_edge);
			  free(index);
			  free(edge_b_loc);
			  free(weight_b_loc);
			  if(count>=ideal || weights_r==0)
			    {
			      if(cur_proc==Mpi_::mpi_rank)
				dim_edge = count_edge;
			      weights_r -= count;
			      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - Number of edges for proc "<<cur_proc<<" is "<<count<<"\n";
			      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - count_edge 0 for proc "<<cur_proc<<" ="<<count_edge<<std::endl;
			      count = 0;
			      count_edge=0;
			      cur_proc++;
			    }
			}
		      else
			{
			  if(cur_proc==Mpi_::mpi_rank)
			    dim_edge = count_edge;
			  weights_r -= count;
			  //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - count_edge 1 for proc "<<cur_proc<<" = "<<count_edge<<std::endl;
			  //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - Number of edges for proc "<<cur_proc<<" is "<<count<<"\n";
			  count = 0;
			  count_edge=0;
			  cur_proc++;
			}
		    }
		  else
		    {
		      if(cur_proc==Mpi_::mpi_rank)
			dim_edge = count_edge;
		      weights_r -= count;
		      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - count_edge 2 for proc "<<cur_proc<<" = "<<count_edge<<std::endl;
		      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - Number of edges for proc "<<cur_proc<<" is "<<count<<"\n";
		      count = 0;
		      count_edge=0;
		      cur_proc++;
		      i=i-1;
		    }
		}
	    }
	  else
	    {
	      if(cur_proc==Mpi_::mpi_rank)
		dim_edge = count_edge;
	      weights_r -= count;
	      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - count_edge 3 for proc "<<cur_proc<<" = "<<count_edge<<std::endl;
	      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - Number of edges for proc "<<cur_proc<<" is "<<count<<"\n";
	      count = 0;
	      count_edge=0;
	      cur_proc++;
	    }
	}

      if(Mpi_::mpi_nb==1)
	dim_edge=dim_edge_g;

      /*std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" taken = ";
      for (int i=0;i<b;i++)
	std::cout<<taken[i]<<" ";
	std::cout<<std::endl;*/

      //--------------------------------------------------------------------------------
      //FILL Parallel data structure
      //--------------------------------------------------------------------------------
      //with taken, edge_block & edges_g we can construct the local data structure has done before for the global one
      //with degcum_block & linked_blocks we can get where communications are needed in the local data structure
      //--------------------------------------------------------------------------------
      //local src_nodes & dst_nodes
      //--------------------------------------------------------------------------------
      src_nodes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      dst_nodes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      indexes_edges = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      unsigned int * edges_t_out=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      unsigned int * edges_t_in=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" - dim_edge="<<dim_edge<<std::endl;
      localSrcDstNodes(b,taken,edge_block,src_nodes_g,dst_nodes_g,edges_g,dim_roots_g,dim_leafs_g,roots_g,leafs_g);
      //std::cout<<"Proc "<<Mpi_::mpi_rank<<" dim_node="<<dim_node<<std::endl;
      //--------------------------------------------------------------------------------
      //free global and distribution tables
      //--------------------------------------------------------------------------------
      free(degcum_out_g);free(degcum_in_g);
      free(neigh_out_g);free(neigh_in_g);
      free(edges_g);
      free(degcum_block);free(linked_blocks);free(edges_subtree);
      //--------------------------------------------------------------------------------
      //construction of local degcum_out and degcum_in :
      //represents the cumulative number of inputs and outputs nodes and edges of a node
      //--------------------------------------------------------------------------------
      unsigned int *deg_in2=(unsigned int *)calloc(dim_node,sizeof(unsigned int));
      unsigned int *deg_out2=(unsigned int *)calloc(dim_node,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  deg_out2[src_nodes[i]]++;
	  deg_in2[dst_nodes[i]]++;
	}
      degcum_in=(unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      degcum_out=(unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      for(unsigned int i=1;i<dim_node+1;i++)
	{
	  degcum_out[i]=degcum_out[i-1]+deg_out2[i-1];
	  degcum_in[i]=degcum_in[i-1]+deg_in2[i-1];
	}
      free(deg_in2),free(deg_out2);
      //--------------------------------------------------------------------------------
      //--------------------------------------------------------------------------------
      //TEMPORARY DATA
      //--------------------------------------------------------------------------------
      //degcum_in_t and degcum_out_t = degcum_in and degcum_out to keep the pure local values (without communications problems)
      //--------------------------------------------------------------------------------
      unsigned int *degcum_in_t=(unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      unsigned int *degcum_out_t=(unsigned int *)calloc(dim_node+1,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_node+1;i++)
	{
	  degcum_in_t[i] = degcum_in[i];
	  degcum_out_t[i] = degcum_out[i];
	}

      //--------------------------------------------------------------------------------
      //neigh_out_t and neigh_in_t and edges_t to keep the pure local values (without communications problems)
      //--------------------------------------------------------------------------------
      unsigned int * t_in=(unsigned int *)calloc(dim_node,sizeof(unsigned int));
      unsigned int * t_out=(unsigned int *)calloc(dim_node,sizeof(unsigned int));

      unsigned int * neigh_in_t=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      unsigned int * neigh_out_t=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      //unsigned int * edges_t_out=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      //unsigned int * edges_t_in=(unsigned int *)calloc(dim_edge,sizeof(unsigned int));

      for (int i=0;i<dim_edge;i++) 
	{
	  neigh_out_t[ degcum_out[src_nodes[i]] + t_out[ src_nodes[i] ]++ ]=dst_nodes[i];
	  neigh_in_t[ degcum_in[dst_nodes[i]] + t_in[ dst_nodes[i] ]++ ]=src_nodes[i];
	  edges_t_out[i]=i;
	  //edges_t_in[i]=;
	}
      free(t_out),free(t_in);
      //--------------------------------------------------------------------------------
      //--------------------------------------------------------------------------------

      //--------------------------------------------------------------------------------
      //we have to add edges and nodes for parallelization and communications
      //to do that search root and leaf nodes of the local structure
      //found their direct (distance 1) neighbors in the global data structure
      //construct degcum_tos and degcum_tor and associated edges_tos, edges_tor, nodes_tos, nodes_tor
      //indexes of new elements have to be higher than local elements (at the end of every local data structure)
      //--------------------------------------------------------------------------------
      neigh_out = (unsigned int *)calloc(0,sizeof(unsigned int));
      neigh_in = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_out = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_in = (unsigned int *)calloc(0,sizeof(unsigned int));
      //edges_tor = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_tor_in = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_tor_out = (unsigned int *)calloc(0,sizeof(unsigned int));
      //edges_tos = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_tos_in = (unsigned int *)calloc(0,sizeof(unsigned int));
      edges_tos_out = (unsigned int *)calloc(0,sizeof(unsigned int));
      nodes_tor = (unsigned int *)calloc(0,sizeof(unsigned int));
      nodes_tos = (unsigned int *)calloc(0,sizeof(unsigned int));
      fillLocalDS(b,dst_nodes_g,src_nodes_g,edge_block,taken,degcum_out_t,degcum_in_t,edges_t_out,edges_t_in,neigh_out_t,neigh_in_t,roots_g,leafs_g,dim_roots_g,dim_leafs_g);
      free(degcum_in_t);free(degcum_out_t);
      free(neigh_out_t);free(neigh_in_t);free(edges_t_out);free(edges_t_in);
      free(roots_g);free(leafs_g);
      free(edge_block);free(weight_block);
      free(taken);
      free(src_nodes_g);free(dst_nodes_g);free(weight_g);
      //---------------------------------------------------------
      //PRINTINGS
      //---------------------------------------------------------
      /*std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" src_nodes = ";
      for (int i=0;i<dim_edge;i++)
	std::cout<<src_nodes[i]<<" ";
      std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" dst_nodes = ";
      for (int i=0;i<dim_edge;i++)
	std::cout<<dst_nodes[i]<<" ";
	std::cout<<std::endl;
      */
      /*std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" indexes_nodes = ";
      for (int i=0;i<dim_ind_no;i++)
	std::cout<<indexes_nodes[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" indexes_edges = ";
      for (int i=0;i<dim_ind_ed;i++)
	std::cout<<indexes_edges[i]<<" ";
	std::cout<<std::endl;*/
      /*
      std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_out = ";
      for (int i=1;i<dim_node+1;i++)
	std::cout<<degcum_out[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_in = ";
      for (int i=1;i<dim_node+1;i++)
	std::cout<<degcum_in[i]<<" ";
	std::cout<<std::endl;*/
      
	/*std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" neigh_out = ";
      for (int i=0;i<dim_edge+dim_tor_out;i++)
	std::cout<<neigh_out[i]<<" ";
	std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" neigh_in = ";
      for (int i=0;i<dim_edge+dim_tor_in;i++)
	std::cout<<neigh_in[i]<<" ";
	std::cout<<std::endl;
      */
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_out = ";
      for (int i=0;i<dim_edge+dim_tor_out;i++)
	std::cout<<edges_out[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_in = ";
      for (int i=0;i<dim_edge+dim_tor_in;i++)
	std::cout<<edges_in[i]<<" ";
	std::cout<<std::endl;*/
      
      //std::cout<<std::endl;
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tor = ";
      for (int i=1;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tor[i]<<" ";
	std::cout<<std::endl;*/
	/*std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tos = ";
      for (int i=0;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tos[i]<<" ";
	std::cout<<std::endl;*/

      //std::cout<<std::endl;
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tor_in = ";
      for (int i=1;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tor_in[i]<<" ";
      std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tos_in = ";
      for (int i=0;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tos_in[i]<<" ";
      std::cout<<std::endl;

      //std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tor_out = ";
      for (int i=1;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tor_out[i]<<" ";
      std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" degcum_tos_out = ";
      for (int i=0;i<Mpi_::mpi_nb+1;i++)
	std::cout<<degcum_tos_out[i]<<" ";
	std::cout<<std::endl;*/
      
      //std::cout<<std::endl;
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tor = ";
      for (int i=0;i<dim_tor_out+dim_tor_in;i++)
	std::cout<<edges_tor[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tos = ";
      for (int i=0;i<dim_tos_out+dim_tos_in;i++)
	std::cout<<edges_tos[i]<<" ";
	std::cout<<std::endl;*/

      //std::cout<<std::endl;
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tor_in = ";
      for (int i=0;i<dim_tor_in;i++)
	std::cout<<edges_tor_in[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tos_in = ";
      for (int i=0;i<dim_tos_in;i++)
	std::cout<<edges_tos_in[i]<<" ";
	std::cout<<std::endl;*/

      //std::cout<<std::endl;
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tor_out = ";
      for (int i=0;i<dim_tor_out;i++)
	std::cout<<edges_tor_out[i]<<" ";
	std::cout<<std::endl;*/
      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" edges_tos_out = ";
      for (int i=0;i<dim_tos_out;i++)
	std::cout<<edges_tos_out[i]<<" ";
	std::cout<<std::endl;*/

      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" nodes_tor = ";
      for (int i=0;i<dim_tor_out+dim_tor_in;i++)
	std::cout<<nodes_tor[i]<<" ";
      std::cout<<std::endl;
      std::cout<<"Proc "<<Mpi_::mpi_rank<<" nodes_tos = ";
      for (int i=0;i<dim_tos_out+dim_tos_in;i++)
	std::cout<<nodes_tos[i]<<" ";
	std::cout<<std::endl;*/
    }
    //-------------------------------------------------------------------------------
    //! Initialization of the DDAG_impl
    /*!
      \param file is the path to the .dot input file
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      degcum_in = NULL;degcum_out = NULL;neigh_in = NULL;neigh_out = NULL;edges_out = NULL;edges_in=NULL;src_nodes = NULL;dst_nodes = NULL;indexes_nodes=NULL;indexes_edges=NULL;
      degcum_tos = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      degcum_tos_in = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      degcum_tos_out = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      degcum_tor = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      degcum_tor_in = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      degcum_tor_out = (unsigned int *)calloc(Mpi_::mpi_nb+1,sizeof(unsigned int));
      /*edges_tor=NULL;*/edges_tor_in=NULL;edges_tor_out=NULL;/*edges_tos=NULL;*/edges_tos_in=NULL;edges_tos_out=NULL;nodes_tor=NULL;nodes_tos=NULL;
      dim_node=0;dim_edge=0;dim_roots=0;dim_leafs=0;dim_comm=0;dim_loc=0;
      dim_tor_out=0;dim_tor_in=0;dim_tos_out=0;dim_tos_in=0;
      //lecture du .dot, construction d'une structure séquentielle temporaire
      //algo distribution
      //remplissage structure locale
      distribute(file);
    }
    //-------------------------------------------------------------------------------

  public:

    //-------------------------------------------------------------------------------
    //! Constructor of the DDAG_impl
    /*!
      \param path to the input .dot file
    */
    //-------------------------------------------------------------------------------
    DDAG_impl(const char * file)
    //-------------------------------------------------------------------------------
    {
      //read file
      //distribution algorithm
      //data structure construction
      init(file);
    }
    //-------------------------------------------------------------------------------
    //! Destructor of the DDAG_impl
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DDAG_impl()
    //-------------------------------------------------------------------------------
    {
      if(NULL!=degcum_in)
	free(degcum_in);
      if(NULL!=degcum_out)
	free(degcum_out);
      if(NULL!=neigh_in)
	free(neigh_in);
      if(NULL!=neigh_out)
	free(neigh_out);
      if(NULL!=edges_out)
	free(edges_out);
      if(NULL!=edges_in)
	free(edges_in);
      if(NULL!=src_nodes)
	free(src_nodes);
      if(NULL!=dst_nodes)
	free(dst_nodes);
      if(NULL!=indexes_nodes)
	free(indexes_nodes);
      if(NULL!=indexes_edges)
	free(indexes_edges);
      if(NULL!=degcum_tos)
	free(degcum_tos);
      if(NULL!=degcum_tos_in)
	free(degcum_tos_in);
      if(NULL!=degcum_tos_out)
	free(degcum_tos_out);
      if(NULL!=degcum_tor)
	free(degcum_tor);
      if(NULL!=degcum_tor_in)
	free(degcum_tor_in);
      if(NULL!=degcum_tor_out)
	free(degcum_tor_out);
      if(NULL!=edges_tor_in)
	free(edges_tor_in);
      if(NULL!=edges_tos_in)
	free(edges_tos_in);
      if(NULL!=edges_tor_out)
	free(edges_tor_out);
      if(NULL!=edges_tos_out)
	free(edges_tos_out);
      if(NULL!=nodes_tor)
	free(nodes_tor);
      if(NULL!=nodes_tos)
      free(nodes_tos);
    }
    //-------------------------------------------------------------------------------
    //! get the number of nodes
    /*!
      \return dim_node the number of nodes in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimNode(){return dim_node;}
    //-------------------------------------------------------------------------------
    //! get the number of edges
    /*!
      \return dim_edge the number of edges in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimEdge(){return dim_edge;}
    //-------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------
    //! get global number of nodes
    /*!
      \return dim_node the number of nodes in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getGDimNode(){return dim_node_g;}
    //-------------------------------------------------------------------------------
    //! get the number of edges
    /*!
      \return dim_edge the number of edges in the local data structure
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getGDimEdge(){return dim_edge_g;}
    //-------------------------------------------------------------------------------
    //! get the number of pure local nodes
    /*!
      \return dim_node the number of pure local nodes
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimLoc(){return dim_loc;}
    //-------------------------------------------------------------------------------
    //! get the number of communication nodes
    /*!
      \return dim_node the number of communication nodes
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimComm(){return dim_comm;}
    //-------------------------------------------------------------------------------
    //! get the number of roots nodes
    /*!
      \return dim_node the number of roots nodes
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimRoots(){return dim_roots;}
    //-------------------------------------------------------------------------------
    //! get the number of leafs nodes
    /*!
      \return dim_node the number of leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimLeafs(){return dim_leafs;}
    //-------------------------------------------------------------------------------
    //! get the number of elements to receive in
    /*!
      \return the number of elements
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimTorIn(){return dim_tor_in;}
    //-------------------------------------------------------------------------------
    //! get the number of elements to receive out
    /*!
      \return the number of elements
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDimTorOut(){return dim_tor_out;}
    //-------------------------------------------------------------------------------
    //! get the global index of a local node
    /*!
      \return the global index of a local node
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getIndexNode(unsigned int ind){return indexes_nodes[ind];}
    //-------------------------------------------------------------------------------
    //! get the global index of a local edge
    /*!
      \return the global index of a local edge
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getIndexEdge(unsigned int ind){return indexes_edges[ind];}
    //-------------------------------------------------------------------------------
    //!  to get the iterator rank on the source node of the index edge
    /*!
      \return the iterator rank on the source node
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getSrcNode(unsigned int r){return src_nodes[r];}
    //-------------------------------------------------------------------------------
    //!  to get the iterator rank on the destination node of the index edge
    /*!
      \return the iterator rank on the destination node
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDstNode(unsigned int r){return dst_nodes[r];}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input edges for the node at rank r
    /*!
      \param r is the rank of the node to get in edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    std::vector<iterator_dag> getInEdges(unsigned int r)
    {
      std::vector<iterator_dag> v;
      for(unsigned int i=degcum_in[r];i<degcum_in[r+1];i++)
	  v.push_back(iterator_dag(edges_in[i]));
      return v;
    }
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output edges for the node at rank r
    /*!
      \param r is the rank of the node to get out edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    std::vector<iterator_dag> getOutEdges(unsigned int r)
    {
      std::vector<iterator_dag> v;
      for(unsigned int i=degcum_out[r];i<degcum_out[r+1];i++)
	  v.push_back(iterator_dag(edges_out[i]));
      return v;
    }
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input nodes for the node at rank r
    /*!
      \param r is the rank of the node to get in nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    std::vector<iterator_dag> getInNodes(unsigned int r)
    {
      std::vector<iterator_dag> v;
      for(unsigned int i=degcum_in[r];i<degcum_in[r+1];i++)
	  v.push_back(iterator_dag(neigh_in[i]));
      return v;
    }
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output nodes for the node at rank r
    /*!
      \param r is the rank of the node to get out nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    std::vector<iterator_dag> getOutNodes(unsigned int r)
    {
      std::vector<iterator_dag> v;
      for(unsigned int i=degcum_out[r];i<degcum_out[r+1];i++)
	  v.push_back(iterator_dag(neigh_out[i]));
      return v;
    }
    //-------------------------------------------------------------------------------
    //!  to get the number of elements to send to a precised processor rank
    /*!
      \param rank is the rank of the processor to send elements to
      \return the number of elements to send to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbNodesToSend(unsigned int rank){return degcum_tos[rank+1]-degcum_tos[rank];}
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to send to other processors
    /*!
      \param rank is the rank of the processor to send data to
      \return a table with indexes to send
    */
    //-------------------------------------------------------------------------------
    unsigned int * getNodesToSend(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tos[rank+1]-degcum_tos[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tos[rank+1]-degcum_tos[rank];i++)
	  ret[i] = nodes_tos[degcum_tos[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to send to a precised processor rank
    /*!
      \param rank is the rank of the processor to send edges to
      \return the number of edges to send to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToSend(unsigned int rank){return (degcum_tos_in[rank+1]+degcum_tos_out[rank+1])-(degcum_tos_in[rank]+degcum_tos_out[rank]);}
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to send in to a precised processor rank
    /*!
      \param rank is the rank of the processor to send edges to
      \return the number of edges to send in to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToSendIn(unsigned int rank){return degcum_tos_in[rank+1]-degcum_tos_in[rank];}
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to send out to a precised processor rank, those edges go inside the processor rank
    /*!
      \param rank is the rank of the processor to send edges to
      \return the number of edges to send out to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToSendOut(unsigned int rank){return degcum_tos_out[rank+1]-degcum_tos_out[rank];}
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to send to other processors, those edges go outside the processor rank
    /*!
      \param rank is the rank of the processor to send data to
      \return a table with indexes to send
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToSend(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc((degcum_tos_in[rank+1]+degcum_tos_out[rank+1])-(degcum_tos_in[rank]+degcum_tos_out[rank]),sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tos_in[rank+1]-degcum_tos_in[rank];i++)
	ret[i] = edges_tos_in[degcum_tos_in[rank]+i];
      for(unsigned int i=0;i<degcum_tos_out[rank+1]-degcum_tos_out[rank];i++)
	ret[degcum_tos_in[rank+1]-degcum_tos_in[rank]+i] = edges_tos_out[degcum_tos_out[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to send in to other processors
    /*!
      \param rank is the rank of the processor to send in data to
      \return a table with indexes to send in
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToSendIn(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tos_in[rank+1]-degcum_tos_in[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tos_in[rank+1]-degcum_tos_in[rank];i++)
	ret[i] = edges_tos_in[degcum_tos_in[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to send out to other processors
    /*!
      \param rank is the rank of the processor to send out data to
      \return a table with indexes to send out
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToSendOut(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tos_out[rank+1]-degcum_tos_out[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tos_out[rank+1]-degcum_tos_out[rank];i++)
	ret[degcum_tos_in[rank+1]-degcum_tos_in[rank]+i] = edges_tos_out[degcum_tos_out[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to get the number of elements to receive to a precised processor rank
    /*!
      \param rank is the rank of the processor to receive elements to
      \return the number of elements to receive to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbNodesToRecv(unsigned int rank){return degcum_tor[rank+1]-degcum_tor[rank];}
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to receive to other processors
    /*!
      \param rank is the rank of the processor to receive data to
      \return a table with indexes to receive
    */
    //-------------------------------------------------------------------------------
    unsigned int * getNodesToRecv(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tor[rank+1]-degcum_tor[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tor[rank+1]-degcum_tor[rank];i++)
	ret[i] = nodes_tor[degcum_tor[rank]+i];;

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to receive to a precised processor rank
    /*!
      \param rank is the rank of the processor to receive edges to
      \return the number of edges to receive to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToRecv(unsigned int rank){return (degcum_tor_in[rank+1]+degcum_tor_out[rank+1])-(degcum_tor_in[rank]+degcum_tor_out[rank]);}
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to receive in to a precised processor rank
    /*!
      \param rank is the rank of the processor to receive edges to
      \return the number of edges to receive in to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToRecvIn(unsigned int rank){return degcum_tor_in[rank+1]-degcum_tor_in[rank];}
    //-------------------------------------------------------------------------------
    //!  to get the number of edges to receive out to a precised processor rank
    /*!
      \param rank is the rank of the processor to receive edges to
      \return the number of edges to receive out to this processor
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getNbEdgesToRecvOut(unsigned int rank){return degcum_tor_out[rank+1]-degcum_tor_out[rank];}
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to receive to other processors
    /*!
      \param rank is the rank of the processor to receive data to
      \return a table with indexes to receive
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToRecv(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc((degcum_tor_in[rank+1]+degcum_tor_out[rank+1])-(degcum_tor_in[rank]+degcum_tor_out[rank]),sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tor_in[rank+1]-degcum_tor_in[rank];i++)
	  ret[i] = edges_tor_in[degcum_tor_in[rank]+i];
      for(unsigned int i=0;i<degcum_tor_out[rank+1]-degcum_tor_out[rank];i++)
	  ret[degcum_tor_in[rank+1]-degcum_tor_in[rank]+i] = edges_tor_out[degcum_tor_out[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to receive in to other processors
    /*!
      \param rank is the rank of the processor to receive in data to
      \return a table with indexes to receive in
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToRecvIn(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tor_in[rank+1]-degcum_tor_in[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tor_in[rank+1]-degcum_tor_in[rank];i++)
	  ret[i] = edges_tor_in[degcum_tor_in[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------
    //!  to fill a table with indexes to receive out to other processors
    /*!
      \param rank is the rank of the processor to receive out data to
      \return a table with indexes to receive out
    */
    //-------------------------------------------------------------------------------
    unsigned int * getEdgesToRecvOut(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int * ret = (unsigned int *)calloc(degcum_tor_out[rank+1]-degcum_tor_out[rank],sizeof(unsigned int));
      for(unsigned int i=0;i<degcum_tor_out[rank+1]-degcum_tor_out[rank];i++)
	  ret[degcum_tor_in[rank+1]-degcum_tor_in[rank]+i] = edges_tor_out[degcum_tor_out[rank]+i];

      return ret;
    }
    //-------------------------------------------------------------------------------

  };

  //-------------------------------------------------------------------------------
  //!  Second specialization of the DDAG_impl class
  /*!
    Computations on edges only
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG_impl<false,true>
  //-------------------------------------------------------------------------------
  {
  protected:

  public:
    DDAG_impl(const char * binFile)
    {
    }
    ~DDAG_impl()
    {
    }

  };

  //-------------------------------------------------------------------------------
  //!  Third specialization of the DDAG_impl class
  /*!
    Computations on nodes only
  */
  //-------------------------------------------------------------------------------
  template<> struct DDAG_impl<true,false>
  //-------------------------------------------------------------------------------
  {
  protected:

  public:
    DDAG_impl(const char * binFile)
    {
    }
    ~DDAG_impl()
    {
    }

  };

}

#endif
