/*! \file pmap_edges.hpp
 *  \brief Definitions of the object PMap_Edges used by the user to map data to the edges of a DDAG. Three template specializations are available.
 */
#ifndef DPMAP_EDGES_H
#define DPMAP_EDGES_H

#include "dpmap_edges_impl.hpp"

namespace skelgis{

  //-------------------------------------------------------------------------------
  //!  base template class of Property map on edges
  /*!
    This template class defines the base PMap_Edges
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam node_access indicates if nodes have to have access to this property during calculation (>=1) or not (=0)

    In the case of a basic type T, node_access can't be >1
    In the case of a pointer type T*, if node_access>0 then it indicates the number of elements to shared at the end and at the beginning of the pointer
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T, int node_access> struct DPMap_Edges;
  //-------------------------------------------------------------------------------
  //!  first specialization of DPMap_Edges
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The node_access is specialized to the value 1
    The size is used with its default value 1
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges<DD,T,1>
  //-------------------------------------------------------------------------------
  {
  protected :
    bool ptr_const;
    DPMap_Edges_impl<DD,T,1> * dpmap;

  public:
    typedef DPMap_Edges_impl<DD,T,1> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T,1>(dag,file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T,1>(dag,val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Edges_impl pointer
    /*!
      \param pointer to a DPMap_Edges_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DPMap_Edges_impl<DD,T,1> * p)
    {
      ptr_const=true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges()
    {
      //do not destroy the pointer if the DPMap was constructed from an existing dpmap_impl
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Edges_impl<DD,T,1> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return dpmap->getSrcNode(it);}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return dpmap->getDstNode(it);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications(){dpmap->start_communications();}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications(){dpmap->end_communications();}
    //-------------------------------------------------------------------------------
    
  };
  //-------------------------------------------------------------------------------
  //!  second specialization of DPMap_Edges
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The node_access is specialized to the value 0
    The size is used with its default value 1
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges<DD,T,0>
  {
  protected :
    bool ptr_const;
    DPMap_Edges_impl<DD,T,0> * dpmap;

  public:
    typedef DPMap_Edges_impl<DD,T,0> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T,0>(dag,file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T,0>(dag,val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Edges_impl pointer
    /*!
      \param pointer to a DPMap_Edges_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DPMap_Edges_impl<DD,T,0> * p)
    {
      ptr_const=true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges()
    {
      //do not destroy the pointer if the DPMap was constructed from an existing dpmap_impl
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Edges_impl<DD,T,0> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return dpmap->getSrcNode(it);}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return dpmap->getDstNode(it);}
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Edges
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam node_access indicates if nodes have to have access to this property during calculation (>=1) or not (=0)

    In the case of a basic type T, node_access can't be >1
    In the case of a pointer type T*, if node_access>0 then it indicates the number of elements to shared at the end and at the beginning of the pointer

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The node_access is not specialized and is [0,...,size] size is precised at constructor call
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T,int node_access> struct DPMap_Edges<DD,T*,node_access>
  {
    protected :
    bool ptr_const;
    DPMap_Edges_impl<DD,T*,node_access> * dpmap;

  public:
    typedef DPMap_Edges_impl<DD,T*,node_access> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a file
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
      \param sizes is the table of sizes of the pointer table mapped on each edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,node_access>(dag,file,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a value
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
      \param size is the size of the pointer table mapped on edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,node_access>(dag,val,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a file
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
      \param size is the size of the pointer table mapped on edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,node_access>(dag,file,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a value
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
      \param size is the size of the pointer table mapped on edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,node_access>(dag,val,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Edges_impl pointer
    /*!
      \param pointer to a DPMap_Edges_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DPMap_Edges_impl<DD,T*,node_access> * p)
    {
      ptr_const=true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges()
    {
      //do not destroy the pointer if the DPMap was constructed from an existing dpmap_impl
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Edges_impl<DD,T*,node_access> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T*& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return dpmap->getSrcNode(it);}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return dpmap->getDstNode(it);}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return dpmap->getDataSize(it);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications(){dpmap->start_communications();}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications(){dpmap->end_communications();}
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    inline void print(){dpmap->print();}
    //-------------------------------------------------------------------------------
    inline void print_in(iterator_dag it){dpmap->print_in(it);}
    inline void print_out(iterator_dag it){dpmap->print_out(it);}
  };

  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Edges
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam node_access indicates if nodes have to have access to this property during calculation (>=1) or not (=0)

    In the case of a basic type T, node_access can't be >1
    In the case of a pointer type T*, if node_access>0 then it indicates the number of elements to shared at the end and at the beginning of the pointer

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The node_access is not specialized and is [0,...,size] size is precised at construction
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges<DD,T*,1>
  {
    protected :
    bool ptr_const;
    DPMap_Edges_impl<DD,T*,1> * dpmap;

  public:
    typedef DPMap_Edges_impl<DD,T*,1> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a file
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
      \param sizes is the table of sizes of the pointer table mapped on each edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,1>(dag,file,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a value
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
      \param size is the size of the pointer table mapped on edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,1>(dag,val,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,1>(dag,file,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,1>(dag,val,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Edges_impl pointer
    /*!
      \param pointer to a DPMap_Edges_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DPMap_Edges_impl<DD,T*,1> * p)
    {
      ptr_const=true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges()
    {
      //do not destroy the pointer if the DPMap was constructed from an existing dpmap_impl
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Edges_impl<DD,T*,1> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T*& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return dpmap->getSrcNode(it);}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return dpmap->getDstNode(it);}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return dpmap->getDataSize(it);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications(){dpmap->start_communications();}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications(){dpmap->end_communications();}
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    inline void print(){dpmap->print();}
    //-------------------------------------------------------------------------------
  };

  //-------------------------------------------------------------------------------
  //!  fourth specialization of DPMap_Edges
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data
    \tparam size is the number of elements at each node

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is specialized to the value 0
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges<DD,T*,0>
  {
  protected :
    bool ptr_const;
    DPMap_Edges_impl<DD,T*,0> * dpmap;

  public:
    typedef DPMap_Edges_impl<DD,T*,0> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a file
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
      \param sizes is the table of sizes of the pointer table mapped on each edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,0>(dag,file,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges from a value
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
      \param size is the size of the pointer table mapped on edges
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,unsigned int * sizes)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,0>(dag,val,sizes);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of edges wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, const char * file,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,0>(dag,file,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DD &dag, T val,const unsigned int size)
    {
      ptr_const=false;
      dpmap = new DPMap_Edges_impl<DD,T*,0>(dag,val,size);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Edges_impl pointer
    /*!
      \param pointer to a DPMap_Edges_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges(DPMap_Edges_impl<DD,T*,0> * p)
    {
      ptr_const=true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges()
    {
      //do not destroy the pointer if the DPMap was constructed from an existing dpmap_impl
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Edges_impl<DD,T*,0> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T*& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return dpmap->getSrcNode(it);}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return dpmap->getDstNode(it);}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return dpmap->getDataSize(it);}
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    inline void print(){dpmap->print();}
    //-------------------------------------------------------------------------------
  };
  
};


#endif
