/*! \file pmap_edges_impl.hpp
 *  \brief Definitions of the object PMap_Edges_impl used by the user to map data to the edges of a DDAG. Three template specializations are available.
 */
#ifndef DPMAP_EDGES_IMPL_H
#define DPMAP_EDGES_IMPL_H

#include "ddag.hpp"
#include "../../iterators/DDAG/iterator.hpp"
#include "../../util/communications.hpp"

namespace skelgis{

  //-------------------------------------------------------------------------------
  //!  base template class of Property map on edges
  /*!
    This template class defines the base PMap_Edges_impl
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam overlap is the overlap size needed to compute the data
    \tparam size is the number of elements at each edge

    for a standard type T (two first specializations) the overlap can be 0 or 1 and the size is the default one = 1
    for a pointer type T (two second specializations) the overlap can be any value and the size is > 1 because a table is gonna be mapped to each edge
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T, int node_access> struct DPMap_Edges_impl;
  //-------------------------------------------------------------------------------
  //!  first specialization of DPMap_Edges_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The overlap is specialized to the value 1
    The size is used with its default value 1
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges_impl<DD,T,1>
  //-------------------------------------------------------------------------------
  {
  public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;
    /*!< object to make MPI communications */
    Communications<T> * comm;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      std::ifstream pmapfile;
      pmapfile.open(file);
      
      std::string line;
      T any=0;
      for(unsigned int i=0;i<dim_edge+dim_tor;i++)
	{
	  for(unsigned int j=0;j<ddag->getIndexEdge(i);j++)
	    {
	      std::getline(pmapfile,line);
	      sscanf(line.data(),"%lf", &any);
	    }
	}

      pmapfile.close();
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      data = new T[dim_edge+dim_tor];
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate edges to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      data = new T[dim_edge+dim_tor];
      for(unsigned int i=0;i<dim_edge+dim_tor;i++)
	{
	  data[i] = val;
	}
    }
    //-------------------------------------------------------------------------------
    //!  init function for MPI communications
    /*!
    */
    //-------------------------------------------------------------------------------
    void init_comm()
    {
      //init the communication object
      comm = new Communications<T>();
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_edge is the number of local edges */
    unsigned int dim_edge;
    /*!< dim_tor is the number of additionnal nodes to receive */
    unsigned int dim_tor;
    /*!< data mapped on edges of the ddag */
    T * data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //communication initialization
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges_impl()
    //-------------------------------------------------------------------------------
    {
      delete [] data;
      delete comm;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return iterator_dag(ddag->getDimEdge());}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return iterator_dag(ddag->getSrcNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return iterator_dag(ddag->getDstNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send
	  unsigned int nb_toSend = ddag->getNbEdgesToSend(j);
	  if(nb_toSend>0)
	    {
	      T * toSend = (T*)calloc(nb_toSend,sizeof(T));
	      unsigned int * indexes = ddag->getEdgesToSend(j);
	      for(unsigned int i=0;i<nb_toSend;i++)
		toSend[i] = data[indexes[i]];

	      comm->ISend(toSend, nb_toSend, j);

	      free(indexes);
	      free(toSend);
	    }
	  //prepare data to receive
	  unsigned int nb_toRecv = ddag->getNbEdgesToRecv(j);
	  if(nb_toRecv>0)
	    comm->IRecv(nb_toRecv, j);
	 
	}
    }
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send
	  unsigned int nb_toSend = ddag->getNbEdgesToSend(j);
	  if(nb_toSend>0)
	    comm->Wait_s(j);
	  unsigned int nb_toRecv = ddag->getNbEdgesToRecv(j);
	  if(nb_toRecv>0)
	    {
	      T * toRecv = (T*)calloc(nb_toRecv,sizeof(T));
	      comm->Wait_r(toRecv,j,nb_toRecv);
	      unsigned int * indexes = ddag->getEdgesToRecv(j);
	      //change values in data
	      for(int k=0;k<nb_toRecv;k++)
		data[indexes[k]] = toRecv[k];
	    
	      free(indexes);
	      free(toRecv);
	    }
	}
    }
    //-------------------------------------------------------------------------------

  };
  //-------------------------------------------------------------------------------
  //!  second specialization of DPMap_Edges_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The overlap is specialized to the value 0
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges_impl<DD,T,0>
  {

    public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      std::ifstream pmapfile;
      pmapfile.open(file);
      
      std::string line;
      T any=0;
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  for(unsigned int j=0;j<ddag->getIndexEdge(i);j++)
	    {
	      std::getline(pmapfile,line);
	      sscanf(line.data(),"%lf", &any);
	    }
	}

      pmapfile.close();
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      data = new T[dim_edge];
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate edges to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      data = new T[dim_edge];
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  data[i] = val;
	}
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_edge is the number of local edges */
    unsigned int dim_edge;
    /*!< data mapped on edges of the ddag */
    T * data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges_impl()
    //-------------------------------------------------------------------------------
    {
      delete [] data;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return iterator_dag(ddag->getDimEdge());}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return iterator_dag(ddag->getSrcNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return iterator_dag(ddag->getDstNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    inline void start_communications(){}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    inline void end_communications(){}
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Edges_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data
    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is not specialized and is [0,...,size]
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T,int node_access> struct DPMap_Edges_impl<DD,T*,node_access>
  {
  public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;
    /*!< object to make MPI communications */
    CommunicationsInOut<T> * comm;
    /*!< size of the pointer mapped to edges */
    unsigned int * sizes;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      std::ifstream entries(file,std::ios::in);
      if (!entries){
	std::cerr << "Impossible to open the " << file <<"\n";
	exit(EXIT_FAILURE);
      }

      unsigned int nb_line = 0;
      while (!entries.eof()){
	//read line
	std::string line;
	getline (entries, line);
	for(unsigned int j=0;j<dim_edge;j++)
	  {
	    if(ddag->getIndexEdge(j)==nb_line)
	      {
		//init data table for this edge
		data[j] = (T*)calloc(sizes[j],sizeof(T));
		int found1 = -1;
		int found2 = -1;
		for(unsigned int i=0;i<sizes[j];i++)
		  {
		    found1 = found2+1;
		    found2 = line.find(",",found1);
		    std::string val = line.substr(found1,found2-found1);
		    //memcpy(&data[j][i], value, sizeof(T));
		    data[j][i] = atof(val.data());
		  }
	      }
	  }
	nb_line++;
      }
      entries.close();
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      data = (T**)calloc(dim_edge+dim_tor,sizeof(T*));
      for(unsigned int j=dim_edge;j<dim_edge+dim_tor;j++)
	{
	  data[j] = (T*)calloc(node_access,sizeof(T));
	  for(unsigned int i=0;i<sizes[j];i++)
	    data[j][i] = 1.0;
	}
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate edges to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      data = (T**)calloc(dim_edge+dim_tor,sizeof(T*));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  data[i] = (T*)calloc(sizes[i],sizeof(T));
	  for(unsigned int j=0;j<sizes[i];j++)
	    data[i][j] = val;
	}
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	{
	  data[i] = (T*)calloc(node_access,sizeof(T));
	  for(unsigned int j=0;j<node_access;j++)
	    data[i][j] = val;
	}
    }
    //-------------------------------------------------------------------------------
    //!  init function for MPI communications
    /*!
    */
    //-------------------------------------------------------------------------------
    void init_comm()
    {
      //init the communication object
      comm = new CommunicationsInOut<T>();
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_edge is the number of local edges */
    unsigned int dim_edge;
    /*!< dim_tor is the number of additionnal nodes to receive */
    unsigned int dim_tor;
    /*!< data mapped on edges of the ddag */
    T ** data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();

      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = node_access;
      
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();

      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = node_access;
      
      //init comm
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s;
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = node_access;
      //for(unsigned int i=0;i<dim_edge+dim_tor;i++)
      //sizes[i] = s;
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s;
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = node_access;
      //for(unsigned int i=0;i<dim_edge+dim_tor;i++)
      //sizes[i] = s;
      //init comm
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges_impl()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge+dim_tor;i++)
	free(data[i]);
      free(data);
      if(sizes!=NULL)
	free(sizes);
      
      delete comm;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return iterator_dag(ddag->getDimEdge());}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return iterator_dag(ddag->getSrcNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return iterator_dag(ddag->getDstNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return sizes[it._rank];}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send in another processor
	  unsigned int nb_toSend_in = ddag->getNbEdgesToSendIn(j);
	  if(nb_toSend_in>0)
	    {
	      //the size of toSend_in is multiplied by node_access
	      T * toSend_in = (T*)calloc(nb_toSend_in*node_access,sizeof(T));
	      unsigned int * indexes_in = ddag->getEdgesToSendIn(j);
	      //to send inside another processor -> then send the node_access last values of ptr
	      for(unsigned int i=0;i<nb_toSend_in*node_access;i=i+node_access)
		{
		  T * ptr = data[indexes_in[i/node_access]];
		  for(unsigned int k=0;k<node_access;k++)
		    toSend_in[i+k] = ptr[sizes[indexes_in[i/node_access]]-node_access+k];
		}

	      //a mettre en place, envoi de proc 0 vers proc1 à rechercher dans la sortie à l'iteration 337
	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" to send in to proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toSend_in*node_access;i++)
		std::cout<<toSend_in[i]<<" ";
		std::cout<<std::endl;*/
		
	      comm->ISendIn(toSend_in, nb_toSend_in*node_access, j);

	      free(indexes_in);
	      free(toSend_in);
	    }
	  //prepare data to receive out (correspond to send in)
	  unsigned int nb_toRecv_out = ddag->getNbEdgesToRecvOut(j);
	  if(nb_toRecv_out>0)
	    comm->IRecvOut(nb_toRecv_out*node_access, j);

	  //prepare data to send out another processor
	  unsigned int nb_toSend_out = ddag->getNbEdgesToSendOut(j);
	  if(nb_toSend_out>0)
	    {
	      T * toSend_out = (T*)calloc(nb_toSend_out*node_access,sizeof(T));
	      unsigned int * indexes_out = ddag->getEdgesToSendOut(j);
	      //to send outside another processor -> then send the node_access first values of ptr
	      for(unsigned int i=0;i<nb_toSend_out*node_access;i=i+node_access)
		{
		  T * ptr = data[indexes_out[i/node_access]];
		  for(unsigned int k=0;k<node_access;k++)
		  toSend_out[i+k] = ptr[k];
		}

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" to send out to proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toSend_out*node_access;i++)
		std::cout<<toSend_out[i]<<" ";
		std::cout<<std::endl;*/

	      comm->ISendOut(toSend_out, nb_toSend_out*node_access, j);

	      free(indexes_out);
	      free(toSend_out);
	    }

	  //prepare data to recv in
	  unsigned int nb_toRecv_in = ddag->getNbEdgesToRecvIn(j);
	  if(nb_toRecv_in>0)
	    comm->IRecvIn(nb_toRecv_in*node_access, j);
	}
    }
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //end data to send in
 	  unsigned int nb_toSend_in = ddag->getNbEdgesToSendIn(j);
	  if(nb_toSend_in>0)
	    comm->Wait_sIn(j);
	  
	  //end data to receive out
	  unsigned int nb_toRecv_out = ddag->getNbEdgesToRecvOut(j);
	  if(nb_toRecv_out>0)
	    {
	      T * toRecv_out = (T*)calloc(nb_toRecv_out*node_access,sizeof(T));
	      comm->Wait_rOut(toRecv_out,j,nb_toRecv_out*node_access);
	      unsigned int * indexes_out = ddag->getEdgesToRecvOut(j);
	      //change values in data
	      for(unsigned int k=0;k<nb_toRecv_out;k++)
		{
		  //pointer to modify
		  T * ptr = data[indexes_out[k]];
		  for(unsigned int i=0;i<node_access;i++)
		    ptr[i] = toRecv_out[k*node_access+i];
		}
	      
	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" received out from proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toRecv_out*node_access;i++)
		std::cout<<toRecv_out[i]<<" ";
		std::cout<<std::endl;*/

	      free(indexes_out);
	      free(toRecv_out);
	    }
	  //end data to send out
 	  unsigned int nb_toSend_out = ddag->getNbEdgesToSendOut(j);
	  if(nb_toSend_out>0)
	    comm->Wait_sOut(j);
	  
	  //end data to receive in
	  unsigned int nb_toRecv_in = ddag->getNbEdgesToRecvIn(j);
	  if(nb_toRecv_in>0)
	    {
	      T * toRecv_in = (T*)calloc(nb_toRecv_in*node_access,sizeof(T));
	      comm->Wait_rIn(toRecv_in,j,nb_toRecv_in*node_access);
	      unsigned int * indexes_in = ddag->getEdgesToRecvIn(j);
	      //change values in data
	      for(int k=0;k<nb_toRecv_in;k++)
		{
		  //pointer to modify
		  T * ptr = data[indexes_in[k]];
		  for(unsigned int i=0;i<node_access;i++)
		    ptr[i] = toRecv_in[k*node_access+i];
		}

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" received in from proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toRecv_in*node_access;i++)
		std::cout<<toRecv_in[i]<<" ";
		std::cout<<std::endl;*/

	      free(indexes_in);
	      free(toRecv_in);
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  int N = sizes[i];
	  T * dat = data[i];
	  printf("Proc %i - Artery %i : ",Mpi_::mpi_rank,ddag->getIndexEdge(i));
	  for(int i=0;i<N;i++)
	    printf("dat[%i] = %f,",i,dat[i]);
	  printf("\n");
	}
    }
    //-------------------------------------------------------------------------------
  };

  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Edges_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data
    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is not specialized and is [0,...,size]
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges_impl<DD,T*,1>
  {
  public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;
    /*!< object to make MPI communications */
    CommunicationsInOut<T> * comm;
    /*!< size of the pointer mapped to edges */
    unsigned int * sizes;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      std::ifstream entries(file,std::ios::in);
      if (!entries){
	std::cerr << "Impossible to open the " << file <<"\n";
	exit(EXIT_FAILURE);
      }

      unsigned int nb_line = 0;
      while (!entries.eof()){
	//read line
	std::string line;
	getline (entries, line);
	for(unsigned int j=0;j<dim_edge;j++)
	  {
	    if(ddag->getIndexEdge(j)==nb_line)
	      {
		//init data table for this edge
		data[j] = (T*)calloc(sizes[j],sizeof(T));
		int found1 = -1;
		int found2 = -1;
		for(unsigned int i=0;i<sizes[j];i++)
		  {
		    found1 = found2+1;
		    found2 = line.find(",",found1);
		    std::string val = line.substr(found1,found2-found1);
		    data[j][i] = atof(val.data());
		    //const char * value = line.data();
		    //memcpy(&data[j][i], value, sizeof(T));
		  }
	      }
	  }
	nb_line++;
      }
      entries.close();
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      data = (T**)calloc(dim_edge+dim_tor,sizeof(T*));
      for(unsigned int j=dim_edge;j<dim_edge+dim_tor;j++)
	{
	  data[j] = (T*)calloc(1,sizeof(T));
	  for(unsigned int i=0;i<sizes[j];i++)
	    data[j][i] = 1.0;
	}
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate edges to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      data = (T**)calloc(dim_edge+dim_tor,sizeof(T*));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  data[i] = (T*)calloc(sizes[i],sizeof(T));
	  for(unsigned int j=0;j<sizes[i];j++)
	    data[i][j] = val;
	}
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	data[i] = (T*)calloc(1,sizeof(T));
    }
    //-------------------------------------------------------------------------------
    //!  init function for MPI communications
    /*!
    */
    //-------------------------------------------------------------------------------
    void init_comm()
    {
      //init the communication object
      comm = new CommunicationsInOut<T>();
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_edge is the number of local edges */
    unsigned int dim_edge;
    /*!< dim_tor is the number of additionnal nodes to receive */
    unsigned int dim_tor;
    /*!< data mapped on edges of the ddag */
    T ** data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();

      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = 1;
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();

      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = 1;
      //init comm
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s;
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = 1;
      //for(unsigned int i=0;i<dim_edge+dim_tor;i++)
      //sizes[i] = s;
      //communication initialization
      init_comm();
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      dim_tor = ddag->getDimTorIn()+ddag->getDimTorOut();
      sizes = (unsigned int *)calloc(dim_edge+dim_tor,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s;
      for(unsigned int i=dim_edge;i<dim_edge+dim_tor;i++)
	sizes[i] = 1;
      //for(unsigned int i=0;i<dim_edge+dim_tor;i++)
      //sizes[i] = s;
      //init comm
      init_comm();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges_impl()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge+dim_tor;i++)
	free(data[i]);
      free(data);
      if(sizes!=NULL)
	free(sizes);
      
      delete comm;
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return iterator_dag(ddag->getDimEdge());}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return iterator_dag(ddag->getSrcNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return iterator_dag(ddag->getDstNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return sizes[it._rank];}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    void start_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //prepare data to send in
	  unsigned int nb_toSend_in = ddag->getNbEdgesToSendIn(j);
	  if(nb_toSend_in>0)
	    {
	      //the size of toSend_in is multiplied by node_access
	      T * toSend_in = (T*)calloc(nb_toSend_in,sizeof(T));
	      unsigned int * indexes_in = ddag->getEdgesToSendIn(j);
	      for(unsigned int i=0;i<nb_toSend_in;i++)
		{
		  T * ptr = data[indexes_in[i]];
		  //as we send to another processor that receive in, we give the node_access last values of ptr
		  toSend_in[i] = ptr[sizes[indexes_in[i]]-1];
		}

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" to send in to proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toSend_in;i++)
		std::cout<<toSend_in[i]<<" ";
		std::cout<<std::endl;*/
		
	      comm->ISendIn(toSend_in, nb_toSend_in, j);

	      free(indexes_in);
	      free(toSend_in);
	    }
	  //prepare data to receive out (correspond to send in)
	  unsigned int nb_toRecv_out = ddag->getNbEdgesToRecvOut(j);
	  if(nb_toRecv_out>0)
	    comm->IRecvOut(nb_toRecv_out, j);

	  //prepare data to send out
	  unsigned int nb_toSend_out = ddag->getNbEdgesToSendOut(j);
	  if(nb_toSend_out>0)
	    {
	      T * toSend_out = (T*)calloc(nb_toSend_out,sizeof(T));
	      unsigned int * indexes_out = ddag->getEdgesToSendOut(j);
	      for(unsigned int i=0;i<nb_toSend_out;i++)
		{
		  T * ptr = data[indexes_out[i]];
		  //as we send to another processor that receive out, we give the node_access first values of ptr
		  toSend_out[i] = ptr[0];
		  }

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" to send out to proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toSend_out;i++)
		std::cout<<toSend_out[i]<<" ";
		std::cout<<std::endl;*/

	      comm->ISendOut(toSend_out, nb_toSend_out, j);

	      free(indexes_out);
	      free(toSend_out);
	    }

	  //prepare data to recv in
	  unsigned int nb_toRecv_in = ddag->getNbEdgesToRecvIn(j);
	  if(nb_toRecv_in>0)
	  comm->IRecvIn(nb_toRecv_in, j);
	}
    }
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    void end_communications()
    //-------------------------------------------------------------------------------
    {
      for(int j=0;j<Mpi_::mpi_nb;j++)
	{
	  //end data to send in
 	  unsigned int nb_toSend_in = ddag->getNbEdgesToSendIn(j);
	  if(nb_toSend_in>0)
	    comm->Wait_sIn(j);
	  
	  //end data to receive out
	  unsigned int nb_toRecv_out = ddag->getNbEdgesToRecvOut(j);
	  if(nb_toRecv_out>0)
	    {
	      T * toRecv_out = (T*)calloc(nb_toRecv_out,sizeof(T));
	      comm->Wait_rOut(toRecv_out,j,nb_toRecv_out);
	      unsigned int * indexes_out = ddag->getEdgesToRecvOut(j);
	      //change values in data
	      for(unsigned int k=0;k<nb_toRecv_out;k++)
		{
		  //pointer to modify
		  T * ptr = data[indexes_out[k]];
		  ptr[0] = toRecv_out[k];
		}

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" received out from proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toRecv_out;i++)
		std::cout<<toRecv_out[i]<<" ";
		std::cout<<std::endl;*/

	      free(indexes_out);
	      free(toRecv_out);
	    }
	  //end data to send out
 	  unsigned int nb_toSend_out = ddag->getNbEdgesToSendOut(j);
	  if(nb_toSend_out>0)
	    comm->Wait_sOut(j);
	  
	  //end data to receive in
	  unsigned int nb_toRecv_in = ddag->getNbEdgesToRecvIn(j);
	  if(nb_toRecv_in>0)
	    {
	      T * toRecv_in = (T*)calloc(nb_toRecv_in,sizeof(T));
	      comm->Wait_rIn(toRecv_in,j,nb_toRecv_in);
	      unsigned int * indexes_in = ddag->getEdgesToRecvIn(j);
	      //change values in data
	      for(int k=0;k<nb_toRecv_in;k++)
		{
		  //pointer to modify
		  T * ptr = data[indexes_in[k]];
		  ptr[0] = toRecv_in[k];
		}

	      /*std::cout<<"Proc "<<Mpi_::mpi_rank<<" received in from proc "<<j<<" = ";
	      for(unsigned int i=0;i<nb_toRecv_in;i++)
		std::cout<<toRecv_in[i]<<" ";
		std::cout<<std::endl;*/

	      free(indexes_in);
	      free(toRecv_in);
	    }
	}
    }
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  int N = sizes[i];
	  T * dat = data[i];
	  printf("Artery %i : ",ddag->getIndexEdge(i));
	  for(int i=0;i<N;i++)
	    printf("dat[%i] = %f,",i,dat[i]);
	  printf("\n");
	}
    }
    //-------------------------------------------------------------------------------
  };

  //-------------------------------------------------------------------------------
  //!  fourth specialization of DPMap_Edges_impl
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tapram overlap is the overlap size needed to compute the data
    \tparam size is the number of elements at each edge

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The overlap is specialized to the value 0
    The size is not specialized, it is the number of elements in the pointer table to map
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Edges_impl<DD,T*,0>
  {
  public :
    /*!< ddag is the pointer to the ddag data structure the property is mapped on */
    DDAG_impl<DD::nod,DD::edg> * ddag;
    /*!< size of the pointer mapped to edges */
    unsigned int * sizes;

    //-------------------------------------------------------------------------------
    //!  to read input initialize file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void read(const char * file)
    //-------------------------------------------------------------------------------
    {
      std::ifstream entries(file,std::ios::in);
      if (!entries){
	std::cerr << "Impossible to open the " << file <<"\n";
	exit(EXIT_FAILURE);
      }

      unsigned int nb_line = 0;
      while (!entries.eof()){
	//read line
	std::string line;
	getline (entries, line);
	for(unsigned int j=0;j<dim_edge;j++)
	  {
	    if(ddag->getIndexEdge(j)==nb_line)
	      {
		//init data table for this edge
		data[j] = (T*)calloc(sizes[j],sizeof(T));
		int found1 = -1;
		int found2 = -1;
		for(unsigned int i=0;i<sizes[j];i++)
		  {
		    found1 = found2+1;
		    found2 = line.find(",",found1);
		    std::string val = line.substr(found1,found2-found1);
		    data[j][i] = atof(val.data());
		    //const char * value = line.data();
		    //memcpy(&data[j][i], value, sizeof(T));
		  }
	      }
	  }
	nb_line++;
      }
      entries.close();
    }
    //-------------------------------------------------------------------------------
    //!  init function with values in file
    /*!
      \param file is the file of intial values for this property on edges (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    void init(const char * file)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      data = (T**)calloc(dim_edge,sizeof(T*));
      read(file);
    }
    //-------------------------------------------------------------------------------
    //!  init function with default value
    /*!
      \param val is the value to intiate edges to
    */
    //-------------------------------------------------------------------------------
    void init(T val)
    //-------------------------------------------------------------------------------
    {
      dim_edge = ddag->getDimEdge();
      data = (T**)calloc(dim_edge,sizeof(T*));
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  data[i] = (T*)calloc(sizes[i],sizeof(T));
	  for(unsigned int j=0;j<sizes[i];j++)
	    data[i][j] = val;
	}
      
    }
    //-------------------------------------------------------------------------------
    
  public:
    /*!< dim_edge is the number of local edges */
    unsigned int dim_edge;
    /*!< data mapped on edges of the ddag */
    T ** data;

    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, unsigned int * s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i]=s[ddag->getIndexEdge(i)];
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor
    /*!
      \param dag is the DDAG the property is going to mapped
      \param file is the file of intial values for this property on edges (each line is the value of the edge "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, const char * file, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      sizes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i] = s;
      //construction of data and read the file
      init(file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Edges
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of edges for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Edges_impl(DD &dag, T val, const unsigned int s)
    //-------------------------------------------------------------------------------
    {
      ddag = dag.getDDAG_impl();
      dim_edge = ddag->getDimEdge();
      sizes = (unsigned int *)calloc(dim_edge,sizeof(unsigned int));
      for(unsigned int i=0;i<dim_edge;i++)
	sizes[i] = s;
      ddag = dag.getDDAG_impl();
      //construction of data and fill with val
      init(val);
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Edges
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Edges_impl()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge;i++)
	free(data[i]);
      free(data);
      if(sizes!=NULL)
	free(sizes);
    }
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on edges
    /*!
      \return the begin iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return iterator_dag(0);}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on edges
    /*!
      \return the end iterator on edges
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return iterator_dag(ddag->getDimEdge());}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the source node of the edge
    /*!
      \return the iterator on the source node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getSrcNode(iterator_dag it){return iterator_dag(ddag->getSrcNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  to get an iterator on the destination node of the edge
    /*!
      \return the iterator on the destination node
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag getDstNode(iterator_dag it){return iterator_dag(ddag->getDstNode(it._rank));}
    //-------------------------------------------------------------------------------
    //!  get the size of the table at iterator it
    /*!
      \param it iterator position to get the size
      \return the size
    */
    //-------------------------------------------------------------------------------
    inline unsigned int getDataSize(iterator_dag it){return sizes[it._rank];}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_edges property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    inline void start_communications(){}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_edges property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_edges property
    */
    //-------------------------------------------------------------------------------
    inline void end_communications(){}
    //-------------------------------------------------------------------------------
    //!  print values of the dpmap, edge after edge using the global indexation
    /*!
    */
    //-------------------------------------------------------------------------------
    void print()
    //-------------------------------------------------------------------------------
    {
      for(unsigned int i=0;i<dim_edge;i++)
	{
	  int N = sizes[i];
	  T * dat = data[i];
	  printf("Artery %i : ",ddag->getIndexEdge(i));
	  for(int i=0;i<N;i++)
	    printf("dat[%i] = %f,",i,dat[i]);
	  printf("\n");
	}
    }
    //-------------------------------------------------------------------------------
  };
  
};


#endif
