/*! \file pmap_nodes.hpp
 *  \brief Definitions of the object PMap_Nodes used by the user to map data to the nodes of a DDAG. Three template specializations are available.
 */
#ifndef DPMAP_NODES_H
#define DPMAP_NODES_H

#include "dpmap_nodes_impl.hpp"

namespace skelgis{

  //-------------------------------------------------------------------------------
  //!  base template class of Property map on nodes
  /*!
    This template class defines the base PMap_Nodes
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam node_access indicates if nodes have to have access to this property during calculation (>=1) or not (=0)
    \tparam size is the number of elements at each node

    In the case of a basic type T, node_access can't be >1
    In the case of a pointer type T*, if node_access>0 then it indicates the number of elements to shared at the end and at the beginning of the pointer
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T, int node_access> struct DPMap_Nodes;
  //-------------------------------------------------------------------------------
  //!  first specialization of DPMap_Nodes
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The node_access is specialized to the value 1
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Nodes<DD,T,1>
  //-------------------------------------------------------------------------------
  {
  protected :
    bool ptr_const;
    DPMap_Nodes_impl<DD,T,1> * dpmap;

  public:

    typedef DPMap_Nodes_impl<DD,T,1> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of nodes wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DD &dag, const char * file)
    {
      ptr_const = false;
      dpmap = new DPMap_Nodes_impl<DD,T,1>(dag,file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of nodes for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DD &dag, T val)
    {
      ptr_const = false;
      dpmap = new DPMap_Nodes_impl<DD,T,1>(dag,val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Nodes_impl pointer
    /*!
      \param pointer to a DPMap_Nodes_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DPMap_Nodes_impl<DD,T,1> * p)
    {
      ptr_const = true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Nodes
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Nodes()
    {
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Nodes_impl<DD,T,1> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on roots nodes
    /*!
      \return the begin iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_roots(){return dpmap->begin_roots();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on roots nodes
    /*!
      \return the end iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_roots(){return dpmap->end_roots();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on leafs nodes
    /*!
      \return the begin iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_leafs(){return dpmap->begin_leafs();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on leafs nodes
    /*!
      \return the end iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_leafs(){return dpmap->end_leafs();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes with communication with other processors
    /*!
      \return the begin iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_comm(){return dpmap->begin_comm();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes with communication with other processors
    /*!
      \return the end iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_comm(){return dpmap->end_comm();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes without communication with other processors
    /*!
      \return the begin iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_loc(){return dpmap->begin_loc();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes without communication with other processors
    /*!
      \return the end iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_loc(){return dpmap->end_loc();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the begin iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin_in();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the end iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end_in();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input edges for the node it
    /*!
      \param it is the iterator on the node to get input edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInEdges(iterator_dag it){return dpmap->getInEdges(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output edges for the node it
    /*!
      \param it is the iterator on the node to get output edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutEdges(iterator_dag it){return dpmap->getOutEdges(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input nodes for the node it
    /*!
      \param it is the iterator on the node to get input nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInNodes(iterator_dag it){return dpmap->getInNodes(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output nodes for the node it
    /*!
      \param it is the iterator on the node to get output nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutNodes(iterator_dag it){return dpmap->getOutNodes(it);}
    //-------------------------------------------------------------------------------
    //!  start the communication process for the dpmap_nodes property
    /*!
      prepare data to send
      launch non blocking MPI communications
    */
    //-------------------------------------------------------------------------------
    inline void start_communications(){dpmap->start_communications();}
    //-------------------------------------------------------------------------------
    //!  end the communication process for the dpmap_nodes property
    /*!
      wait for the end of non-blocking communications
      put received values in the dpmap_nodes property
    */
    //-------------------------------------------------------------------------------
    inline void end_communications(){dpmap->end_communications();}
    //-------------------------------------------------------------------------------
    
  };
  //-------------------------------------------------------------------------------
  //!  second specialization of DPMap_Nodes
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is a simple type in this case as float, double, int etc.
    The node_access is specialized to the value 0
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Nodes<DD,T,0>
  {
    protected :
    bool ptr_const;
    DPMap_Nodes_impl<DD,T,0> * dpmap;

  public:

    typedef DPMap_Nodes_impl<DD,T,0> impl_type;

    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param file is the file with initial values of nodes wanted (each line is the value of the node "number of this line" associated in the .dot file)
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DD &dag, const char * file)
    {
      ptr_const = false;
      dpmap = new DPMap_Nodes_impl<DD,T,0>(dag,file);
    }
    //-------------------------------------------------------------------------------
    //!  constructor of DPMap_Nodes
    /*!
      \param dag is the DDAG object associated to this map
      \param val is the default value of nodes for the map
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DD &dag, T val)
    {
      ptr_const = false;
      dpmap = new DPMap_Nodes_impl<DD,T,0>(dag,val);
    }
    //-------------------------------------------------------------------------------
    //!  constructor from a DPMap_Nodes_impl pointer
    /*!
      \param pointer to a DPMap_Nodes_impl
    */
    //-------------------------------------------------------------------------------
    DPMap_Nodes(DPMap_Nodes_impl<DD,T,0> * p)
    {
      ptr_const = true;
      dpmap = p;
    }
    //-------------------------------------------------------------------------------
    //!  destructor of DPMap_Nodes
    /*!
    */
    //-------------------------------------------------------------------------------
    ~DPMap_Nodes()
    {
      if(dpmap!=NULL && !ptr_const)
	{delete dpmap;dpmap=NULL;}
    };
    //-------------------------------------------------------------------------------
    //!  to get the pointer on dpmap_impl
    /*!
      \return the dpmap implementation pointer
    */
    //-------------------------------------------------------------------------------
    inline DPMap_Nodes_impl<DD,T,0> * getDPMap(){return dpmap;}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on roots nodes
    /*!
      \return the begin iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_roots(){return dpmap->begin_roots();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on roots nodes
    /*!
      \return the end iterator on roots nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_roots(){return dpmap->end_roots();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on leafs nodes
    /*!
      \return the begin iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_leafs(){return dpmap->begin_leafs();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on leafs nodes
    /*!
      \return the end iterator on leafs nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_leafs(){return dpmap->end_leafs();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes with communication with other processors
    /*!
      \return the begin iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_comm(){return dpmap->begin_comm();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes with communication with other processors
    /*!
      \return the end iterator on communication nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_comm(){return dpmap->end_comm();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on nodes without communication with other processors
    /*!
      \return the begin iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin_loc(){return dpmap->begin_loc();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on nodes without communication with other processors
    /*!
      \return the end iterator on local nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end_loc(){return dpmap->end_loc();}
    //-------------------------------------------------------------------------------
    //!  to get the begin iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the begin iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag begin(){return dpmap->begin_in();}
    //-------------------------------------------------------------------------------
    //!  to get the end iterator on inside nodes (i.e. nodes that are not roots or leafs = loc and comm nodes)
    /*!
      \return the end iterator on inside nodes
    */
    //-------------------------------------------------------------------------------
    inline iterator_dag end(){return dpmap->end_in();}
    //-------------------------------------------------------------------------------
    //!  to get the value of the property map at the iterator position
    /*!
      \return a reference on the value (to be modified)
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator_dag it) { return dpmap->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input edges for the node it
    /*!
      \param it is the iterator on the node to get input edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInEdges(iterator_dag it){return dpmap->getInEdges(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output edges for the node it
    /*!
      \param it is the iterator on the node to get output edges
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutEdges(iterator_dag it){return dpmap->getOutEdges(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on input nodes for the node it
    /*!
      \param it is the iterator on the node to get input nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getInNodes(iterator_dag it){return dpmap->getInNodes(it);}
    //-------------------------------------------------------------------------------
    //!  to get a vector of iterator on output nodes for the node it
    /*!
      \param it is the iterator on the node to get output nodes
      \return a vector of iterator
    */
    //-------------------------------------------------------------------------------
    inline std::vector<iterator_dag> getOutNodes(iterator_dag it){return dpmap->getOutNodes(it);}
    //-------------------------------------------------------------------------------

  };
  //-------------------------------------------------------------------------------
  //!  third specialization of DPMap_Nodes
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map
    \tparam node_access indicates if nodes have to have access to this property during calculation (>=1) or not (=0)

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The node_access is not specialized and is [0,...,size], with size a constant value given at the construction
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T,int node_access> struct DPMap_Nodes<DD,T*,node_access>
  {
  protected :

  public:

    typedef DPMap_Nodes_impl<DD,T*,node_access> impl_type;

    DPMap_Nodes(){};
    ~DPMap_Nodes(){};
    
  };
  //-------------------------------------------------------------------------------
  //!  fourth specialization of DPMap_Nodes
  /*!
    \tparam DD is the type of the DDAG
    \tparam T is the type of data in the property map

    The type T is specialized with a specific pointer type in this case as float*, double*, int* etc.
    The node_access is specialized to the value 0
  */
  //-------------------------------------------------------------------------------
  template<class DD, class T> struct DPMap_Nodes<DD,T*,0>
  {
  protected :

  public:
    DPMap_Nodes(){};
    ~DPMap_Nodes(){};

    typedef DPMap_Nodes_impl<DD,T*,0> impl_type;
  };

};


#endif
