/*! \file dmatrix.hpp
 *  \brief Definitions of the object DMatrix used by the user. Four template specializations are available
 */
#ifndef DMATRIX_H
#define DMATRIX_H

#include "dmatrix_impl.hpp"

namespace skelgis{

  //================================================================================
  //! DMatrix class
  /*!
    template of the distributed matrix of SkelGIS. This is the first specialization of the DMatrix.
    \tparam T is the type of data to store in the DMatrix
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,bool line=true> struct DMatrix
  //-------------------------------------------------------------------------------
  {
  protected:
    //! Pointer to DMatrix_impl object
    /*!
      The pointer to Dmatrix_impl object is the real distributed matrix object. DMatrix class is the user interface to this object.
    */
    DMatrix_impl<T,R,line> * m;

    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRight(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank+1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank-1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r--;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width -1 ;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width -1 ;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width - 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width - 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    T right(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + 1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T left(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - 1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T up(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T down(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T rightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width +1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T leftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width -1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T rightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width +1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T leftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width -1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allX(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[b];
	  b++;
	}
      b++;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = m->data[b];
	  b++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allX(unsigned int rank, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
        {
          t[i] = m->data[b];
          b++;
        }
      b++;
      for(int i=R;i<2*R;i++)
        {
          t[i] = m->data[b];
          b++;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allY(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[b];
	  b = b + width;
	}
      b = b + width;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = m->data[b];
	  b = b + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allY(unsigned int rank,unsigned int width, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
        {
          t[i] = m->data[b];
          b = b + width;
        }
      b = b + width;
      for(int i=R;i<2*R;i++)
        {
          t[i] = m->data[b];
          b = b + width;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to leftUp, up, rightUp, right, rightDown, down, leftDown and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * eight(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = leftUp(rank,width);
      res[1] = up(rank,width);
      res[2] = rightUp(rank,width);
      res[3] = right(rank);
      res[4] = rightDown(rank,width);
      res[5] = down(rank,width);
      res[6] = leftDown(rank,width);
      res[7] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to up, right, down and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * four(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = up(rank,width);
      res[1] = right(rank);
      res[2] = down(rank,width);
      res[3] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inright(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + R;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inleft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - R;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inup(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - R*width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T indown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + R*width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from pointer
    /*!
      This Constructor is used in the user function when a skeleton ApplyList is called.
      \param _m is a DMatrix_impl pointer
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix_impl<T,R,line> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface with another DMatrix
    /*!
      It is possible to construct a DMatrix instance from another DMatrix instance
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix<T,R,line>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a HEADER and a default value
    /*!
      The HEADER essentially define the size of the DMatrix, while the default value is given to each element of the DMatrix.
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,R,line>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a binary file
    /*!
      A specific binary file format is read by SkelGIS. A Dmatrix instance can be constructed from a binary file.
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(const char * binFile, bool loc=false){m = new DMatrix_impl<T,R,line>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix(){}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from pointer
    /*!
      The equivalent to the constructor from a pointer
      \param _m is a DMatrix_impl object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix_impl<T,R,line> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from another DMatrix
    /*!
      The equivalent to the constructor from another DMatrix
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix<T,R,line>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a HEADER and a default value
    /*!
      The equivalent to the constructor from a HEADER and a default value
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,R,line>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a binary file
    /*!
      The equivalent to the constructor from a binary file
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(const char * binFile, bool loc=false){m = new DMatrix_impl<T,R,line>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! to erase and clear the DMatrix
    /*!
      The equivalent to the erase of a std::vector in the STL.
      The erase will empty the DMatrix and will deallocate the memory space associated.
      This has to be done when the DMatrix is not usefull anymore.
    */
    //-------------------------------------------------------------------------------
    inline void erase(){delete m;}
    //-------------------------------------------------------------------------------
    //! Methods to get the value of the matrix at the iterator it (alternative to [it])
    /*!
      This method exists for each type of iterator in SkelGIS. This method is const as the value returned could not be modified.
      \param it is the iterator position
      \return the value at the iterator it
    */
    //-------------------------------------------------------------------------------
    inline T getValue(iterator<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_cont<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_step<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_rev<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_left<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_right<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_up<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_down<T,R> it) const {return m->data[it._rank];}
    //-------------------------------------------------------------------------------
    //! Methods to set the value of the matrix at the iterator it (alternative to [it]=)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \param value is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setValue(iterator<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_cont<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_step<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_rev<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_left<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_right<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_up<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_down<T,R> it,T value){m->data[it._rank] = value;}
    //-------------------------------------------------------------------------------
    //! Set value in the middle of the global matrix
    /*!
      \param val is the value to affect
    */
    //-------------------------------------------------------------------------------
    inline void setGlobalMiddleValue(T val){m->setGlobalMiddleValue(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values of the physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setPhysicalBorder(T val){m->setPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the right of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setRightPhysicalBorder(T val){m->setRightPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the left of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setLeftPhysicalBorder(T val){m->setLeftPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the up of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setUpPhysicalBorder(T val){m->setUpPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the down of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setDownPhysicalBorder(T val){m->setDownPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> begin(){return m->begin();}
    inline iterator_cont<T,R> begin_cont(){return m->begin_cont();}
    inline iterator_rev<T,R> begin_rev(){return m->begin_rev();}
    inline iterator_step<T,R> begin_step(int step,int nb){return m->begin_step(step,nb);}
    inline iterator_phb_left<T,R> begin_phb_left(){return m->begin_phb_left();}
    inline iterator_phb_right<T,R> begin_phb_right(){return m->begin_phb_right();}
    inline iterator_phb_up<T,R> begin_phb_up(){return m->begin_phb_up();}
    inline iterator_phb_down<T,R> begin_phb_down(){return m->begin_phb_down();}
    inline iterator_line<T,R> begin_line(){return m->begin_line();}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> end(){return m->end();}
    inline iterator_cont<T,R> end_cont(){return m->end_cont();}
    inline iterator_rev<T,R> end_rev(){return m->end_rev();}
    inline iterator_step<T,R> end_step(){return m->end_step();}
    inline iterator_phb_left<T,R> end_phb_left(){return m->end_phb_left();}
    inline iterator_phb_right<T,R> end_phb_right(){return m->end_phb_right();}
    inline iterator_phb_up<T,R> end_phb_up(){return m->end_phb_up();}
    inline iterator_phb_down<T,R> end_phb_down(){return m->end_phb_down();}
    inline iterator_line<T,R> end_line(){return m->end_line();}
    //-------------------------------------------------------------------------------
    //! Operator [] to get and set an element of the matrix at iterator it
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \return the reference to the value at it
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_cont<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_rev<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_step<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_left<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_right<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_up<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_down<T,R> it) { return m->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //! Operator = to assign the current DMatrix interface object to another one
    /*!
      \param right is the other DMatrix object
      \return the reference to the new current DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline DMatrix<T,R,line>& operator=(DMatrix<T,R,line> & right){m = right.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! Print the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs".
     */
    //-------------------------------------------------------------------------------
    inline void print(){m->print();}
    //-------------------------------------------------------------------------------
    //! Print all the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs". 
      It will also print the physical border of the DMatrix and the parallel overlap of the DMatrix.
     */
    //-------------------------------------------------------------------------------
    inline void printAll(){m->printAll();}
    //-------------------------------------------------------------------------------
    //! Saving in a bin file the distributed matrix interface
    /*!
      This method writes the DMatrix in a single binary output file.
      \param binFile is a binary output file to save the DMatrix in
    */
    //-------------------------------------------------------------------------------
    inline void write(char * binFile){m->write(binFile);}
    //-------------------------------------------------------------------------------
    //! get borders
    /*!
      To get the needed parallel overlap from other processors.
      MPI exchanges done.
    */
    //-------------------------------------------------------------------------------
    inline void getBorders(){m->getBorders();}
    //-------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRight(iterator<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_cont<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_step<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_rev<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return allRight(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeft(iterator<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_cont<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_step<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_rev<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return allLeft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllUp(iterator<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_cont<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_step<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_rev<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return allUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllDown(iterator<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_cont<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_step<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_rev<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return allDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightDown(iterator<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_step<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return allRightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftDown(iterator<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return allLeftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightUp(iterator<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_step<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return allRightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftUp(iterator<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return allLeftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRight(iterator<T,R> it){return right(it._rank);}
    inline T getRight(iterator_cont<T,R> it){return right(it._rank);}
    inline T getRight(iterator_step<T,R> it){return right(it._rank);}
    inline T getRight(iterator_rev<T,R> it){return right(it._rank);}
    inline T getRight(iterator_phb_left<T,R> it){return right(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeft(iterator<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_cont<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_step<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_rev<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_phb_right<T,R> it){return left(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getUp(iterator<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_cont<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_step<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_rev<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_phb_down<T,R> it){return up(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getDown(iterator<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_cont<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_step<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_rev<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_phb_up<T,R> it){return down(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightDown(iterator<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_cont<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_step<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_rev<T,R> it){return rightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftDown(iterator<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_cont<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_step<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_rev<T,R> it){return leftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightUp(iterator<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_cont<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_step<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_rev<T,R> it){return rightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftUp(iterator<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_cont<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_step<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_rev<T,R> it){return leftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline T * getAllX(iterator<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_cont<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_step<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_rev<T,R> it){return allX(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline void getAllX(iterator<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_cont<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_rev<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_step<T,R> it, T * t){allX(it._rank,t);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline T * getAllY(iterator<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_cont<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_step<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_rev<T,R> it){return allY(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline void getAllY(iterator<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_cont<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_rev<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_step<T,R> it, T * t){allY(it._rank,it._width,t);}
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get8(iterator<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_cont<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_step<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_rev<T,R> it){return eight(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get4(iterator<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_cont<T,R> it){return four(it._rank,it._width);}
     inline T * get4(iterator_step<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_rev<T,R> it){return four(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInRight(iterator_phb_left<T,R> it){return inright(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInLeft(iterator_phb_right<T,R> it){return inleft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInUp(iterator_phb_down<T,R> it){return inup(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInDown(iterator_phb_up<T,R> it){return indown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRight(iterator<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_cont<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_rev<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_step<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return m->getAllRight(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeft(iterator<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_cont<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_rev<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_step<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return m->getAllLeft(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllUp(iterator<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_cont<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_rev<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_step<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return m->getAllUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllDown(iterator<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_cont<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_rev<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_step<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return m->getAllDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRightDown(iterator<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_step<T,R> it){return m->getAllRightDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeftDown(iterator<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return m->getAllLeftDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRightUp(iterator<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_step<T,R> it){return m->getAllRightUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeftUp(iterator<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return m->getAllLeftUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRight(iterator<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_cont<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_rev<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_step<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_phb_left<T,R> it){return m->getRight(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeft(iterator<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_cont<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_rev<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_step<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_phb_right<T,R> it){return m->getLeft(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getUp(iterator<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_cont<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_rev<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_step<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_phb_down<T,R> it){return m->getUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getDown(iterator<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_cont<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_rev<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_step<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_phb_up<T,R> it){return m->getDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRightDown(iterator<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_cont<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_rev<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_step<T,R> it){return m->getRightDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeftDown(iterator<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_cont<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_rev<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_step<T,R> it){return m->getLeftDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRightUp(iterator<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_cont<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_rev<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_step<T,R> it){return m->getRightUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeftUp(iterator<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_cont<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_rev<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_step<T,R> it){return m->getLeftUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllX(iterator<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_cont<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_rev<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_step<T,R> it){return m->getAllX(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    /*inline void getAllX(iterator<T,R> it, T * t){return m->getAllX(it,t);}
    inline void getAllX(iterator_cont<T,R> it, T * t){return m->getAllX(it,t);}
    inline void getAllX(iterator_rev<T,R> it, T * t){return m->getAllX(it,t);}
    inline void getAllX(iterator_step<T,R> it, T * t){return m->getAllX(it,t);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllY(iterator<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_cont<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_rev<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_step<T,R> it){return m->getAllY(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    /*inline void getAllY(iterator<T,R> it, T * t){return m->getAllY(it,t);}
    inline void getAllY(iterator_cont<T,R> it, T * t){return m->getAllY(it,t);}
    inline void getAllY(iterator_rev<T,R> it, T * t){return m->getAllY(it,t);}
    inline void getAllY(iterator_step<T,R> it, T * t){return m->getAllY(it,t);}*/
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    /*inline T * get8(iterator<T,R> it){return m->get8(it);}
    inline T * get8(iterator_cont<T,R> it){return m->get8(it);}
    inline T * get8(iterator_rev<T,R> it){return m->get8(it);}
    inline T * get8(iterator_step<T,R> it){return m->get8(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    /*inline T * get4(iterator<T,R> it){return m->get4(it);}
    inline T * get4(iterator_cont<T,R> it){return m->get4(it);}
    inline T * get4(iterator_rev<T,R> it){return m->get4(it);}
    inline T * get4(iterator_step<T,R> it){return m->get4(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInRight(iterator_phb_left<T,R> it){return m->getInRight(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInLeft(iterator_phb_right<T,R> it){return m->getInLeft(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInUp(iterator_phb_down<T,R> it){return m->getInUp(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInDown(iterator_phb_up<T,R> it){return m->getInDown(it);}
    //-------------------------------------------------------------------------------
    //! Get the iterator on the matrix at position (x,y)
    /*!
      This method exists for the three basic types of iterators.
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> getIterator(int x,int y){return m->getIterator(x,y);}
    inline iterator<T,R> getIterator_cont(int x,int y){return m->getIterator_cont(x,y);}
    inline iterator<T,R> getIterator_rev(int x,int y){return m->getIterator_rev(x,y);}
    //-------------------------------------------------------------------------------
    //! Get the indexes of the iterator on the matrix
    /*!
      This method exists for the three basic types of iterators.
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    //-------------------------------------------------------------------------------
    inline void getIndexes(iterator<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_cont<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_rev<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    //-------------------------------------------------------------------------------
    //! Get the global header of the DMatrix
    /*!
      \return the global header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getGlobalHeader(){return m->head;}
    //-------------------------------------------------------------------------------
    //! Get the header of the DMatrix
    /*!
      \return the header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getHeader(){return m->loc_head;}
    //-------------------------------------------------------------------------------
    //! Get the border header of the DMatrix
    /*!
      \return the border header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getBorderHeader(){return m->border_head;}
    //-------------------------------------------------------------------------------
    //! Get the number of columns in the domain decomposition
    /*!
      \return the number of columns in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getCols(){return m->cols;}
    //-------------------------------------------------------------------------------
    //! Get the number of rows in the domain decomposition
    /*!
      \return the number of rows in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getRows(){return m->rows;}
    //-------------------------------------------------------------------------------
    //! Get the column number of the DMatrix
    /*!
      \return the column number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getCol(){return m->col;}
    //-------------------------------------------------------------------------------
    //! Get the row number of the DMatrix
    /*!
      \return the row number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getRow(){return m->row;}
    //-------------------------------------------------------------------------------
    //! Get the data table of the DMatrix
    /*!
      \return the data table of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline T * getData(){return m->data;}
    //-------------------------------------------------------------------------------
    //! Get the recovery of the DMatrix (=R)
    /*!
      \return the recovery of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getBorder(){return R;}
    //-------------------------------------------------------------------------------
    //! Get the DMatrix_impl associated to the DMatrix interface
    /*!
      \return the DMatrix_impl
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,R,line> * getDMatrix(){return m;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //! DMatrix class
  /*!
    template of the distributed matrix of SkelGIS. This is the second specialization of the DMatrix.
    \tparam T is the type of data to store in the DMatrix
    R overlap template parameter is not asked and specialized to 0.
    The default value of line is used (true).
  */
  //-------------------------------------------------------------------------------
  template<class T> struct DMatrix<T,0>
  //-------------------------------------------------------------------------------
  {
  protected:
    //! Pointer to DMatrix_impl object
    /*!
      The pointer to Dmatrix_impl object is the real distributed matrix object. DMatrix class is the user interface to this object.
    */
    DMatrix_impl<T,0> * m;

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from pointer
    /*!
      This Constructor is used in the user function when a skeleton ApplyList is called.
      \param _m is a DMatrix_impl pointer
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix_impl<T,0> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface with another DMatrix
    /*!
      It is possible to construct a DMatrix instance from another DMatrix instance
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix<T,0>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a HEADER and a default value
    /*!
      The HEADER essentially define the size of the DMatrix, while the default value is given to each element of the DMatrix.
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,0>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a binary file
    /*!
      A specific binary file format is read by SkelGIS. A Dmatrix instance can be constructed from a binary file.
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(const char * binFile, bool loc=false){m = new DMatrix_impl<T,0>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix(){}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from pointer
    /*!
      The equivalent to the constructor from a pointer
      \param _m is a DMatrix_impl object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix_impl<T,0> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from another DMatrix
    /*!
      The equivalent to the constructor from another DMatrix
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix<T,0>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a HEADER and a default value
    /*!
      The equivalent to the constructor from a HEADER and a default value
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,0>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a binary file
    /*!
      The equivalent to the constructor from a binary file
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(const char * binFile, bool loc=false){m = new DMatrix_impl<T,0>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! to erase and clear the DMatrix
    /*!
      The equivalent to the erase of a std::vector in the STL.
      The erase will empty the DMatrix and will deallocate the memory space associated.
      This has to be done when the DMatrix is not usefull anymore.
    */
    //-------------------------------------------------------------------------------
    inline void erase(){delete m;}
    //-------------------------------------------------------------------------------
    //! Methods to get the value of the matrix at the iterator it (alternative to [it])
    /*!
      This method exists for each type of iterator in SkelGIS. This method is const as the value returned could not be modified.
      \param it is the iterator position
      \return the value at the iterator it
    */
    //-------------------------------------------------------------------------------
    inline T getValue(iterator<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_cont<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_rev<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_step<T,0> it) const {return m->data[it._rank];}
    //-------------------------------------------------------------------------------
    //! Methods to set the value of the matrix at the iterator it (alternative to [it]=)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \param value is the value to set
    */
    //-------------------------------------------------------------------------------
    inline T setValue(iterator<T,0> it,T value){m->data[it._rank] = value;}
    inline T setValue(iterator_cont<T,0> it,T value){m->data[it._rank] = value;}
    inline T setValue(iterator_rev<T,0> it,T value){m->data[it._rank] = value;}
    inline T setValue(iterator_step<T,0> it,T value){m->data[it._rank] = value;}
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> begin(){return m->begin();}
    inline iterator_cont<T,0> begin_cont(){return m->begin_cont();}
    inline iterator_rev<T,0> begin_rev(){return m->begin_rev();}
    inline iterator_step<T,0> begin_step(int step,int nb){return m->begin_step(step,nb);}
    inline iterator_line<T,0> begin_line(){return m->begin_line();}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> end(){return m->end();}
    inline iterator_cont<T,0> end_cont(){return m->end_cont();}
    inline iterator_rev<T,0> end_rev(){return m->end_rev();}
    inline iterator_step<T,0> end_step(){return m->end_step();}
    inline iterator_line<T,0> end_line(){return m->end_line();}
    //-------------------------------------------------------------------------------
    //! Operator [] to get and set an element of the matrix at iterator it
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \return the reference to the value at it
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_cont<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_rev<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_step<T,0> it) { return m->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //! Operator = to assign the current DMatrix interface object to another one
    /*!
      \param right is the other DMatrix object
      \return the reference to the new current DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline DMatrix<T,0,true>& operator=(DMatrix<T,0,true> & right){ m = right.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! Print the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs".
     */
    //-------------------------------------------------------------------------------
    inline void print(){m->print();}
    //-------------------------------------------------------------------------------
    //! Print all the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs". 
      It will also print the physical border of the DMatrix and the parallel overlap of the DMatrix.
     */
    //-------------------------------------------------------------------------------
    inline void printAll(){m->printAll();}
    //-------------------------------------------------------------------------------
    //! Saving in a bin file the distributed matrix interface
    /*!
      This method writes the DMatrix in a single binary output file.
      \param binFile is a binary output file to save the DMatrix in
    */
    //-------------------------------------------------------------------------------
    inline void write(char * binFile){m->write(binFile);}
    //-------------------------------------------------------------------------------
    //! Get the iterator on the matrix at position (x,y)
    /*!
      This method exists for the three basic types of iterators.
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> getIterator(int x,int y){return m->getIterator(x,y);}
    inline iterator<T,0> getIterator_cont(int x,int y){return m->getIterator_cont(x,y);}
    inline iterator<T,0> getIterator_rev(int x,int y){return m->getIterator_rev(x,y);}
    //-------------------------------------------------------------------------------
    //! Get the indexes of the iterator on the matrix
    /*!
      This method exists for the three basic types of iterators.
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    //-------------------------------------------------------------------------------
    inline void getIndexes(iterator<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_cont<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_rev<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    //-------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------
    //! Get the global header of the DMatrix
    /*!
      \return the global header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getGlobalHeader(){return m->head;}
    //-------------------------------------------------------------------------------
    //! Get the header of the DMatrix
    /*!
      \return the header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getHeader(){return m->loc_head;}
    //-------------------------------------------------------------------------------
    //! Get the number of columns in the domain decomposition
    /*!
      \return the number of columns in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getCols(){return m->cols;}
    //-------------------------------------------------------------------------------
    //! Get the number of rows in the domain decomposition
    /*!
      \return the number of rows in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getRows(){return m->rows;}
    //-------------------------------------------------------------------------------
    //! Get the column number of the DMatrix
    /*!
      \return the column number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getCol(){return m->col;}
    //-------------------------------------------------------------------------------
    //! Get the row number of the DMatrix
    /*!
      \return the row number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getRow(){return m->row;}
    //-------------------------------------------------------------------------------
    //! Get the data table of the DMatrix
    /*!
      \return the data table of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline T * getData(){return m->data;}
    //-------------------------------------------------------------------------------
    //! Get the DMatrix_impl associated to the DMatrix interface
    /*!
      \return the DMatrix_impl
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,0,true> * getDMatrix(){return m;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //! DMatrix class
  /*!
    template of the distributed matrix of SkelGIS. This is the third specialization of the DMatrix.
    \tparam T is the type of data to store in the DMatrix
    \tparam R is the overlap distance needed by the calculation, in other words, the size of the physical border needed
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct DMatrix<T,R,false>
  //-------------------------------------------------------------------------------
  {
  protected:
    //! Pointer to DMatrix_impl object
    /*!
      The pointer to Dmatrix_impl object is the real distributed matrix object. DMatrix class is the user interface to this object.
    */
    DMatrix_impl<T,R,false> * m;

    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRight(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank+1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank-1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r--;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank + width -1 ;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r + width -1 ;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allRightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width + 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width + 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor values are wanted
      \return table of neighbor values
    */
    //-------------------------------------------------------------------------------
    T * allLeftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R];
      unsigned int r = rank - width - 1;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[r];
	  r = r - width - 1;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T right(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + 1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T left(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - 1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T up(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width;
      return m->data[r];
    }//-------------------------------------------------------------------------------
    //! Get the nearest neighbor down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T down(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T rightDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width +1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left down value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T leftDown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + width -1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor right up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T rightUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width +1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest neighbor left up value for the element to a precise rank
    /*!
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T leftUp(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - width -1;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allX(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[b];
	  b++;
	}
      b++;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = m->data[b];
	  b++;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element to a precise rank
    /*!
      Equivalent to allLeft and allRight with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allX(unsigned int rank, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R;
      for(int i=0;i<R;i++)
        {
          t[i] = m->data[b];
          b++;
        }
      b++;
      for(int i=R;i<2*R;i++)
        {
          t[i] = m->data[b];
          b++;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    T * allY(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[R*2];
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
	{
	  res[i] = m->data[b];
	  b = b + width;
	}
      b = b + width;
      for(int i=R;i<2*R;i++)
	{
	  res[i] = m->data[b];
	  b = b + width;
	}
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element to a precise rank
    /*!
      Equivalent to allDown and allUp with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    void allY(unsigned int rank,unsigned int width, T * t)
    //-------------------------------------------------------------------------------
    {
      unsigned int b = rank - R*width;

      for(int i=0;i<R;i++)
        {
          t[i] = m->data[b];
          b = b + width;
        }
      b = b + width;
      for(int i=R;i<2*R;i++)
        {
          t[i] = m->data[b];
          b = b + width;
        }
    }
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to leftUp, up, rightUp, right, rightDown, down, leftDown and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * eight(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = leftUp(rank,width);
      res[1] = up(rank,width);
      res[2] = rightUp(rank,width);
      res[3] = right(rank);
      res[4] = rightDown(rank,width);
      res[5] = down(rank,width);
      res[6] = leftDown(rank,width);
      res[7] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element to a precise rank
    /*!
      Equivalent to up, right, down and left with a single call.
      \param rank is the rank at which the neighbor value is wanted
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    T * four(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      T * res = new T[8];
      res[0] = up(rank,width);
      res[1] = right(rank);
      res[2] = down(rank,width);
      res[3] = left(rank);
      return res;
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inright(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + R;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inleft(unsigned int rank)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - R;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T inup(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank - R*width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain to a precise rank
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param rank is the rank at which the neighbor value is wanted
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T indown(unsigned int rank,unsigned int width)
    //-------------------------------------------------------------------------------
    {
      unsigned int r = rank + R*width;
      return m->data[r];
    }
    //-------------------------------------------------------------------------------

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from pointer
    /*!
      This Constructor is used in the user function when a skeleton ApplyList is called.
      \param _m is a DMatrix_impl pointer
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix_impl<T,R,false> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface with another DMatrix
    /*!
      It is possible to construct a DMatrix instance from another DMatrix instance
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix<T,R,false>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a HEADER and a default value
    /*!
      The HEADER essentially define the size of the DMatrix, while the default value is given to each element of the DMatrix.
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,R,false>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a binary file
    /*!
      A specific binary file format is read by SkelGIS. A Dmatrix instance can be constructed from a binary file.
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(const char * binFile, bool loc=false){m = new DMatrix_impl<T,R,false>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix(){}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from pointer
    /*!
      The equivalent to the constructor from a pointer
      \param _m is a DMatrix_impl object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix_impl<T,R,false> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from another DMatrix
    /*!
      The equivalent to the constructor from another DMatrix
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix<T,R,false>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a HEADER and a default value
    /*!
      The equivalent to the constructor from a HEADER and a default value
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(HEADER h,const T value, bool loc=false){ m = new DMatrix_impl<T,R,false>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a binary file
    /*!
      The equivalent to the constructor from a binary file
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    inline void init(const char * binFile, bool loc=false){m = new DMatrix_impl<T,R,false>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! to erase and clear the DMatrix
    /*!
      The equivalent to the erase of a std::vector in the STL.
      The erase will empty the DMatrix and will deallocate the memory space associated.
      This has to be done when the DMatrix is not usefull anymore.
    */
    //-------------------------------------------------------------------------------
    inline void erase(){delete m;}
    //-------------------------------------------------------------------------------
    //! Methods to get the value of the matrix at the iterator it (alternative to [it])
    /*!
      This method exists for each type of iterator in SkelGIS. This method is const as the value returned could not be modified.
      \param it is the iterator position
      \return the value at the iterator it
    */
    //-------------------------------------------------------------------------------
    inline T getValue(iterator<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_cont<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_step<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_rev<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_left<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_right<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_up<T,R> it) const {return m->data[it._rank];}
    inline T getValue(iterator_phb_down<T,R> it) const {return m->data[it._rank];}
    //-------------------------------------------------------------------------------
    //! Methods to set the value of the matrix at the iterator it (alternative to [it]=)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \param value is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setValue(iterator<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_cont<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_step<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_rev<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_left<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_right<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_up<T,R> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_phb_down<T,R> it,T value){m->data[it._rank] = value;}
    //-------------------------------------------------------------------------------
    //! Set value in the middle of the global matrix
    /*!
      \param val is the value to affect
    */
    //-------------------------------------------------------------------------------
    inline void setGlobalMiddleValue(T val){m->setGlobalMiddleValue(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values of the physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setPhysicalBorder(T val){m->setPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the right of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setRightPhysicalBorder(T val){m->setRightPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the left of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setLeftPhysicalBorder(T val){m->setLeftPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the up of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setUpPhysicalBorder(T val){m->setUpPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Set all the values on the down of physical border of the matrix
    /*!
      \param val is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setDownPhysicalBorder(T val){m->setDownPhysicalBorder(val);}
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> begin(){return m->begin();}
    inline iterator_cont<T,R> begin_cont(){return m->begin_cont();}
    inline iterator_rev<T,R> begin_rev(){return m->begin_rev();}
    inline iterator_step<T,R> begin_step(int step,int nb){return m->begin_step(step,nb);}
    inline iterator_phb_left<T,R> begin_phb_left(){return m->begin_phb_left();}
    inline iterator_phb_right<T,R> begin_phb_right(){return m->begin_phb_right();}
    inline iterator_phb_up<T,R> begin_phb_up(){return m->begin_phb_up();}
    inline iterator_phb_down<T,R> begin_phb_down(){return m->begin_phb_down();}
    inline iterator_line<T,R> begin_line(){return m->begin_line();}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> end(){return m->end();}
    inline iterator_cont<T,R> end_cont(){return m->end_cont();}
    inline iterator_rev<T,R> end_rev(){return m->end_rev();}
    inline iterator_step<T,R> end_step(){return m->end_step();}
    inline iterator_phb_left<T,R> end_phb_left(){return m->end_phb_left();}
    inline iterator_phb_right<T,R> end_phb_right(){return m->end_phb_right();}
    inline iterator_phb_up<T,R> end_phb_up(){return m->end_phb_up();}
    inline iterator_phb_down<T,R> end_phb_down(){return m->end_phb_down();}
    inline iterator_line<T,R> end_line(){return m->end_line();}
    //-------------------------------------------------------------------------------
    //! Operator [] to get and set an element of the matrix at iterator it
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \return the reference to the value at it
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_cont<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_rev<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_step<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_left<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_right<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_up<T,R> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_phb_down<T,R> it) { return m->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //! Operator = to assign the current DMatrix interface object to another one
    /*!
      \param right is the other DMatrix object
      \return the reference to the new current DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline DMatrix<T,R,false>& operator=(DMatrix<T,R,false> & right){m = right.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! Print the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs".
     */
    //-------------------------------------------------------------------------------
    inline void print(){m->print();}
    //-------------------------------------------------------------------------------
    //! Print all the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs". 
      It will also print the physical border of the DMatrix and the parallel overlap of the DMatrix.
     */
    //-------------------------------------------------------------------------------
    inline void printAll(){m->printAll();}
    //-------------------------------------------------------------------------------
    //! Saving in a bin file the distributed matrix interface
    /*!
      This method writes the DMatrix in a single binary output file.
      \param binFile is a binary output file to save the DMatrix in
    */
    //-------------------------------------------------------------------------------
    inline void write(char * binFile){m->write(binFile);}
    //-------------------------------------------------------------------------------
    //! get borders
    /*!
      To get the needed parallel overlap from other processors.
      MPI exchanges done.
    */
    //-------------------------------------------------------------------------------
    inline void getBorders(){m->getBorders();}
    //-------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRight(iterator<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_cont<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_step<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_rev<T,R> it){return allRight(it._rank);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return allRight(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeft(iterator<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_cont<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_step<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_rev<T,R> it){return allLeft(it._rank);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return allLeft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllUp(iterator<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_cont<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_step<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_rev<T,R> it){return allUp(it._rank,it._width);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return allUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllDown(iterator<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_cont<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_step<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_rev<T,R> it){return allDown(it._rank,it._width);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return allDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightDown(iterator<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_step<T,R> it){return allRightDown(it._rank,it._width);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return allRightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftDown(iterator<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return allLeftDown(it._rank,it._width);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return allLeftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllRightUp(iterator<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_step<T,R> it){return allRightUp(it._rank,it._width);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return allRightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor left up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    inline T * getAllLeftUp(iterator<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return allLeftUp(it._rank,it._width);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return allLeftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRight(iterator<T,R> it){return right(it._rank);}
    inline T getRight(iterator_cont<T,R> it){return right(it._rank);}
    inline T getRight(iterator_step<T,R> it){return right(it._rank);}
    inline T getRight(iterator_rev<T,R> it){return right(it._rank);}
    inline T getRight(iterator_phb_left<T,R> it){return right(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeft(iterator<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_cont<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_step<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_rev<T,R> it){return left(it._rank);}
    inline T getLeft(iterator_phb_right<T,R> it){return left(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getUp(iterator<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_cont<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_step<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_rev<T,R> it){return up(it._rank,it._width);}
    inline T getUp(iterator_phb_down<T,R> it){return up(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getDown(iterator<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_cont<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_step<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_rev<T,R> it){return down(it._rank,it._width);}
    inline T getDown(iterator_phb_up<T,R> it){return down(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightDown(iterator<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_cont<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_step<T,R> it){return rightDown(it._rank,it._width);}
    inline T getRightDown(iterator_rev<T,R> it){return rightDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftDown(iterator<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_cont<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_step<T,R> it){return leftDown(it._rank,it._width);}
    inline T getLeftDown(iterator_rev<T,R> it){return leftDown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getRightUp(iterator<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_cont<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_step<T,R> it){return rightUp(it._rank,it._width);}
    inline T getRightUp(iterator_rev<T,R> it){return rightUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    inline T getLeftUp(iterator<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_cont<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_step<T,R> it){return leftUp(it._rank,it._width);}
    inline T getLeftUp(iterator_rev<T,R> it){return leftUp(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline T * getAllX(iterator<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_cont<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_step<T,R> it){return allX(it._rank);}
    inline T * getAllX(iterator_rev<T,R> it){return allX(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    inline void getAllX(iterator<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_cont<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_rev<T,R> it, T * t){allX(it._rank,t);}
    inline void getAllX(iterator_step<T,R> it, T * t){allX(it._rank,t);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline T * getAllY(iterator<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_cont<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_step<T,R> it){return allY(it._rank,it._width);}
    inline T * getAllY(iterator_rev<T,R> it){return allY(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \param result table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    inline void getAllY(iterator<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_cont<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_rev<T,R> it, T * t){allY(it._rank,it._width,t);}
    inline void getAllY(iterator_step<T,R> it, T * t){allY(it._rank,it._width,t);}
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get8(iterator<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_cont<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_step<T,R> it){return eight(it._rank,it._width);}
    inline T * get8(iterator_rev<T,R> it){return eight(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    inline T * get4(iterator<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_cont<T,R> it){return four(it._rank,it._width);}
     inline T * get4(iterator_step<T,R> it){return four(it._rank,it._width);}
    inline T * get4(iterator_rev<T,R> it){return four(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInRight(iterator_phb_left<T,R> it){return inright(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInLeft(iterator_phb_right<T,R> it){return inleft(it._rank);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInUp(iterator_phb_down<T,R> it){return inup(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    inline T getInDown(iterator_phb_up<T,R> it){return indown(it._rank,it._width);}
    //-------------------------------------------------------------------------------
    //! Get all neighbor right values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right of the current it.
      This method exists for each type of iterator with right values in SkelGIS.
      \param it is the iterator position
      \return a table of right values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRight(iterator<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_cont<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_rev<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_step<T,R> it){return m->getAllRight(it);}
    inline T * getAllRight(iterator_phb_left<T,R> it){return m->getAllRight(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor left values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left of the current it.
      This method exists for each type of iterator with left values in SkelGIS.
      \param it is the iterator position
      \return a table of left values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeft(iterator<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_cont<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_rev<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_step<T,R> it){return m->getAllLeft(it);}
    inline T * getAllLeft(iterator_phb_right<T,R> it){return m->getAllLeft(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the up of the current it.
      This method exists for each type of iterator with up values in SkelGIS.
      \param it is the iterator position
      \return a table of up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllUp(iterator<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_cont<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_rev<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_step<T,R> it){return m->getAllUp(it);}
    inline T * getAllUp(iterator_phb_down<T,R> it){return m->getAllUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the down of the current it.
      This method exists for each type of iterator with down values in SkelGIS.
      \param it is the iterator position
      \return a table of down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllDown(iterator<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_cont<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_rev<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_step<T,R> it){return m->getAllDown(it);}
    inline T * getAllDown(iterator_phb_up<T,R> it){return m->getAllDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor right down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right down of the current it.
      This method exists for each type of iterator with right down values in SkelGIS.
      \param it is the iterator position
      \return a table of right down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRightDown(iterator<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_cont<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_rev<T,R> it){return m->getAllRightDown(it);}
    inline T * getAllRightDown(iterator_step<T,R> it){return m->getAllRightDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor left down values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left down of the current it.
      This method exists for each type of iterator with left down values in SkelGIS.
      \param it is the iterator position
      \return a table of left down values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeftDown(iterator<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_cont<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_rev<T,R> it){return m->getAllLeftDown(it);}
    inline T * getAllLeftDown(iterator_step<T,R> it){return m->getAllLeftDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all neighbor right up values for the element at iterator it
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the right up of the current it.
      This method exists for each type of iterator with right up values in SkelGIS.
      \param it is the iterator position
      \return a table of right up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllRightUp(iterator<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_cont<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_rev<T,R> it){return m->getAllRightUp(it);}
    inline T * getAllRightUp(iterator_step<T,R> it){return m->getAllRightUp(it);}*/
    //-------------------------------------------------------------------------------
    /*!
      The returned table will contain R elements. 
      These elements are the R values on the left up of the current it.
      This method exists for each type of iterator with left up values in SkelGIS.
      \param it is the iterator position
      \return a table of left up values
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllLeftUp(iterator<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_cont<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_rev<T,R> it){return m->getAllLeftUp(it);}
    inline T * getAllLeftUp(iterator_step<T,R> it){return m->getAllLeftUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right value in SkelGIS.
      \param it is the iterator position
      \return the right neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRight(iterator<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_cont<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_rev<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_step<T,R> it){return m->getRight(it);}
    inline T getRight(iterator_phb_left<T,R> it){return m->getRight(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left value in SkelGIS.
      \param it is the iterator position
      \return the left neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeft(iterator<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_cont<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_rev<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_step<T,R> it){return m->getLeft(it);}
    inline T getLeft(iterator_phb_right<T,R> it){return m->getLeft(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with up value in SkelGIS.
      \param it is the iterator position
      \return the up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getUp(iterator<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_cont<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_rev<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_step<T,R> it){return m->getUp(it);}
    inline T getUp(iterator_phb_down<T,R> it){return m->getUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with down value in SkelGIS.
      \param it is the iterator position
      \return the down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getDown(iterator<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_cont<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_rev<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_step<T,R> it){return m->getDown(it);}
    inline T getDown(iterator_phb_up<T,R> it){return m->getDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right down value in SkelGIS.
      \param it is the iterator position
      \return the right down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRightDown(iterator<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_cont<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_rev<T,R> it){return m->getRightDown(it);}
    inline T getRightDown(iterator_step<T,R> it){return m->getRightDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left down neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left down value in SkelGIS.
      \param it is the iterator position
      \return the left down neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeftDown(iterator<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_cont<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_rev<T,R> it){return m->getLeftDown(it);}
    inline T getLeftDown(iterator_step<T,R> it){return m->getLeftDown(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with right up value in SkelGIS.
      \param it is the iterator position
      \return the right up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getRightUp(iterator<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_cont<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_rev<T,R> it){return m->getRightUp(it);}
    inline T getRightUp(iterator_step<T,R> it){return m->getRightUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest left up neighbor value for the element at iterator it
    /*!
      This method exists for each type of iterator with left up value in SkelGIS.
      \param it is the iterator position
      \return the left up neighbor element value
    */
    //-------------------------------------------------------------------------------
    /*inline T getLeftUp(iterator<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_cont<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_rev<T,R> it){return m->getLeftUp(it);}
    inline T getLeftUp(iterator_step<T,R> it){return m->getLeftUp(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along X axe for the element at iterator it
    /*!
      Equivalent to getAllLeft and getAllRight with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along X
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllX(iterator<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_cont<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_rev<T,R> it){return m->getAllX(it);}
    inline T * getAllX(iterator_step<T,R> it){return m->getAllX(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all values along Y axe for the element at iterator it
    /*!
      Equivalent to getAllUp and getAllDown with a single call.
      This method exists for each type of iterator with left and right values in SkelGIS.
      \param it is the iterator position
      \return a table of all neighbor values along Y
    */
    //-------------------------------------------------------------------------------
    /*inline T * getAllY(iterator<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_cont<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_rev<T,R> it){return m->getAllY(it);}
    inline T * getAllY(iterator_step<T,R> it){return m->getAllY(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all 8 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getLeftUp, getUp, getRightUp, getRight, getRightDown, getDown, getLeftDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 8 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    /*inline T * get8(iterator<T,R> it){return m->get8(it);}
    inline T * get8(iterator_cont<T,R> it){return m->get8(it);}
    inline T * get8(iterator_rev<T,R> it){return m->get8(it);}
    inline T * get8(iterator_step<T,R> it){return m->get8(it);}*/
    //-------------------------------------------------------------------------------
    //! Get all 4 directions neighbor values for the element at iterator it
    /*!
      Equivalent to getUp, getRight, getDown and getLeft with a single call.
      \param it is the iterator position
      \return a table of all 4 directions neighbor values
    */
    //-------------------------------------------------------------------------------
    /*inline T * get4(iterator<T,R> it){return m->get4(it);};
    inline T * get4(iterator_cont<T,R> it){return m->get4(it);}
    inline T * get4(iterator_rev<T,R> it){return m->get4(it);}
    inline T * get4(iterator_step<T,R> it){return m->get4(it);}*/
    //-------------------------------------------------------------------------------
    //! Get the nearest right neighbor value inside the domain, for physical left border iterators
    /*!
      This work for every size of R and will always give the nearest right inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInRight(iterator_phb_left<T,R> it){return m->getInRight(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest left neighbor value inside the domain, for physical right border iterators
    /*!
      This work for every size of R and will always give the nearest left inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInLeft(iterator_phb_right<T,R> it){return m->getInLeft(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest up neighbor value inside the domain, for physical down border iterators
    /*!
      This work for every size of R and will always give the nearest up inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInUp(iterator_phb_down<T,R> it){return m->getInUp(it);}
    //-------------------------------------------------------------------------------
    //! Get the nearest down neighbor value inside the domain, for physical up border iterators
    /*!
      This work for every size of R and will always give the nearest down inside-domain value.
      \param it is the physical border iterator position
      \return the neighbor value
    */
    //-------------------------------------------------------------------------------
    //inline T getInDown(iterator_phb_up<T,R> it){return m->getInDown(it);}
    //-------------------------------------------------------------------------------
    //! Get the iterator on the matrix at position (x,y)
    /*!
      This method exists for the three basic types of iterators.
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R> getIterator(int x,int y){return m->getIterator(x,y);}
    inline iterator<T,R> getIterator_cont(int x,int y){return m->getIterator_cont(x,y);}
    inline iterator<T,R> getIterator_rev(int x,int y){return m->getIterator_rev(x,y);}
    //-------------------------------------------------------------------------------
    //! get the indexes of the iterator on the matrix
    /*!
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    //-------------------------------------------------------------------------------
    inline void getIndexes(iterator<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_cont<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_rev<T,R> it,int &col,int &li){m->getIndexes(it,col,li);}
    //-------------------------------------------------------------------------------
    //! Get the indexes of the iterator on the matrix
    /*!
      This method exists for the three basic types of iterators.
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    //-------------------------------------------------------------------------------
    inline HEADER getGlobalHeader(){return m->head;}
    //-------------------------------------------------------------------------------
    //! Get the header of the DMatrix
    /*!
      \return the header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getHeader(){return m->loc_head;}
    //-------------------------------------------------------------------------------
    //! Get the border header of the DMatrix
    /*!
      \return the border header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getBorderHeader(){return m->border_head;}
    //-------------------------------------------------------------------------------
    //! Get the number of columns in the domain decomposition
    /*!
      \return the number of columns in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getCols(){return m->cols;}
    //-------------------------------------------------------------------------------
    //! Get the number of rows in the domain decomposition
    /*!
      \return the number of rows in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getRows(){return m->rows;}
    //-------------------------------------------------------------------------------
    //! Get the column number of the DMatrix
    /*!
      \return the column number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getCol(){return m->col;}
    //-------------------------------------------------------------------------------
    //! Get the row number of the DMatrix
    /*!
      \return the row number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getRow(){return m->row;}
    //-------------------------------------------------------------------------------
    //! Get the data table of the DMatrix
    /*!
      \return the data table of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline T * getData(){return m->data;}
    //-------------------------------------------------------------------------------
    //! Get the recovery of the DMatrix (=R)
    /*!
      \return the recovery of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getBorder(){return R;}
    //-------------------------------------------------------------------------------
    //! Get the DMatrix_impl associated to the DMatrix interface
    /*!
      \return the DMatrix_impl
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,R,false> * getDMatrix(){return m;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //! DMatrix class
  /*!
    template of the distributed matrix of SkelGIS. This is the fourth specialization of the DMatrix.
    \tparam T is the type of data to store in the DMatrix
    R overlap template parameter is not asked and specialized to 0.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block).
  */
  //-------------------------------------------------------------------------------
  template<class T> struct DMatrix<T,0,false>
  //-------------------------------------------------------------------------------
  {
  protected:
    //! Pointer to DMatrix_impl object
    /*!
      The pointer to Dmatrix_impl object is the real distributed matrix object. DMatrix class is the user interface to this object.
    */
    DMatrix_impl<T,0,false> * m;

  public:
    //! default constructor of the distributed matrix interface
    //-------------------------------------------------------------------------------
    DMatrix(){}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from pointer
    /*!
      This Constructor is used in the user function when a skeleton ApplyList is called.
      \param _m is a DMatrix_impl pointer
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix_impl<T,0,false> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface with another DMatrix
    /*!
      It is possible to construct a DMatrix instance from another DMatrix instance
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    DMatrix(DMatrix<T,0,false>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a HEADER and a default value
    /*!
      The HEADER essentially define the size of the DMatrix, while the default value is given to each element of the DMatrix.
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,0,false>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! constructor of the distributed matrix interface from a binary file
    /*!
      A specific binary file format is read by SkelGIS. A Dmatrix instance can be constructed from a binary file.
      \param binFile is a binary input file to construct the DMatrix with
      \param loc is a boolean value to indicate if the DMatrix_impl has to stay local to one MPI process (used in ApplyReduction).
    */
    //-------------------------------------------------------------------------------
    DMatrix(const char * binFile, bool loc=false){m = new DMatrix_impl<T,0,false>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! destructor of the distributed matrix
    //-------------------------------------------------------------------------------
    ~DMatrix(){}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from pointer
    /*!
      The equivalent to the constructor from a pointer
      \param _m is a DMatrix_impl object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix_impl<T,0,false> * _m){m=_m;}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from another DMatrix
    /*!
      The equivalent to the constructor from another DMatrix
      \param _m is a DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline void init(DMatrix<T,0,false>& _m){m=_m.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a HEADER and a default value
    /*!
      The equivalent to the constructor from a HEADER and a default value
      \param h is the header to construct the object (see skelgis/util/utility.hpp)
      \param value is the default value to put in the matrix
    */
    //-------------------------------------------------------------------------------
    inline void init(HEADER h,const T value, bool loc=false){m = new DMatrix_impl<T,0,false>(h,value,loc);}
    //-------------------------------------------------------------------------------
    //! initialization of the distributed matrix interface from a binary file
    /*!
      The equivalent to the constructor from a binary file
      \param binFile is a binary input file to construct the DMatrix with
    */
    //-------------------------------------------------------------------------------
    inline void init(const char * binFile, bool loc=false){m = new DMatrix_impl<T,0,false>(binFile,loc);}
    //-------------------------------------------------------------------------------
    //! to erase and clear the DMatrix
    /*!
      The equivalent to the erase of a std::vector in the STL.
      The erase will empty the DMatrix and will deallocate the memory space associated.
      This has to be done when the DMatrix is not usefull anymore.
    */
    //-------------------------------------------------------------------------------
    inline void erase(){delete m;}
    //-------------------------------------------------------------------------------
    //! Methods to get the value of the matrix at the iterator it (alternative to [it])
    /*!
      This method exists for each type of iterator in SkelGIS. This method is const as the value returned could not be modified.
      \param it is the iterator position
      \return the value at the iterator it
    */
    //-------------------------------------------------------------------------------
    inline T getValue(iterator<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_cont<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_step<T,0> it) const {return m->data[it._rank];}
    inline T getValue(iterator_rev<T,0> it) const {return m->data[it._rank];}
    //-------------------------------------------------------------------------------
    //! Methods to set the value of the matrix at the iterator it (alternative to [it]=)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \param value is the value to set
    */
    //-------------------------------------------------------------------------------
    inline void setValue(iterator<T,0> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_cont<T,0> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_step<T,0> it,T value){m->data[it._rank] = value;}
    inline void setValue(iterator_rev<T,0> it,T value){m->data[it._rank] = value;}
    //-------------------------------------------------------------------------------
    //! Get the beginning iterator of the DMatrix (on the first element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the first element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> begin(){return m->begin();}
    inline iterator_cont<T,0> begin_cont(){return m->begin_cont();}
    inline iterator_rev<T,0> begin_rev(){return m->begin_rev();}
    inline iterator_step<T,0> begin_step(int step,int nb){return m->begin_step(step,nb);}
    inline iterator_line<T,0> begin_line(){return m->begin_line();}
    //-------------------------------------------------------------------------------
    //! Get the ending iterator of the DMatrix (on the last element of the DMatrix)
    /*!
      This method exists for each type of iterator in SkelGIS.
      \return the iterator of the last element
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> end(){return m->end();}
    inline iterator_cont<T,0> end_cont(){return m->end_cont();}
    inline iterator_rev<T,0> end_rev(){return m->end_rev();}
    inline iterator_step<T,0> end_step(){return m->end_step();}
    inline iterator_line<T,0> end_line(){return m->end_line();}
    //-------------------------------------------------------------------------------
    //! Operator [] to get and set an element of the matrix at iterator it
    /*!
      This method exists for each type of iterator in SkelGIS.
      \param it is the iterator position
      \return the reference to the value at it
    */
    //-------------------------------------------------------------------------------
    inline T& operator[](iterator<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_cont<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_rev<T,0> it) { return m->data[it._rank]; }
    inline T& operator[](iterator_step<T,0> it) { return m->data[it._rank]; }
    //-------------------------------------------------------------------------------
    //! Operator = to assign the current DMatrix interface object to another one
    /*!
      \param right is the other DMatrix object
      \return the reference to the new current DMatrix object
    */
    //-------------------------------------------------------------------------------
    inline DMatrix<T,0,false>& operator=(DMatrix<T,0,false> & right){m = right.getDMatrix();}
    //-------------------------------------------------------------------------------
    //! Print the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs".
     */
    //-------------------------------------------------------------------------------
    inline void print(){m->print();}
    //-------------------------------------------------------------------------------
    //! Print all the matrix by block
    /*!
      This method will print the DMatrix by processors in the directory "outputs". 
      It will also print the physical border of the DMatrix and the parallel overlap of the DMatrix.
     */
    //-------------------------------------------------------------------------------
    inline void printAll(){m->printAll();}
    //-------------------------------------------------------------------------------
    //! Saving in a bin file the distributed matrix interface
    /*!
      This method writes the DMatrix in a single binary output file.
      \param binFile is a binary output file to save the DMatrix in
    */
    //-------------------------------------------------------------------------------
    inline void write(char * binFile){m->write(binFile);}
    //-------------------------------------------------------------------------------
    //! Get the iterator on the matrix at position (x,y)
    /*!
      This method exists for the three basic types of iterators.
      \param x is row index
      \pram y is column index
      \return returns the iterator at position (x,y)
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0> getIterator(int x,int y){return m->getIterator(x,y);}
    inline iterator<T,0> getIterator_cont(int x,int y){return m->getIterator_cont(x,y);}
    inline iterator<T,0> getIterator_rev(int x,int y){return m->getIterator_rev(x,y);}
    //-------------------------------------------------------------------------------
    //! Get the indexes of the iterator on the matrix
    /*!
      This method exists for the three basic types of iterators.
      \param it is the iterator from which the indexes will be taken
      \param x is the iterator position x returned
      \param y is the iterator position y returned
    */
    //-------------------------------------------------------------------------------
    inline void getIndexes(iterator<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_cont<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    inline void getIndexes(iterator_rev<T,0> it,int &col,int &li){m->getIndexes(it,col,li);}
    //-------------------------------------------------------------------------------
    //! Get the global header of the DMatrix
    /*!
      \return the global header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getGlobalHeader(){return m->head;}
    //-------------------------------------------------------------------------------
    //! Get the header of the DMatrix
    /*!
      \return the header
    */
    //-------------------------------------------------------------------------------
    inline HEADER getHeader(){return m->loc_head;}
    //-------------------------------------------------------------------------------
    //! Get the number of columns in the domain decomposition
    /*!
      \return the number of columns in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getCols(){return m->cols;}
    //-------------------------------------------------------------------------------
    //! Get the number of rows in the domain decomposition
    /*!
      \return the number of rows in the domain decomposition
    */
    //-------------------------------------------------------------------------------
    inline int getRows(){return m->rows;}
    //-------------------------------------------------------------------------------
    //! Get the column number of the DMatrix
    /*!
      \return the column number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getCol(){return m->col;}
    //-------------------------------------------------------------------------------
    //! Get the row number of the DMatrix
    /*!
      \return the row number of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline int getRow(){return m->row;}
    //-------------------------------------------------------------------------------
    //! Get the data table of the DMatrix
    /*!
      \return the data table of the DMatrix
    */
    //-------------------------------------------------------------------------------
    inline T * getData(){return m->data;}
    //-------------------------------------------------------------------------------
    //! Get the DMatrix_impl associated to the DMatrix interface
    /*!
      \return the DMatrix_impl
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,0,false> * getDMatrix(){return m;}
    //-------------------------------------------------------------------------------
  };
  
}

#endif
