/*! \file iterator.hpp
 *  \brief Basic iterator object definition. Two template specializations. Users do not have to construct an iterator object directly (these objects are constructed in the DMatrix) however operators and methods of the iterator are used by the user.
 */
#ifndef ITERATOR_H
#define ITERATOR_H

namespace skelgis{

  //================================================================================
  //!  Iterator class
  /*!
    template of the basic iterator of SkelGIS. This is the first specialization of iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator(const iterator<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1)%_width<(_width-R))
	  ++_rank;
      else
	  _rank = _rank+2*R+1;

      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator<T,R> _tmp = *this;
      if((_tmp._rank+1)%_tmp._width<(_tmp._width-R))
	  ++_tmp._rank;
      else
	  _tmp._rank = _tmp._rank+2*R+1;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator += of the iterator
    /*!
      This operation increments the position of the iterator with a given index and assigns the new rank to the current iterator.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator+=(int _d)
    //-------------------------------------------------------------------------------
    {
      if((_rank+_d)%_width<(_width-R))
	_rank = _rank + _d;
      else
	_rank = _rank+2*R+_d;

      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator -- pre-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator--()
    //-------------------------------------------------------------------------------
    {
      if((_rank-1)%_width>(R-1))
	--_rank;
      else
	_rank = _rank - 2*R -1;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator -- post-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator--(int)
    //-------------------------------------------------------------------------------
    {
      iterator<T,R> _tmp = *this;
      if((_tmp._rank-1)%_tmp._width>(R-1))
	_tmp._rank;
      else
	_tmp._rank = _tmp._rank - 2*R -1;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator -= of the iterator
    /*!
      This operation decrements the position of the iterator with a given index and assigns the new rank to the current iterator.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,R>& operator-=(int _d)
    //-------------------------------------------------------------------------------
    {
      if((_rank-_d)%_width>(R-1) && (_rank-_d)%_width<_rank%_width)
	_rank = _rank - _d;
      else
	_rank = _rank - 2*R - _d;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,R>& operator=(iterator<T,R> right)
    //-------------------------------------------------------------------------------
    {
      _rank = right._rank;
      _width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator<T,R>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator<T,R>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator<T,R>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator<T,R>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator<T,R>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator<T,R>& toComp) {return _rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //================================================================================
  //!  Iterator class
  /*!
    template of the basic iterator of SkelGIS. This is the second specialization of iterator class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    R is specialized with the value 0
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator<T,0>
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator():_rank(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator(const unsigned int r):_rank(r){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator(const iterator<T,0>& it):_rank(it._rank){}
    //------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0>& operator++(){++_rank;return *this;}
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0>& operator++(int){iterator<T,0> _tmp=*this;++_rank;return _tmp;}
    //-------------------------------------------------------------------------------
    //! operator += of the iterator
    /*!
      This operation increments the position of the iterator with a given index and assigns the new rank to the current iterator.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,0>& operator+=(int _d){_rank = _rank + _d;return *this;}
    //-------------------------------------------------------------------------------
    //! operator -- pre-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0>& operator--(){--_rank;return *this;}
    //-------------------------------------------------------------------------------
    //! operator -- post-decrementation of the iterator
    /*!
      This operation decrements the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0>& operator--(int){iterator<T,0> _tmp=*this;--_rank;return _tmp;}
    //-------------------------------------------------------------------------------
    //! operator -= of the iterator
    /*!
      This operation decrements the position of the iterator with a given index and assigns the new rank to the current iterator.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator<T,0>& operator-=(int _d){_rank = _rank - _d;return *this;}
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator<T,0>& operator=(iterator<T,0> right)
    //-------------------------------------------------------------------------------
    {
      this->_rank = right._rank;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator<T,0>& toTest) {return _rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator<T,0>& toTest) {return _rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator<T,0>& toComp) {return _rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator<T,0>& toComp) {return _rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator<T,0>& toComp) {return _rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator<T,0>& toComp) {return _rank>toComp._rank;}
  };
}

#endif

