/*! \file iterator_phb.hpp
 *  \brief This file contains the definitions of the four specific physical border iterators of skelgis. Two template specializations for each iterator. Users do not have to construct an iterator object directly (these objects are constructed in the DMatrix) however operators and methods of the iterator are used by the user.
 */
#ifndef ITERATOR_PHB_H
#define ITERATOR_PHB_H

namespace skelgis{

  //===============================================================================
  //!  Iterator class
  /*!
    template of the left physical border iterator of SkelGIS. 
    This is the first specialization of iterator_phb_left class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_phb_left
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_phb_left():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_phb_left(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_phb_left(const iterator_phb_left<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_phb_left(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_left<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1)%_width%R==0)
	_rank = _rank+_width-(R-1);
      else
	++_rank;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_left<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_phb_left<T,R> _tmp = *this;
      if((_rank+1)%_width%R==0)
	_rank = _rank+_width-(R-1);
      else
	++_rank;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_phb_left<T,R>& operator=(iterator_phb_left<T,R> right)
    //-------------------------------------------------------------------------------
    {
      this->_rank = right._rank;
      this->_width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_phb_left<T,R> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_phb_left<T,R>& toTest) {return this->_rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_phb_left<T,R>& toComp) {return this->_rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_phb_left<T,R>& toComp) {return this->_rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_phb_left<T,R>& toComp) {return this->_rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_phb_left<T,R>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //===============================================================================
  //!  Iterator class
  /*!
    template of the left physical border iterator of SkelGIS. Second specialization. In this case the overlap R is null, then their are no physical borders.
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_phb_left<T,0>{};
  //-------------------------------------------------------------------------------

  //===============================================================================
  //!  Iterator class
  /*!
    template of the up physical border iterator of SkelGIS. 
    This is the first specialization of iterator_phb_up class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_phb_up
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_phb_up():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_phb_up(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_phb_up(const iterator_phb_up<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_phb_up(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_up<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_up<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_cont<T,R> _tmp = *this;
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_phb_up<T,R>& operator=(iterator_phb_up<T,R> right)
    //-------------------------------------------------------------------------------
    {
      this->_rank = right._rank;
      this->_width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_phb_up<T,R> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_phb_up<T,R>& toTest) {return this->_rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_phb_up<T,R>& toComp) {return this->_rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_phb_up<T,R>& toComp) {return this->_rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_phb_up<T,R>& toComp) {return this->_rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_phb_up<T,R>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //===============================================================================
  //!  Iterator class
  /*!
    template of the up physical border iterator of SkelGIS. Second specialization. In this case the overlap R is null, then their are no physical borders.
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_phb_up<T,0>{};
  //-------------------------------------------------------------------------------

  //===============================================================================
  //!  Iterator class
  /*!
    template of the right physical border iterator of SkelGIS. 
    This is the first specialization of iterator_phb_right class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_phb_right
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_phb_right():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_phb_right(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_phb_right(const iterator_phb_right<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_phb_right(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_right<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1-_width+R)%_width%R==0)
	_rank = _rank+_width-(R-1);
      else
	++_rank;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_right<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_cont<T,R> _tmp = *this;
      if((_rank+1-_width+R)%_width%R<=(R-1))
	++_rank;
      else
	_rank = _rank+_width-(R-1);
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_phb_right<T,R>& operator=(iterator_phb_right<T,R> & right)
    //-------------------------------------------------------------------------------
    {
      this->_rank = right._rank;
      this->_width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_phb_right<T,R> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_phb_right<T,R>& toTest) {return this->_rank!=toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_phb_right<T,R>& toComp) {return this->_rank<=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_phb_right<T,R>& toComp) {return this->_rank<toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_phb_right<T,R>& toComp) {return this->_rank>=toComp._rank;}
    //-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_phb_right<T,R>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //===============================================================================
  //!  Iterator class
  /*!
    template of the right physical border iterator of SkelGIS. Second specialization. In this case the overlap R is null, then their are no physical borders.
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_phb_right<T,0>{};
  //-------------------------------------------------------------------------------
  

  //===============================================================================
  //!  Iterator class
  /*!
    template of the down physical border iterator of SkelGIS. 
    This is the first specialization of iterator_phb_down class.
    \tparam T is the type of data stored in the DMatrix to manipulate with the iterator
    \tparam R is the overlap distance needed by the calculation, and the same overlap than in the DMatrix to manipulate with this iterator.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct iterator_phb_down
  //-------------------------------------------------------------------------------
  {
  public:
    unsigned int _rank; /*!< current rank of the iterator */
    unsigned int _width; /*!< width of the DMatrix associated to this iterator */

    //! default constructor of the iterator
    //-------------------------------------------------------------------------------
    iterator_phb_down():_rank(0),_width(0){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator
    /*!
      \param r is the initial rank to assign to the iterator
      \param w is the DMatrix width to assign to the iterator
    */
    //-------------------------------------------------------------------------------
    iterator_phb_down(const unsigned int r,const unsigned int w):_rank(r),_width(w){}
    //-------------------------------------------------------------------------------
    //! constructor of the iterator from another iterator
    /*!
      \param it is the other iterator instance from which to construct the new one
    */
    //-------------------------------------------------------------------------------
    iterator_phb_down(const iterator_phb_down<T,R>& it):_rank(it._rank),_width(it._width){}
    //-------------------------------------------------------------------------------
    //! Destructor of the iterator
    //-------------------------------------------------------------------------------
    ~iterator_phb_down(){}
    //-------------------------------------------------------------------------------
    //! operator ++ pre-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_down<T,R>& operator++()
    //-------------------------------------------------------------------------------
    {
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator ++ post-incrementation of the iterator
    /*!
      This operation increments the position of the iterator in the DMatrix.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    iterator_phb_down<T,R>& operator++(int)
    //-------------------------------------------------------------------------------
    {
      iterator_cont<T,R> _tmp = *this;
      if((_rank+1)%_width<(_width-R))
	++_rank;
      else
	_rank = _rank+2*R+1;
      return _tmp;
    }
    //-------------------------------------------------------------------------------
    //! operator = of the iterator
    /*!
      This operation assigns another iterator informations to the current one.
      \return a reference to the resulting iterator itself
    */
    //-------------------------------------------------------------------------------
    inline iterator_phb_down<T,R>& operator=(iterator_phb_down<T,R> & right)
    //-------------------------------------------------------------------------------
    {
      this->_rank = right._rank;
      this->_width = right._width;
      return *this;
    }
    //-------------------------------------------------------------------------------
    //! operator == of the iterator
    /*!
      This operation checks the equality of two iterators with their rank values.
      \return true if the two iterator have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator==(const iterator_phb_down<T,R> & toTest) {return this->_rank==toTest._rank;}
    //-------------------------------------------------------------------------------
    //! operator != of the iterator
    /*!
      This operation checks the inequality of two iterators with their rank values.
      \return true if the two iterator do not have the same rank. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator!=(const iterator_phb_down<T,R>& toTest) {return this->_rank!=toTest._rank;}
//-------------------------------------------------------------------------------
    //! operator <= of the iterator
    /*!
      This operation checks if the current iterator rank is smaller or equal to the other one.
      \return true is the current iterator rank is smaller or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <=(const iterator_phb_down<T,R>& toComp) {return this->_rank<=toComp._rank;}
//-------------------------------------------------------------------------------
    //! operator < of the iterator
    /*!
      This operation checks if the current iterator rank is smaller than the other one.
      \return true is the current iterator rank is smaller. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator <(const iterator_phb_down<T,R>& toComp) {return this->_rank<toComp._rank;}
//-------------------------------------------------------------------------------
    //! operator >= of the iterator
    /*!
      This operation checks if the current iterator rank is biger or equal to the other one.
      \return true is the current iterator rank is biger or equal. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >=(const iterator_phb_down<T,R>& toComp) {return this->_rank>=toComp._rank;}
//-------------------------------------------------------------------------------
    //! operator > of the iterator
    /*!
      This operation checks if the current iterator rank is biger than the other one.
      \return true is the current iterator rank is biger. False otherwize.
    */
    //-------------------------------------------------------------------------------
    inline bool operator >(const iterator_phb_down<T,R>& toComp) {return this->_rank>toComp._rank;}
    //-------------------------------------------------------------------------------
  };

  //===============================================================================
  //!  Iterator class
  /*!
    template of the down physical border iterator of SkelGIS. Second specialization. In this case the overlap R is null, then their are no physical borders.
  */
  //-------------------------------------------------------------------------------
  template<class T> struct iterator_phb_down<T,0>{};
  //-------------------------------------------------------------------------------
  

}

#endif
