/*! \file apply_reduction.hpp
 *  \brief Definitions of the skeleton object ApplyReduction. This skeleton takes a user function and a DMatrix as inputs and return a single element.
 */
#ifndef APPLY_REDUCTION_H
#define APPLY_REDUCTION_H

namespace skelgis{
  
  //================================================================================
  //Types of user functions
  //================================================================================
  //! _ApplyReduction_Func structure
  /*!
    These objects are not directly used by the user, but the macros used by the user, to define its functions, use these. Four specializations of this structure are available.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template <typename T,int R,bool line=true> struct _ApplyReduction_Func
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix <T,R>&, T& ) const =0;
  };
  //================================================================================
  //! _ApplyReduction_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template <typename T> struct _ApplyReduction_Func<T,0>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,0>&, T&) const =0;
  };
  //================================================================================
  //! _ApplyReduction_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block)
  */
  //-------------------------------------------------------------------------------
  template <typename T, int R> struct _ApplyReduction_Func<T,R,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,R,false>&, T &) const =0;
  };
  //================================================================================
  //! _ApplyReduction_Func structure
  /*!
    \tparam T is the type of data to store in the input DMatrix
    R overlap template parameter is not asked and specialized to 0
    The line template parameter is specialized with the false value. In this case the parallel data distribution will be done along height and width (by block)
  */
  //-------------------------------------------------------------------------------
  template <typename T> struct _ApplyReduction_Func<T,0,false>
  //-------------------------------------------------------------------------------
  {
    virtual void operator()(DMatrix<T,0,false>&, T&) const =0;
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //Specializations of skeleton ApplyReduction
  //================================================================================
  //! ApplyReduction structure
  /*!
    These objects are directly called by the user to apply a function on a DMatrix.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
    \tparam line is the parallel distribution wanted, the default value is true, then the parallel distribution will be divided along height but not along width
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,bool line=true> struct ApplyReduction{};
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyReduction structure
  /*!
    First specialization with line to default values.
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct ApplyReduction<T,R>
  //-------------------------------------------------------------------------------
  {
    static T apply(const _ApplyReduction_Func<T,R>& func, DMatrix<T,R>& m)
    {
      T res;
      iterator<T,R> it = m.begin();
      MPI_Status status;
      int MyOwnTag = Mpi_::mpi_nb;

      //-----exchanges of data
      m.getBorders();
      //-----
      //-----call to f
      func(m,res);
      //-----

      if(MyOwnTag>1)
	{
	  if (Mpi_::mpi_rank==0)
	    {
	      HEADER h=m.getHeader();
	      h.width=Mpi_::mpi_nb;
	      h.height=1;
	      
	      DMatrix<T,R> mRes(h,1,true);
	      //affectation of the proc 0 to the temporary result matrix
	      it = mRes.begin();
	      mRes[it] = res;
	      ++it;
	      
	      T msg;
	      for (int i=1; i < Mpi_::mpi_nb; i++)
		{
		  char * get = new char[sizeof(T)];
		  //receive result from other proc , other than proc0
		  MPI_Recv(get, sizeof(T), MPI_BYTE, i, MyOwnTag, MPI_COMM_WORLD, &status);
		  memcpy(&msg, get, sizeof(T));
		  delete [] get;
		  
		  mRes[it] = msg;
		  ++it;
		}
	      //the temporary matrix now is analysed by the user function, to get only one output result
	      func(mRes,res);
	    }
	  else
	    MPI_Send(reinterpret_cast<char *>(&res), sizeof(T), MPI_BYTE, 0, MyOwnTag, MPI_COMM_WORLD);
	}

      char * get_glob = new char[sizeof(T)];
      get_glob = reinterpret_cast<char *>(&res);
      MPI_Bcast(get_glob,sizeof(T),MPI_BYTE,0,MPI_COMM_WORLD);
      memcpy(&res, get_glob, sizeof(T));

      return res;
    }
  };
  //-------------------------------------------------------------------------------
  
  //================================================================================
  //! ApplyReduction structure
  /*!
    Second specialization with line to default values and R=0
    \tparam T is the type of data to store in the input DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T> struct ApplyReduction<T,0>
  //-------------------------------------------------------------------------------
  {
    inline static T apply(const _ApplyReduction_Func<T,0>& func, DMatrix<T,0>& m)
    {
      T res;
      iterator<T,0> it = m.begin();
      MPI_Status status;
      int MyOwnTag = Mpi_::mpi_nb;

      //-----call to f
      func(m,res);
      //-----

      if(MyOwnTag>1)
	{
	  if (Mpi_::mpi_rank==0)
	    {
	      HEADER h=m.getHeader();
	      h.width=Mpi_::mpi_nb;
	      h.height=1;
	      
	      DMatrix<T,0> mRes(h,1,true);
	      //affectation of the proc 0 to the temporary result matrix
	      it = mRes.begin();
	      mRes[it] = res;
	      ++it;
	      
	      T msg;
	      for (int i=1; i < Mpi_::mpi_nb; i++)
		{
		  char * get = new char[sizeof(T)];
		  //receive result from other proc , other than proc0
		  MPI_Recv(get, sizeof(T), MPI_BYTE, i, MyOwnTag, MPI_COMM_WORLD, &status);
		  memcpy(&msg, get, sizeof(T));
		  delete [] get;
		  
		  mRes[it] = msg;
		  ++it;
		}
	      //the temporary matrix now is analysed by the user function, to get only one output result
	      func(mRes,res);
	    }
	  else
	    MPI_Send(reinterpret_cast<char *>(&res), sizeof(T), MPI_BYTE, 0, MyOwnTag, MPI_COMM_WORLD);
	}

      char * get_glob = new char[sizeof(T)];
      get_glob = reinterpret_cast<char *>(&res);
      MPI_Bcast(get_glob,sizeof(T),MPI_BYTE,0,MPI_COMM_WORLD);
      memcpy(&res, get_glob, sizeof(T));

      return res;
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyReduction structure
  /*!
    Third specialization with line=false
    \tparam T is the type of data to store in the input DMatrix
    \tparam R is the overlap distance needed by the calculation for the input DMatrix.
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct ApplyReduction<T,R,false>
  //-------------------------------------------------------------------------------
  {
    static T apply(const _ApplyReduction_Func<T,R,false>& func, DMatrix<T,R,false>& m)
    {
      T res;
      iterator<T,R> it = m.begin();
      MPI_Status status;
      int MyOwnTag = Mpi_::mpi_nb;

      //-----exchanges of data
      m.getBorders();
      //-----
      //-----call to f
      func(m,res);
      //-----

      if(MyOwnTag>1)
	{
	  if (Mpi_::mpi_rank==0)
	    {
	      HEADER h=m.getHeader();
	      h.width=Mpi_::mpi_nb;
	      h.height=1;
	      
	      DMatrix<T,R,false> mRes(h,1,true);
	      //affectation of the proc 0 to the temporary result matrix
	      it = mRes.begin();
	      mRes[it] = res;
	      ++it;
	      
	      T msg;
	      for (int i=1; i < Mpi_::mpi_nb; i++)
		{
		  char * get = new char[sizeof(T)];
		  //receive result from other proc , other than proc0
		  MPI_Recv(get, sizeof(T), MPI_BYTE, i, MyOwnTag, MPI_COMM_WORLD, &status);
		  memcpy(&msg, get, sizeof(T));
		  delete [] get;
		  
		  mRes[it] = msg;
		  ++it;
		}
	      //the temporary matrix now is analysed by the user function, to get only one output result
	      func(mRes,res);
	    }
	  else
	    MPI_Send(reinterpret_cast<char *>(&res), sizeof(T), MPI_BYTE, 0, MyOwnTag, MPI_COMM_WORLD);
	}

      char * get_glob = new char[sizeof(T)];
      get_glob = reinterpret_cast<char *>(&res);
      MPI_Bcast(get_glob,sizeof(T),MPI_BYTE,0,MPI_COMM_WORLD);
      memcpy(&res, get_glob, sizeof(T));

      return res;
    }
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! ApplyReduction structure
  /*!
    Fourth specialization with line=false and R=0
    \tparam T is the type of data to store in the input DMatrix
  */
  //-------------------------------------------------------------------------------
  template<class T> struct ApplyReduction<T,0,false>
  //-------------------------------------------------------------------------------
  {
    inline static T apply(const _ApplyReduction_Func<T,0,false>& func, DMatrix<T,0,false>& m)
    {
      T res;
      iterator<T,0> it = m.begin();
      MPI_Status status;
      int MyOwnTag = Mpi_::mpi_nb;

      //-----call to f
      func(m,res);
      //-----

      if(MyOwnTag>1)
	{
	  if (Mpi_::mpi_rank==0)
	    {
	      HEADER h=m.getHeader();
	      h.width=Mpi_::mpi_nb;
	      h.height=1;
	      
	      DMatrix<T,0,false> mRes(h,1,true);
	      //affectation of the proc 0 to the temporary result matrix
	      it = mRes.begin();
	      mRes[it] = res;
	      ++it;
	      
	      T msg;
	      for (int i=1; i < Mpi_::mpi_nb; i++)
		{
		  char * get = new char[sizeof(T)];
		  //receive result from other proc , other than proc0
		  MPI_Recv(get, sizeof(T), MPI_BYTE, i, MyOwnTag, MPI_COMM_WORLD, &status);
		  memcpy(&msg, get, sizeof(T));
		  delete [] get;
		  
		  mRes[it] = msg;
		  ++it;
		}
	      //the temporary matrix now is analysed by the user function, to get only one output result
	      func(mRes,res);
	    }
	  else
	    MPI_Send(reinterpret_cast<char *>(&res), sizeof(T), MPI_BYTE, 0, MyOwnTag, MPI_COMM_WORLD);
	}

      char * get_glob = new char[sizeof(T)];
      get_glob = reinterpret_cast<char *>(&res);
      MPI_Bcast(get_glob,sizeof(T),MPI_BYTE,0,MPI_COMM_WORLD);
      memcpy(&res, get_glob, sizeof(T));

      return res;
    }
  };
  //-------------------------------------------------------------------------------

}

#endif
