/*! \file dmatrix_vector.hpp
 *  \brief Definitions of the object DMatrix_vector used by the user when he's manipulating AppyList skeleton.
 */
#ifndef DMATRIX_VECTOR_H
#define DMATRIX_VECTOR_H

#include <vector>

namespace skelgis{

  //================================================================================
  //! DMatrix_vector class
  /*!
    template of DMatrix_vector of SkelGIS. This is the first specialization of the DMatrix_vector.
    This object has to be used for ApplyList skeleton.
    \tparam T is the type of data stored in all DMatrix of the DMatrix_vector
    \tparam R is the overlap distance for all DMatrix in DMatrix_vector
    \tparam line is the parallel distribution used for all Dmatrix in DMatrix_vector
  */
  //-------------------------------------------------------------------------------
  template<class T,int R,bool line=true> struct DMatrix_vector
  //-------------------------------------------------------------------------------
  {
  protected :
    //! std::vector to store pointers on DMatrix_impl
    /*!
      This is what is actually stored in the DMatrix_vector to hide pointers.
    */
    std::vector<DMatrix_impl<T,R>* > data;

  public:
    //-------------------------------------------------------------------------------
    //! default constructor of DMatrix_vector
    //-------------------------------------------------------------------------------
    DMatrix_vector(){}
    //-------------------------------------------------------------------------------
    //! destructor of the DMatrix_vector
    //-------------------------------------------------------------------------------
    ~DMatrix_vector(){}
    //-------------------------------------------------------------------------------
    //! push_back method of the DMatrix_vector class
    /*!
      This method add a new DMatrix element at the end of the DMatrix_vector. It works as the push_back method of the std::vector object.
      \param m is the new DMatrix to add
    */
    //-------------------------------------------------------------------------------
    inline void push_back(DMatrix<T,R>& m){data.push_back(m.getDMatrix());}
    //-------------------------------------------------------------------------------
    //! operator [] of DMatrix_vector
    /*!
      Similar to table brackets or to std::vector brackets to access an element of the DMatrix_vector
      \param n is the index element to access in the DMatrix_vector
      \return Returns a pointer to the element at position n in the DMatrix_vector.
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,R> * operator[](int n){return data[n];}
    //-------------------------------------------------------------------------------
    //! size method of the DMatrix_vector class
    /*!
      This method returns the size of the DMatrix_vector, in other words the number of elements.
    */
    //-------------------------------------------------------------------------------
    inline int size(){return data.size();}
    //-------------------------------------------------------------------------------
    //! clear method of the DMatrix_vector class
    /*!
      Removes all elements from the DMatrix_vector, leaving the container with a size of 0
    */
    //-------------------------------------------------------------------------------
    inline void clear(){data.clear();}
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------

  //================================================================================
  //! DMatrix_vector class
  /*!
    template of DMatrix_vector of SkelGIS. This is the first specialization of the DMatrix_vector.
    specialization with line=false
    a specialization with R=0 is not usefull for DMatrix_vector
    This object has to be used for ApplyList skeleton.
    \tparam T is the type of data stored in all DMatrix of the DMatrix_vector
    \tparam R is the overlap distance for all DMatrix in DMatrix_vector
  */
  //-------------------------------------------------------------------------------
  template<class T,int R> struct DMatrix_vector<T,R,false>
  //-------------------------------------------------------------------------------
  {
  protected :
    //! std::vector to store pointers on DMatrix_impl
    /*!
      This is what is actually stored in the DMatrix_vector to hide pointers.
    */
    std::vector<DMatrix_impl<T,R,false>* > data;

  public:
    //-------------------------------------------------------------------------------
    //! default constructor of DMatrix_vector
    //-------------------------------------------------------------------------------
    DMatrix_vector(){}
    //-------------------------------------------------------------------------------
    //! destructor of the DMatrix_vector
    //-------------------------------------------------------------------------------
    ~DMatrix_vector(){}
    //-------------------------------------------------------------------------------
    //-------------------------------------------------------------------------------
    //! push_back method of the DMatrix_vector class
    /*!
      This method add a new DMatrix element at the end of the DMatrix_vector. It works as the push_back method of the std::vector object.
      \param m is the new DMatrix to add
    */
    //-------------------------------------------------------------------------------
    inline void push_back(DMatrix<T,R,false>& m){data.push_back(m.getDMatrix());}
    //-------------------------------------------------------------------------------
    //! operator [] of DMatrix_vector
    /*!
      Similar to table brackets or to std::vector brackets to access an element of the DMatrix_vector
      \param n is the index element to access in the DMatrix_vector
      \return Returns a pointer to the element at position n in the DMatrix_vector.
    */
    //-------------------------------------------------------------------------------
    inline DMatrix_impl<T,R,false> * operator[](int n){return data[n];}
    //-------------------------------------------------------------------------------
    //! size method of the DMatrix_vector class
    /*!
      This method returns the size of the DMatrix_vector, in other words the number of elements.
    */
    //-------------------------------------------------------------------------------
    inline int size(){return data.size();}
    //-------------------------------------------------------------------------------
    //! clear method of the DMatrix_vector class
    /*!
      Removes all elements from the DMatrix_vector, leaving the container with a size of 0
    */
    //-------------------------------------------------------------------------------
    inline void clear(){data.clear();}
    //-------------------------------------------------------------------------------
  };
  //-------------------------------------------------------------------------------

}

#endif
